/* Shadekeep Mapping Tool ©2018-present Howard A. Kistler */

/* Constants */

const VERSION_NUMBER = '2.4';

const MODE_PAINT = 0;
const MODE_LINE  = 1;
const MODE_RECT  = 2;
const MODE_AREA  = 3;
const MODE_SKIP  = 4;
const MODE_SMART = 5;
const MODE_LROOM = 6;
const MODE_SROOM = 7;

const GRAIN_VERTICAL   = 0;
const GRAIN_HORIZONTAL = 1;

const LINEWEIGHT_PROPORTIONAL = 0;
const LINEWEIGHT_HAIRLINE     = 1;

const LAYER_ALL      = 0;
const LAYER_BASE     = 1;
const LAYER_OVERLAY  = 2;
const LAYER_FEATURE  = 3;
const LAYER_TERRAIN  = 4;
const LAYER_SYMBOLS  = 5;
const LAYER_DECOR    = 6;
const LAYER_MEGAHEX  = 9;
const LAYER_ANNOTATE = 19;

const HEX_EMPTY = 0;
const HEX_WHOLE = 1;
const HEX_FHALF = 10;
const HEX_PHALF = 20;
const HEX_APEX  = 30;
const HEX_BULK  = 40;
const HEX_CHANL = 50;
const HEX_JUNCT = 60;
const HEX_CORNR = 70;
const HEX_FINSH = 80;
const HEX_WIDE  = 90;
const HEX_JOIN  = 100;
const HEX_SLICE = 110;
const HEX_TWALL = 120;
const HEX_TEDGE = 130;
const HEX_TCORN = 140;
const HEX_BORDR = 200;
const HEX_LRSID = 210;
const HEX_LRWAL = 220;
const HEX_LRCNR = 230;
const HEX_SRSID = 240;
const HEX_SRCNR = 250;
const HEX_PILAR = 300;
const HEX_PSEUD = 400;

const MATERIAL_EMPTY = 0;

const FEATURE_EMPTY   = 0;
const FEATURE_DOOR    = 10;
const FEATURE_STAIRS  = 20;
const FEATURE_STAIR   = 30;
const FEATURE_SHAFT   = 40;
const FEATURE_UNDERS  = 50;
const FEATURE_UNDER   = 60;
const FEATURE_BANDS   = 70;
const FEATURE_BAND    = 80;
const FEATURE_DNDOOR  = 90;
const FEATURE_SPIRAL  = 110;
const FEATURE_WINDOW  = 120;
const FEATURE_MIDOOR  = 130;
const FEATURE_MIWIND  = 140;
const FEATURE_AXONS   = 150;
const FEATURE_AXON    = 160;
const FEATURE_TDOOR   = 170;
const FEATURE_TWIND   = 180;
const FEATURE_EDOOR   = 190;
const FEATURE_EWIND   = 200;
const FEATURE_ADOOR   = 210;
const FEATURE_AWIND   = 220;
const FEATURE_PILAR   = 300;

const TERRAIN_EMPTY    = 0;
const TERRAIN_BOUNDER  = 2000;
const TERRAIN_WATER    = 2100;
const TERRAIN_MOUNTOUT = 2200;
const TERRAIN_MOUNTFIL = 2250;
const TERRAIN_TREES    = 2300;
const TERRAIN_SWAMP    = 2400;
const TERRAIN_PLAINS   = 2500;
const TERRAIN_DESERT   = 2600;
const TERRAIN_CAPITOL  = 3000;
const TERRAIN_CITY     = 3100;
const TERRAIN_VILLAGE  = 3200;
const TERRAIN_FORT     = 3300;
const TERRAIN_MINE     = 3400;
const TERRAIN_RUIN     = 3500;
const TERRAIN_INTEREST = 4000;

const DECOR_EMPTY    = 0;
const DECOR_TRAP     = 6000;
const DECOR_TRAPDOOR = 6100;
const DECOR_FLOORPIT = 6150;
const DECOR_TRIPWIRE = 6200;
const DECOR_CHESTL   = 6300;
const DECOR_CHESTS   = 6350;
const DECOR_KEY      = 6400;
const DECOR_PENTAGRM = 6500;
const DECOR_GATE     = 6550;
const DECOR_LIGHT    = 6600;
const DECOR_SCONCE   = 6650;

const SYMBOL_EMPTY   = 0;
const SYMBOL_MARKER  = 100;
const SYMBOL_EVENT   = 1000;
const SYMBOL_ARROWEL = 8000;
const SYMBOL_ARROWES = 8050;
const SYMBOL_ARROWVL = 8100;
const SYMBOL_ARROWVS = 8150;
const SYMBOL_PROHIBL = 8200;
const SYMBOL_PROHIBS = 8250;
const SYMBOL_ACCENTL = 8500;
const SYMBOL_ACCENTC = 8510;
const SYMBOL_ACCENTT = 8520;
const SYMBOL_ACCENTD = 8530;

const ANNOTATION_BASE = 9000;

const PALETTE_SIZE = 16;

const PREVIEW_SCALE = 16;
const PREVIEW_SIZEX = 24;
const PREVIEW_SIZEY = 24;

const COMPONENT_UNSET = -1;
const FEATURE_UNSET   = -9;

const SHIFTMAP_LEFT   = 'sml';
const SHIFTMAP_RIGHT  = 'smr';
const SHIFTMAP_UP     = 'smu';
const SHIFTMAP_DOWN   = 'smd';
const SHIFTMAP_RIGHTK = 'smrk';
const SHIFTMAP_DOWNK  = 'smdk';

const NEWLINE = "\r\n";

const MAPFILE_EXTENSION = ".skmr";

// Data file constants
const CSVPREAMBLE = "data:text/csv;charset=utf-8,";
const FILEHEADER  = "SHADEKEEP:MAP:RAW";
const FILEKEYVERS = "SHAMATVER:";
const FILEKEYSIZE = "SIZE:";
const FILEKEYNAME = "NAME:";
const FILEKEYAUTH = "AUTHOR:";
const FILEKEYITER = "ITERATION:";
const FILEKEYGRAN = "GRAIN:";
const FILEKEYLWGT = "LINEWEIGHT:";
const FILEKEYSHAD = "SHADE:";
const FILEKEYZOOM = "ZOOM:";
const FILEKEYTRNO = "TRANSLUCENTON:";
const FILEKEYTRNA = "TRANSLUCENTAMT:";
const FILEKEYMHEX = "MHEX:";
const FILEKEYMHWT = "MHWT:";
const FILEKEYMHDT = "MHDT:";
const DATAHEADER  = "X,Y,[Terrains],{Overlays},(Features)";
const DATAHEX     = "H:";

// Backdrop image constants
const BDROPHDR    = "[-BACKDROPIMAGE-]"; // Backdrop data file section header
const BDROPDIMNW  = "BDROPDIMNW:";       // img.naturalWidth
const BDROPDIMNH  = "BDROPDIMNH:";       // img.naturalWidth
const BDROPDIMRW  = "BDROPDIMRW:";       // img.width (rendered width)
const BDROPDIMRH  = "BDROPDIMRH:";       // img.height (rendered height)
const BDROPOFFX   = "BDROPOFFX:";        // backdropOffset.x
const BDROPOFFY   = "BDROPOFFY:";        // backdropOffset.y
const BDROPDATA   = "BDROPDATA:";        // encoded image data block
const BDROPSHOW   = "BDROPSHOW:";        // is backdrop currrently showing
const BDROPLOCK   = "BDROPLOCK:";        // is backdrop currrently locked

const SEP_DATA = ",";
const SEP_LIST = "|";
const SEP_PAIR = ":";

const ZOOM_MIN  = 4;
const ZOOM_MAX  = 64;
const ZOOM_STEP = 2;

// The six cardinal directions relating to hex orientation
const DIR_HXC0 = 0; // Above in vertical grain, Upper Left in horizontal grain
const DIR_HXC1 = 1; // Upper Right in vertical grain, Upper Right in horizontal grain
const DIR_HXC2 = 2; // Lower Right in vertical grain, Right in horizontal grain
const DIR_HXC3 = 3; // Below in vertical grain, Lower Right in horizontal grain
const DIR_HXC4 = 4; // Lower Left in vertical grain, Lower Left in horizontal grain
const DIR_HXC5 = 5; // Upper Left in vertical grain, Left in horizontal grain
const NO_DIR   = -9;
const OFFGRID  = -1;

const MAX_GRID = 500;

const ANGLE_STEP = ((Math.PI * 2) / 360.0);
const CALCSQRT = Math.sqrt(3);
const CALCTHIN = CALCSQRT / 2;

const DEFAULT_LINE_THICKNESS = 1;

const MEGAHEX_TYPE_STANDARD = 0;
const MEGAHEX_TYPE_SQUARE = 1;
const MEGAHEX_WEIGHT_DEFAULT = 2;
const MEGADOT_WEIGHT_DEFAULT = 2;

const CARDINAL_N = 'N';
const CARDINAL_S = 'S';
const CARDINAL_E = 'E';
const CARDINAL_W = 'W';

const CLR_R = 'R';
const CLR_G = 'G';
const CLR_B = 'B';

const OPAQUE    = 1.0;
var   TRANSLUCE = 0.5;

const DEBUG_MODE = true;

/* Page Reload Interceptor */

window.addEventListener("beforeunload", function(evt){
	var msg = "Are you sure you wish to reload the page?<br/>You will lose any unsaved work.";
	(evt || window.event).returnValue = msg;
	return msg;
});

/* Input Interceptor & Hotkeys */

const HOTKEY_LEFT     = "hkleft";
const HOTKEY_RIGHT    = "hkright";
const HOTKEY_DOWN     = "hkdown";
const HOTKEY_UP       = "hkup";
const HOTKEY_ZOOM_IN  = "hkzoomin";
const HOTKEY_ZOOM_OUT = "hkzoomout";

const HOTKEY_MAP = {
	'a': HOTKEY_LEFT,
	'd': HOTKEY_RIGHT,
	's': HOTKEY_DOWN,
	'w': HOTKEY_UP,
	'ArrowLeft': HOTKEY_LEFT,
	'ArrowRight': HOTKEY_RIGHT,
	'ArrowDown': HOTKEY_DOWN,
	'ArrowUp': HOTKEY_UP,
	'+': HOTKEY_ZOOM_IN,
	'=': HOTKEY_ZOOM_IN,
	'-': HOTKEY_ZOOM_OUT
};

function debugMsg(msg) {
	if(DEBUG_MODE == true) {
		console.log(msg);
	}
}

function hotkeyPress(event) {
	if(document.activeElement.type == 'text') { return; }
	if(activeLayer == LAYER_ANNOTATE) {
		debugMsg(event.key.charCodeAt());
		annonum = 9000 + event.key.charCodeAt();
		if(event.keyCode == 32) {
			annonum = 9000;
		}
		if(document.getElementById('annotate_' + annonum) != null) {
		}
		else {
			annonum = 9000 + event.keyCode;
		}
		if(document.getElementById('annotate_' + annonum) != null) {
			setAnnotation(annonum);
			doubleBuffer();
			plotTool(cnv, lastHexCol, lastHexRow);
		}
		return;
	}
	if(HOTKEY_MAP.hasOwnProperty(event.key)) {
		event.preventDefault();
		refreshTool = false;
		switch(HOTKEY_MAP[event.key]) {
			case HOTKEY_LEFT:
				if((activeLayer == LAYER_BASE) || (activeLayer == LAYER_OVERLAY)) {
					rotateComponent(false);
					refreshTool = true;
				}
				else if(activeLayer == LAYER_FEATURE) {
					rotateFeature(false);
					refreshTool = true;
				}
				else if(activeLayer == LAYER_SYMBOLS) {
					rotateSymbol(false);
					refreshTool = true;
				}
				else if(activeLayer == LAYER_DECOR) {
					rotateDecor(false);
					refreshTool = true;
				}
				else if(activeLayer == LAYER_MEGAHEX) {
					shiftMegahexes(CARDINAL_W);
				}
				break;
			case HOTKEY_RIGHT:
				if((activeLayer == LAYER_BASE) || (activeLayer == LAYER_OVERLAY)) {
					rotateComponent(true);
					refreshTool = true;
				}
				else if(activeLayer == LAYER_FEATURE) {
					rotateFeature(true);
					refreshTool = true;
				}
				else if(activeLayer == LAYER_SYMBOLS) {
					rotateSymbol(true);
					refreshTool = true;
				}
				else if(activeLayer == LAYER_DECOR) {
					rotateDecor(true);
					refreshTool = true;
				}
				else if(activeLayer == LAYER_MEGAHEX) {
					shiftMegahexes(CARDINAL_E);
				}
				break;
			case HOTKEY_DOWN:
				if((activeLayer == LAYER_BASE) || (activeLayer == LAYER_OVERLAY)) {
					flipComponent();
					refreshTool = true;
				}
				else if(activeLayer == LAYER_FEATURE) {
					flipFeature();
					refreshTool = true;
				}
				else if(activeLayer == LAYER_SYMBOLS) {
					flipSymbol();
					refreshTool = true;
				}
				else if(activeLayer == LAYER_DECOR) {
					flipDecor();
					refreshTool = true;
				}
				else if(activeLayer == LAYER_MEGAHEX) {
					shiftMegahexes(CARDINAL_S);
				}
				break;
			case HOTKEY_UP:
				if(activeLayer == LAYER_MEGAHEX) {
					shiftMegahexes(CARDINAL_N);
				}
				break;
			case HOTKEY_ZOOM_IN:
				zoom(true);
				break;
			case HOTKEY_ZOOM_OUT:
				zoom(false);
				break;
			default:
				debugMsg(HOTKEY_MAP[event.key]);
				break;
		}
		if(refreshTool) {
			doubleBuffer();
			plotTool(cnv, lastHexCol, lastHexRow);
		}
		if(typeof event.stopPropagation != 'undefined') {
			event.stopPropagation();
		}
		else {
			event.cancelBubble = true;
		}
	}
}

/* Objects */

function Point(x, y) {
	this.x = x;
	this.y = y;
}

function HexElement(typ, mater) {
	this.ctype = typ;
	this.material = mater;
	this.setMaterial = function(m) { this.material = m; }
}

function Hex(x, y) {
	this.x = x;
	this.y = y;
	this.bases = [];
	this.overlays = [];
	this.features = [];
	this.annotations = [];
	this.clearBases = function(mater) {
		this.bases = [];
		if(typeof mater != "undefined") {
			this.bases.push(new HexElement(HEX_WHOLE, mater));
		}
		else {
			this.bases.push(new HexElement(HEX_EMPTY, MATERIAL_EMPTY));
		}
	}
	this.clearOverlays = function() {
		this.overlays = [];
	}
	this.clearFeatures = function() {
		this.features = [];
	}
	this.clearAnnotations = function() {
		this.annotations = [];
	}
	this.clearAll = function() {
		this.clearBases();
		this.clearOverlays();
		this.clearFeatures();
		this.clearAnnotations();
	}
	this.addBase = function(typ, mater) {
		var needToAdd = true;
		if(this.bases.length > 0) {
			/* whole and empty hexes are special cases, as they obliterate all other previously assigned component types */
			if(typ == HEX_WHOLE || typ == HEX_EMPTY) {
				for(var i = 0; i < this.bases.length; i++) {
					if(this.bases[i].ctype == typ) {
						if(this.bases[i].material == mater) {
							needToAdd = false;
						}
						else {
							this.bases[i].setMaterial(mater);
							needToAdd = false;
						}
					}
					else {
						this.bases = [];
						needToAdd = true;
					}
				}
			}
			else {
				for(var i = 0; i < this.bases.length; i++) {
					if(this.bases[i].ctype == typ) {
						this.bases[i].setMaterial(mater);
						needToAdd = false;
					}
				}
			}
		}
		if(needToAdd) {
			this.bases.push(new HexElement(typ, mater));
		}
	}
	this.addOverlay = function(typ, mater) {
		var needToAdd = true;
		if(this.overlays.length > 0) {
			/* whole and empty hexes are special cases, as they obliterate all other previously assigned component types */
			if(typ == HEX_WHOLE || typ == HEX_EMPTY) {
				for(var i = 0; i < this.overlays.length; i++) {
					if(this.overlays[i].ctype == typ) {
						if(this.overlays[i].material == mater) {
							needToAdd = false;
						}
						else {
							this.overlays[i].setMaterial(mater);
							needToAdd = false;
						}
					}
					else {
						this.overlays = [];
						needToAdd = true;
					}
				}
			}
			else {
				for(var i = 0; i < this.overlays.length; i++) {
					if(this.overlays[i].ctype == typ) {
						this.overlays[i].setMaterial(mater);
						needToAdd = false;
					}
				}
			}
		}
		if(needToAdd) {
			this.overlays.push(new HexElement(typ, mater));
		}
	}
	this.addFeature = function(typ, mater) {
		var needToAdd = true;
		if(this.features.length > 0) {
			for(var i = 0; i < this.features.length; i++) {
				if(this.features[i].ctype == typ) {
					this.features[i].setMaterial(mater);
					needToAdd = false;
				}
			}
		}
		if(needToAdd) {
			this.features.push(new HexElement(typ, mater));
		}
	}
	this.addAnnotation = function(typ, mater) {
		if(this.annotations.length < 2) {
			this.annotations.push(new HexElement(parseInt(typ), mater));
		}
	}
	this.clone = function() {
		var newhex = new Hex(this.x, this.y);
		for(var b = 0; b < this.bases.length; b++) {
			newhex.addBase(this.bases[b].ctype, this.bases[b].material);
		}
		for(var o = 0; o < this.overlays.length; o++) {
			newhex.addOverlay(this.overlays[o].ctype, this.overlays[o].material);
		}
		for(var f = 0; f < this.features.length; f++) {
			newhex.addFeature(this.features[f].ctype, this.features[f].material);
		}
		for(var a = 0; a < this.annotations.length; a++) {
			newhex.addAnnotation(this.annotations[a].ctype, this.annotations[a].material);
		}
		return newhex;
	}
}

function RoomDef(x, y, w, h) {
	this.x = x;
	this.y = y;
	this.width = w;
	this.height = h;

	this.startX = this.x;
	this.startY = this.y;
	this.endX   = this.x + (this.width - 1);
	this.endY   = this.y + (this.height - 1);

	this.overlaps = function(room, buffer) {
		if(buffer == null) {
			buffer = 0;
		}
		return !(((this.startX > room.endX + buffer) || (this.endX + buffer < room.startX)) || ((this.startY > room.endY + buffer) || (this.endY + buffer < room.startY)));
	}
}

// Canvas Component Pointers
var cframe;
var canvas;
var cnv;
var buffer;
var buf;
var mhexes;
var mhx;
var preview;
var prv;
var fpreview;
var fpr;
var tpreview;
var tpr;
var spreview;
var spr;
var dpreview;
var dpr;

var mapGrain = GRAIN_VERTICAL; // GRAIN_VERTICAL || GRAIN_HORIZONTAL
var lineWeight = LINEWEIGHT_PROPORTIONAL; // LINEWEIGHT_PROPORTIONAL || LINEWEIGHT_HAIRLINE
var megahexWeight = MEGAHEX_WEIGHT_DEFAULT;
var megahexDotWeight = MEGADOT_WEIGHT_DEFAULT;

var modeStartX = -1;
var modeStartY = -1;

var drawsize = 16;
var drawscale = drawsize;
var halfscale = drawscale / 2;
var xoffset = drawscale;
var yoffset = (drawscale * CALCTHIN);

var outlineThickness = getOutlineThickness();

var defaultGridGrey = 127;
var defaultGridColor = "rgb(" + defaultGridGrey + ", " + defaultGridGrey + ", " + defaultGridGrey + ")";

var hexOutline    = defaultGridColor;
var hexBackground = "#ffffff";
var toolOutline   = "#000000";

var scratchTemp = 0;

var materialNames = [
	"Empty",
	"Black",
	"White",
	"Stone",
	"Earth",
	"Wood",
	"Metal",
	"Verdigris",
	"Azure",
	"Iron",
	"Yellow",
	"Orange",
	"Purple",
	"Blue",
	"Chartreuse",
	"Brown",
	"Crimson",
	"Terracotta",
	"Umber",
	"Gold",
	"Viridian",
	"Lapis",
	"Basalt",
	"Rose",
	"Lilac",
	"Bark",
	"Ochre",
	"Sun",
	"Green",
	"Cobalt",
	"Cyan",
	"Magenta",
	"Violet",
	"Pure White",
	"10% Grey",
	"20% Grey",
	"30% Grey",
	"40% Grey",
	"50% Grey",
	"60% Grey",
	"70% Grey",
	"80% Grey",
	"90% Grey",
	"Pure Black"
];

var hexBase = [
	hexBackground,
	"#000000",
	"#ffffff",
	"#cccccc",
	"#ccbcab",
	"#e6cda6",
	"#99a8c2",
	"#add498",
	"#7fc5e8",
	"#aaaaaa",
	"#e3e395",
	"#f78b0f",
	"#c693e6",
	"#5e88f2",
	"#6de8ad",
	"#875822",
	"#f05757",
	"#b07f71",
	"#ba3e12",
	"#cfc50b",
	"#4e9e29",
	"#5a79ad",
	"#6c6c6c",
	"#e394a5",
	"#f6bfff",
	"#bf9567",
	"#db6300",
	"#f0f01f",
	"#a0e35d",
	"#324ccf",
	"#00afef",
	"#ec268f",
	"#8e5aed",
	"#ffffff",
	"#e6e6e6",
	"#cccccc",
	"#b3b3b3",
	"#999999",
	"#808080",
	"#666666",
	"#4d4d4d",
	"#333333",
	"#191919",
	"#000000"
];

var hexOverlay = [
	hexBackground,
	"#000000",
	"#ffffff",
	"#808080",
	"#857c74",
	"#a88b5c",
	"#606978",
	"#7ea36d",
	"#5da0c2",
	"#606060",
	"#a1a158",
	"#ba680b",
	"#8d64a6",
	"#456ac7",
	"#3bb379",
	"#5e3a11",
	"#c23434",
	"#876359",
	"#8a2500",
	"#8f8800",
	"#336e18",
	"#3d537a",
	"#454545",
	"#a86876",
	"#c7a6cc",
	"#b0691c",
	"#874610",
	"#d3d312",
	"#7bad49",
	"#203187",
	"#0081b0",
	"#ba2273",
	"#6e44bd",
	"#ffffff",
	"#e6e6e6",
	"#cccccc",
	"#b3b3b3",
	"#999999",
	"#808080",
	"#666666",
	"#4d4d4d",
	"#333333",
	"#191919",
	"#000000"
];

var currentMaterPalette = 0;
var maxMaterPalette     = Math.ceil((materialNames.length - 1) / PALETTE_SIZE) - 1;
var currentCompPalette  = 0;
var maxCompPalette      = 1;
var currentFeatPalette  = 0;
var maxFeatPalette      = 1;
var currentSymbPalette  = 0;
var maxSymbPalette      = 1;

hexCursor = "#ff8f4f";

var defRows  = 34;
var defCols  = 51;
var gridx    = defCols;
var gridy    = defRows;
var hexarray = [];
var hexundo  = [];

var activeMode       = MODE_PAINT;
var activeLayer      = LAYER_BASE;
var activeComponent  = HEX_WHOLE;
var activeMaterial   = 1;
var activeFeature    = FEATURE_DOOR;
var activeTerrain    = TERRAIN_WATER;
var activeSymbol     = SYMBOL_ARROWEL;
var activeDecor      = DECOR_TRAP;
var activeAnnotation = 9065;

var lastHexCol = 0;
var lastHexRow = 0;
var paintOn = false;
var megahexesOn = false;

var leftmargin = 8;
var topmargin  = 8;

var mhexType = MEGAHEX_TYPE_STANDARD;
var mhexOverlayX = 1;
var mhexOverlayY = -6;
var mhexOverlayS = 0;
var mhexOutline = "#000000";

var backdropImage = null;
var backdropShowing = true;
var backdropOffset = new Point(0, 0);
var backdropLocked = false;
var translucencyEnabled = false;

var haveUndo = false;

var loadedMapName = "default";
var defaultAuthorName = "default";
var loadedAuthorName = defaultAuthorName;
var fileIteration = 0;

var prefRestoreCreator = defaultAuthorName;
var prefRestoreGridshade = defaultGridGrey;

/* Interface Construction */

function writePaletteButton(buttonId, func, args, img, label, isActive, hoverTip) {
	document.write(
		'<div role="button" class="palettetool' + (isActive ? ' toolactive' : '') + '"' +
			' id="' + buttonId + '"' +
			((hoverTip != undefined && hoverTip != null && hoverTip > '') ? ' title="' + hoverTip + '"' : '') +
			' onclick="' + func + '(' + (typeof args == 'string' ? "'" + args + "'" : args) + ');"><img src="img/' + img + '.png"/><br/>' + label +
		'</div>');
}

function writeCanvasButton(buttonId, canvasId, label, func, args) {
	document.write('<div class="palettetool notclickable" id="' + buttonId + '"' + ((func != '') ? ' onclick="' + func + '(' + args + ');"' : '') + '><canvas id="' + canvasId + '" height="48px" width="48px"></canvas><br/>' + label + '</div>');
}

function writeMaterialButton(materialId, isActive) {
	writePaletteButton('material' + materialId, 'setMaterial', materialId, 'material_' + materialId, materialNames[materialId], isActive);
}

function writeComponentButton(componentId, label, isActive) {
	writePaletteButton('tool' + componentId, 'setComponent', componentId, 'comp_' + componentId, label, isActive);
}

function writeFeatureButton(featureId, label, isActive) {
	writePaletteButton('feature' + featureId, 'setFeature', featureId, 'feature_' + featureId, label, isActive);
}

function writeTerrainButton(terrainId, label, isActive) {
	writePaletteButton('terrain' + terrainId, 'setTerrain', terrainId, 'terrain_' + terrainId, label, isActive);
}

function writeSymbolButton(symbolId, label, isActive) {
	writePaletteButton('symbol' + symbolId, 'setSymbol', symbolId, 'symbol_' + symbolId, label, isActive);
}

function writeDecorButton(decorId, label, isActive) {
	writePaletteButton('decor' + decorId, 'setDecor', decorId, 'decor_' + decorId, label, isActive);
}

function writeCustomMaterialButton(materialId, isActive) {
	writePaletteButton('customMaterial' + materialId, 'setCustomMaterial', materialId, 'material_' + materialId, materialNames[materialId], isActive);
}

function writeAnnotationButton(annotationId, isActive) {
	document.write('<div role="button" class="annotatetool' + (isActive ? ' toolactive' : '') + '" id="annotate_' + annotationId + '" onclick="setAnnotation(' + "'" + annotationId + "'" + ');"><span class="annosymb">' + (annotationId == ANNOTATION_BASE ? ' ' : String.fromCharCode(annotationId - ANNOTATION_BASE)) + '</span></div>');
}

function writeRotateButton(buttonId, func, isRotateRight) {
	writePaletteButton(buttonId, func, isRotateRight, (isRotateRight ? 'rotate_r' : 'rotate_l'), (isRotateRight ? 'Rotate Right' : 'Rotate Left'), false, (isRotateRight ? 'Hotkeys: Cursor Right, D' : 'Hotkeys: Cursor Left, A'));
}

function writeFlipButton(buttonId, func) {
	writePaletteButton(buttonId, func, '', 'flip', 'Flip', false, 'Hotkeys: Cursor Down, S');
}

/* Cookie Support */
function setCookie(cookieName, cookieValue) {
	var dt = new Date();
	var expdays = 365; // expire cookie in this many days
	dt.setTime(dt.getTime() + (expdays * 24 * 60 * 60 * 1000));
	expiry = dt.toUTCString();
	document.cookie = cookieName + '=' + encodeURIComponent(cookieValue) + '; expires=' + expiry + '; path=/; SameSite=Lax';
}

function getCookie(cookieName) {
	var cookieVals = document.cookie.split(';');
	var cookieItem = '';
	for(var cookieIndex in cookieVals) {
		cookieItem = cookieVals[cookieIndex].trim();
		var cookiePair = cookieItem.split('=');
		if(cookiePair[0] == cookieName) {
			return decodeURIComponent(cookiePair[1]);
		}
	}
	return null;
}

function deleteCookie(cookieName) {
	document.cookie = cookieName + '=; expires=Thu, 01 Jan 1970 00:00:01 GMT; path=/; SameSite=Lax';
}

function purgeCookies() {
	var cookieVals = document.cookie.split(';');
	var cookieItem = '';
	var cookiePair = [];
	for(var cookieIndex in cookieVals) {
		cookieItem = cookieVals[cookieIndex].trim();
		var cookiePair = cookieItem.split('=');
		document.cookie = cookiePair[0] + '=; expires=Thu, 01 Jan 1970 00:00:01 GMT; path=/; SameSite=Lax';
	}
}

function setPreferences() {
	var creatorName = document.getElementById('prefInputCreator').value;
	if(creatorName > '') {
		defaultAuthorName = creatorName;
		setCookie('creator', creatorName);
	}
	var defaultGrain = document.getElementById('prefMapgrain').value;
	setCookie('grainofmap', defaultGrain);
	var defaultLineWeight = document.getElementById('prefLineweight').value;
	setCookie('lineweight', defaultLineWeight);
	var defaultRows = document.getElementById('prefInputRows').value;
	if(defaultRows > '') {
		try {
			var dRows = parseInt(defaultRows);
			if(dRows > 0 && dRows <= MAX_GRID) {
				defRows = dRows;
				setCookie('baserows', defRows);
			}
		}
		catch(err) { debugMsg('Illegal entry for Rows: ' + defaultRows); }
	}
	var defaultCols = document.getElementById('prefInputCols').value;
	if(defaultCols > '') {
		try {
			var dCols = parseInt(defaultCols);
			if(dCols > 0 && dCols <= MAX_GRID) {
				defCols = dCols;
				setCookie('basecols', defCols);
			}
		}
		catch(err) { debugMsg('Illegal entry for Rows: ' + defaultRows); }
	}
	var chosenZoomLevel = document.getElementById('prefZoomlevel').value;
	setCookie('zoomlevel', chosenZoomLevel);
	var gridShadeGrey = document.getElementById('prefHexshade').value;
	setCookie('gridshade', gridShadeGrey);
}

function getPreferences() {
	var creatorCookie = getCookie('creator');
	if(creatorCookie != null) {
		defaultAuthorName = creatorCookie;
		document.getElementById('prefInputCreator').value = defaultAuthorName;
	}
	var rowsCookie = getCookie('baserows');
	if(rowsCookie != null) {
		defRows = parseInt(rowsCookie);
		gridy = defRows;
		document.getElementById('maprows').value = defRows;
	}
	var colsCookie = getCookie('basecols');
	if(colsCookie != null) {
		defCols = parseInt(colsCookie);
		gridx = defCols;
		document.getElementById('mapcols').value = defCols;
	}
	var grainCookie = getCookie('grainofmap');
	if(grainCookie != null) {
		mapGrain = grainCookie;
		document.getElementById('mapgrain').value = mapGrain;
		toggleRoomDrawMode();
	}
	var lineweightCookie = getCookie('lineweight');
	if(lineweightCookie != null) {
		lineWeight = lineweightCookie;
		document.getElementById('lineweight').value = lineWeight;
	}
	var zoomCookie = getCookie('zoomlevel');
	if(zoomCookie != null) {
		drawsize = parseInt(zoomCookie);
		recalculate();
	}
	var gridshadeCookie = getCookie('gridshade');
	if(gridshadeCookie != null) {
		defaultGridGrey = parseInt(gridshadeCookie);
		document.getElementById('hexshade').value = defaultGridGrey;
		document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((defaultGridGrey / 255) * 100)) + '%';
		document.getElementById('prefHexshade').value = defaultGridGrey;
		document.getElementById('showShadeLevelPref').innerHTML = (100 - Math.round((defaultGridGrey / 255) * 100)) + '%';
		defaultGridColor = "rgb(" + defaultGridGrey + "," + defaultGridGrey + "," + defaultGridGrey + ")";
		hexOutline = defaultGridColor;
	}
}

/* Methods */

function initialiseGrid() {
	hexarray = [];
	for(x = 0; x < gridx; x++) {
		hexarray[x] = [];
		for(y = 0; y < gridy; y++) {
			var hex = new Hex(x, y);
			hex.addBase(HEX_EMPTY, MATERIAL_EMPTY);
			hexarray[x][y] = hex;
		}
	}
	hexundo = hexarray.slice();
}

function initialiseFont(ctx, scl) {
	var fontStr = "bold 1px RobotoMono";
	ctx.font = fontStr;
	ctx.fillStyle = 0;
	ctx.textAlign = "center";
	ctx.textBaseline = "middle";
	ctx.fillText('A', -10, -10);
}

function initialise() {
	// load preferences from cookies
	getPreferences();
	// initialise graphical components
	cframe = document.getElementById('canvasframe');
	canvas = document.getElementById('canvas');
	cnv = canvas.getContext('2d');
	buffer = document.getElementById('buffer');
	buf = buffer.getContext('2d');
	mhexes = document.getElementById('mhexes');
	mhx = mhexes.getContext('2d');
	preview = document.getElementById('canvaspreview');
	prv = preview.getContext('2d');
	fpreview = document.getElementById('canvasfeature');
	fpr = fpreview.getContext('2d');
	tpreview = document.getElementById('canvasterrain');
	tpr = tpreview.getContext('2d');
	spreview = document.getElementById('canvassymbol');
	spr = spreview.getContext('2d');
	dpreview = document.getElementById('canvasdecor');
	dpr = dpreview.getContext('2d');
	updatePreview();
	updateFeaturePreview();
	updateTerrainPreview();
	updateDecorPreview();
	updateSymbolPreview();
	// install mouse listeners
	canvas.addEventListener('mousemove',  function(evt) { processMouseMove(evt); });
	canvas.addEventListener('mousedown',  function(evt) { togglePaintMode(evt, true); });
	canvas.addEventListener('mouseup',    function(evt) { togglePaintMode(evt, false); });
	canvas.addEventListener('mouseleave', function(evt) { togglePaintMode(evt, false); hideCursor(evt); });
	document.getElementById('mapname').value = 'default';
	document.getElementById('mapauthor').value = defaultAuthorName;
	document.getElementById('maprows').value = gridy;
	document.getElementById('mapcols').value = gridx;
	document.getElementById('zoomlevel').value = drawsize;
	document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
	document.getElementById('hexshade').value = defaultGridGrey;
	document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((defaultGridGrey / 255) * 100)) + '%';
	document.getElementById('translevel').value = TRANSLUCE;
	document.getElementById('showOpacity').innerHTML = TRANSLUCE;
	// force load the annotation font
	initialiseFont(cnv, drawsize);
	// initialise the grid
	initialiseGrid();
	// prep the canvas
	resizeCanvas();
	drawMap();
	refreshDisplaySpace();
	updateBackdropWidgets();
}

function doubleBuffer() {
	cnv.clearRect(0, 0, canvas.width, canvas.height);
	if(backdropShowing && backdropImage != null) {
		if(backdropImage.src != null && backdropImage.src > '') {
			cnv.drawImage(backdropImage, backdropOffset.x, backdropOffset.y, backdropImage.width, backdropImage.height);
		}
	}
	if(translucencyEnabled) { cnv.globalAlpha = TRANSLUCE; }
	cnv.drawImage(buffer, 0, 0);
	if(megahexesOn) {
		cnv.drawImage(mhexes, 0, 0);
	}
	if(translucencyEnabled) { cnv.globalAlpha = OPAQUE; }
}

function refreshDisplaySpace() {
	var pagex = window.innerWidth || document.documentElement.clientWidth || document.getElementsByTagName('body')[0].clientWidth;
	var pagey = window.innerHeight || document.documentElement.clientHeight || document.getElementsByTagName('body')[0].clientHeight;
	var usableHeight = pagey - 64;
	var displayHeight = Math.max(usableHeight, 820);
	document.getElementById('wrapper').style.height = displayHeight + "px";
	document.getElementById('editor').style.height = displayHeight + "px";
	document.getElementById('canvasframe').style.height = displayHeight + "px";
}

function setForm(formname) {
	if(document.getElementById('tab' + formname).className == "tab tabActive") {
	}
	else {
		document.getElementById('tabMap').className = "tab tabInactive";
		document.getElementById('tabCompose').className = "tab tabInactive";
//		document.getElementById('tabSpace').className = "tab tabInactive";
//		document.getElementById('tabCustom').className = "tab tabInactive";
//		document.getElementById('tabOpts').className = "tab tabInactive";
		document.getElementById('tabHelp').className = "tab tabInactive";
		document.getElementById('tabHotkeys').className = "tab tabInactive";
		document.getElementById('tabAbout').className = "tab tabInactive";
		document.getElementById('tab' + formname).className = "tab tabActive";
		document.getElementById('paletteMap').style.display = "none";
		document.getElementById('paletteCompose').style.display = "none";
//		document.getElementById('paletteSpace').style.display = "none";
//		document.getElementById('paletteCustom').style.display = "none";
//		document.getElementById('paletteOpts').style.display = "none";
		document.getElementById('paletteHelp').style.display = "none";
		document.getElementById('paletteHotkeys').style.display = "none";
		document.getElementById('paletteAbout').style.display = "none";
		document.getElementById('palette' + formname).style.display = "block";
	}
}

function getOutlineThickness() {
	if(lineWeight == LINEWEIGHT_PROPORTIONAL) {
		return Math.max(drawscale / 16, 1);
	}
	else if(lineWeight == LINEWEIGHT_HAIRLINE) {
		return 1;
	}
	else {
		return DEFAULT_LINE_THICKNESS;
	}
}

function getHexCol(x, r) {
	if(mapGrain == GRAIN_VERTICAL) {
		return Math.round((x - (leftmargin + xoffset)) / (drawscale * 1.5));
	}
	else {
		return Math.round((x - (leftmargin + xoffset) - ((r % 2) * (drawscale * CALCTHIN))) / (drawscale * CALCSQRT));
	}
}
function getHexRow(y, c) {
	if(mapGrain == GRAIN_VERTICAL) {
		return Math.round((y - (topmargin + yoffset) - ((c % 2) * (drawscale * CALCTHIN))) / (drawscale * CALCSQRT));
	}
	else {
		return Math.round((y - (topmargin + yoffset)) / (drawscale * 1.5));
	}
}

function getHexRawX(row, col, scl, xoff, lmrg) {
	if(mapGrain == GRAIN_VERTICAL) {
		return (col * (scl * 1.5)) + xoff + lmrg;
	}
	else {
		shft = (row % 2) * (scl * CALCTHIN);
		return (col * scl * CALCSQRT) + xoff + shft + lmrg;
	}
}
function getHexRawY(row, col, scl, yoff, tmrg) {
	if(mapGrain == GRAIN_VERTICAL) {
		shft = (col % 2) * (scl * CALCTHIN);
		return (row * scl * CALCSQRT) + yoff + shft + tmrg;
	}
	else {
		return (row * (scl * 1.5)) + yoff + tmrg;
	}
}

function getHexX(row, col) { return getHexRawX(row, col, drawscale, xoffset, leftmargin); }
function getHexY(row, col) { return getHexRawY(row, col, drawscale, yoffset, topmargin); }

function getHexAt(x, y, direction) {
	var pt = new Point(x, y);
	if(mapGrain == GRAIN_VERTICAL) {
		if(x % 2 == 0) {
			switch(direction) {
				case DIR_HXC0:
					pt.x = x;
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC1:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC2:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = y;
					break;
				case DIR_HXC3:
					pt.x = x;
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC4:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = y;
					break;
				case DIR_HXC5:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				default:
					break;
			}
		}
		else {
			switch(direction) {
				case DIR_HXC0:
					pt.x = x;
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC1:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = y;
					break;
				case DIR_HXC2:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC3:
					pt.x = x;
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC4:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC5:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = y;
					break;
				default:
					break;
			}
		}
	}
	else {
		if(y % 2 == 0) {
			switch(direction) {
				case DIR_HXC0:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC1:
					pt.x = x;
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC2:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = y;
					break;
				case DIR_HXC3:
					pt.x = x;
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC4:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC5:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = y;
					break;
				default:
					break;
			}
		}
		else {
			switch(direction) {
				case DIR_HXC0:
					pt.x = x;
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC1:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC2:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = y;
					break;
				case DIR_HXC3:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC4:
					pt.x = x;
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC5:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = y;
					break;
				default:
					break;
			}
		}
	}
	return pt;
}

function getMouseHex(x, y) {
	var boundbox = cframe.getBoundingClientRect();
	var mousex = (x - boundbox.left) + cframe.scrollLeft;
	var mousey = (y - boundbox.top) + cframe.scrollTop;
	var col, row;
	if(mapGrain == GRAIN_VERTICAL) {
		col = getHexCol(mousex, null);
		row = getHexRow(mousey, col);
	}
	else {
		row = getHexRow(mousey, null);
		col = getHexCol(mousex, row);
	}
	var mouseHex = { x: col, y: row };
	return mouseHex;
}

function getHexPoints(dscale) {
	var hscale = Math.max(dscale / 2, 1);
	var qscale = Math.max(dscale / 4, 1);
	var escale = Math.max(dscale / 8, 1);
	// returns a set of points with the apex points, the midpoints for each face, and the sub-mid points for each apex-midpoint pair
	// return the origin (0,0) as the 0 index item, the remaining points are the hexagon starting at the upper left and moving clockwise
	if(mapGrain == GRAIN_VERTICAL) {
		pptx = [
			0,
			-hscale,
			-qscale,
			0,
			qscale,
			hscale,
			hscale + escale,
			hscale + qscale,
			hscale + qscale + escale,
			dscale,
			hscale + qscale + escale,
			hscale + qscale,
			hscale + escale,
			hscale,
			qscale,
			0,
			-qscale,
			-hscale,
			-(hscale + escale),
			-(hscale + qscale),
			-(hscale + qscale + escale),
			-dscale,
			-(hscale + qscale + escale),
			-(hscale + qscale),
			-(hscale + escale)
		];
		ppty = [
			0,
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-((hscale + qscale) * CALCTHIN),
			-(hscale * CALCTHIN),
			-(qscale * CALCTHIN),
			0,
			(qscale * CALCTHIN),
			(hscale * CALCTHIN),
			((hscale + qscale) * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			((hscale + qscale) * CALCTHIN),
			(hscale * CALCTHIN),
			(qscale * CALCTHIN),
			0,
			-(qscale * CALCTHIN),
			-(hscale * CALCTHIN),
			-((hscale + qscale) * CALCTHIN)
		];
	}
	else {
		pptx = [
			0,
			-(dscale * CALCTHIN),
			-((hscale + qscale) * CALCTHIN),
			-(hscale * CALCTHIN),
			-(qscale * CALCTHIN),
			0,
			(qscale * CALCTHIN),
			(hscale * CALCTHIN),
			((hscale + qscale) * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			((hscale + qscale) * CALCTHIN),
			(hscale * CALCTHIN),
			(qscale * CALCTHIN),
			0,
			-(qscale * CALCTHIN),
			-(hscale * CALCTHIN),
			-((hscale + qscale) * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN)
		];
		ppty = [
			0,
			-hscale,
			-(hscale + escale),
			-(hscale + qscale),
			-(hscale + qscale + escale),
			-dscale,
			-(hscale + qscale + escale),
			-(hscale + qscale),
			-(hscale + escale),
			-hscale,
			-qscale,
			0,
			qscale,
			hscale,
			(hscale + escale),
			(hscale + qscale),
			(hscale + qscale + escale),
			dscale,
			(hscale + qscale + escale),
			(hscale + qscale),
			(hscale + escale),
			hscale,
			qscale,
			0,
			-qscale
		];
	}
	return [pptx, ppty];
}

function getHexFinePoints(ctrx, ctry, dscale) {
	var hscale = Math.max(dscale / 2, 1);
	var qscale = Math.max(dscale / 4, 1);
	var escale = Math.max(dscale / 8, 1);
	// return the centerpoint as the 0 index item, the remaining points are the hexagon starting at the upper left and moving clockwise
	// returns a large set of points, with the apex points, the midpoints for each face, and the sub-mid points for each apex-midpoint pair
	// there points are translated in reference to the hex centerpoint
	if(mapGrain == GRAIN_VERTICAL) {
		pptx = [
			ctrx,
			ctrx - hscale,
			ctrx - qscale,
			ctrx,
			ctrx + qscale,
			ctrx + hscale,
			ctrx + hscale + escale,
			ctrx + hscale + qscale,
			ctrx + hscale + qscale + escale,
			ctrx + dscale,
			ctrx + hscale + qscale + escale,
			ctrx + hscale + qscale,
			ctrx + hscale + escale,
			ctrx + hscale,
			ctrx + qscale,
			ctrx,
			ctrx - qscale,
			ctrx - hscale,
			ctrx - (hscale + escale),
			ctrx - (hscale + qscale),
			ctrx - (hscale + qscale + escale),
			ctrx - dscale,
			ctrx - (hscale + qscale + escale),
			ctrx - (hscale + qscale),
			ctrx - (hscale + escale)
		];
		ppty = [
			ctry,
			ctry - (dscale * CALCTHIN),
			ctry - (dscale * CALCTHIN),
			ctry - (dscale * CALCTHIN),
			ctry - (dscale * CALCTHIN),
			ctry - (dscale * CALCTHIN),
			ctry - ((hscale + qscale) * CALCTHIN),
			ctry - (hscale * CALCTHIN),
			ctry - (qscale * CALCTHIN),
			ctry,
			ctry + (qscale * CALCTHIN),
			ctry + (hscale * CALCTHIN),
			ctry + ((hscale + qscale) * CALCTHIN),
			ctry + (dscale * CALCTHIN),
			ctry + (dscale * CALCTHIN),
			ctry + (dscale * CALCTHIN),
			ctry + (dscale * CALCTHIN),
			ctry + (dscale * CALCTHIN),
			ctry + ((hscale + qscale) * CALCTHIN),
			ctry + (hscale * CALCTHIN),
			ctry + (qscale * CALCTHIN),
			ctry,
			ctry - (qscale * CALCTHIN),
			ctry - (hscale * CALCTHIN),
			ctry - ((hscale + qscale) * CALCTHIN)
		];
	}
	else {
		pptx = [
			ctrx,
			ctrx - (dscale * CALCTHIN),
			ctrx - ((hscale + qscale) * CALCTHIN),
			ctrx - (hscale * CALCTHIN),
			ctrx - (qscale * CALCTHIN),
			ctrx,
			ctrx + (qscale * CALCTHIN),
			ctrx + (hscale * CALCTHIN),
			ctrx + ((hscale + qscale) * CALCTHIN),
			ctrx + (dscale * CALCTHIN),
			ctrx + (dscale * CALCTHIN),
			ctrx + (dscale * CALCTHIN),
			ctrx + (dscale * CALCTHIN),
			ctrx + (dscale * CALCTHIN),
			ctrx + ((hscale + qscale) * CALCTHIN),
			ctrx + (hscale * CALCTHIN),
			ctrx + (qscale * CALCTHIN),
			ctrx,
			ctrx - (qscale * CALCTHIN),
			ctrx - (hscale * CALCTHIN),
			ctrx - ((hscale + qscale) * CALCTHIN),
			ctrx - (dscale * CALCTHIN),
			ctrx - (dscale * CALCTHIN),
			ctrx - (dscale * CALCTHIN),
			ctrx - (dscale * CALCTHIN)
		];
		ppty = [
			ctry,
			ctry - hscale,
			ctry - (hscale + escale),
			ctry - (hscale + qscale),
			ctry - (hscale + qscale + escale),
			ctry - dscale,
			ctry - (hscale + qscale + escale),
			ctry - (hscale + qscale),
			ctry - (hscale + escale),
			ctry - hscale,
			ctry - qscale,
			ctry,
			ctry + qscale,
			ctry + hscale,
			ctry + (hscale + escale),
			ctry + (hscale + qscale),
			ctry + (hscale + qscale + escale),
			ctry + dscale,
			ctry + (hscale + qscale + escale),
			ctry + (hscale + qscale),
			ctry + (hscale + escale),
			ctry + hscale,
			ctry + qscale,
			ctry,
			ctry - qscale
		];
	}
	return [pptx, ppty];
}

function getHexCoords(scaling) {
	return getHexPoints(scaling);
}

function getHexFineCoords(col, row) {
	plotx = getHexX(row, col);
	ploty = getHexY(row, col);
	return getHexFinePoints(plotx, ploty, drawscale);
}

function getHexRawCoords(scl, xoff, yoff) {
	plotx = getHexRawX(0, 0, scl, xoff, 0);
	ploty = getHexRawY(0, 0, scl, yoff, 0);
	return getHexFinePoints(plotx, ploty, scl);
}

function getNorthernHexPoints(ctrx, ctry, dscale) {
	var hscale = Math.max(dscale / 2, 1);
	var qscale = Math.max(dscale / 4, 1);
	var escale = Math.max(dscale / 8, 1);
	pptx = [
		ctrx,
		ctrx - hscale,
		ctrx - qscale,
		ctrx,
		ctrx + qscale,
		ctrx + hscale,
		ctrx + hscale + escale,
		ctrx + hscale + qscale,
		ctrx + hscale + qscale + escale,
		ctrx + dscale,
		ctrx + hscale + qscale + escale,
		ctrx + hscale + qscale,
		ctrx + hscale + escale,
		ctrx + hscale,
		ctrx + qscale,
		ctrx,
		ctrx - qscale,
		ctrx - hscale,
		ctrx - (hscale + escale),
		ctrx - (hscale + qscale),
		ctrx - (hscale + qscale + escale),
		ctrx - dscale,
		ctrx - (hscale + qscale + escale),
		ctrx - (hscale + qscale),
		ctrx - (hscale + escale)
	];
	ppty = [
		ctry,
		ctry - (dscale * CALCTHIN),
		ctry - (dscale * CALCTHIN),
		ctry - (dscale * CALCTHIN),
		ctry - (dscale * CALCTHIN),
		ctry - (dscale * CALCTHIN),
		ctry - ((hscale + qscale) * CALCTHIN),
		ctry - (hscale * CALCTHIN),
		ctry - (qscale * CALCTHIN),
		ctry,
		ctry + (qscale * CALCTHIN),
		ctry + (hscale * CALCTHIN),
		ctry + ((hscale + qscale) * CALCTHIN),
		ctry + (dscale * CALCTHIN),
		ctry + (dscale * CALCTHIN),
		ctry + (dscale * CALCTHIN),
		ctry + (dscale * CALCTHIN),
		ctry + (dscale * CALCTHIN),
		ctry + ((hscale + qscale) * CALCTHIN),
		ctry + (hscale * CALCTHIN),
		ctry + (qscale * CALCTHIN),
		ctry,
		ctry - (qscale * CALCTHIN),
		ctry - (hscale * CALCTHIN),
		ctry - ((hscale + qscale) * CALCTHIN)
	];
	return [pptx, ppty];
}

function rotatePoint(pt, angle) {
	var sine = Math.sin(angle * ANGLE_STEP);
	var cosine = Math.cos(angle * ANGLE_STEP);
	var rotx = (pt.x * cosine) - (pt.y * sine);
	var roty = (pt.x * sine) + (pt.y * cosine);
	return new Point(rotx, roty);
}

function rotatePolygon(pts, angle) {
	if(angle == 0) { return pts; }
	var rotpoly = [];
	for(var p = 0; p < pts.length; p++) {
		rotpoly.push(rotatePoint(pts[p], angle));
	}
	return rotpoly;
}

function renderPolygon(ctx, poly, ctrx, ctry, style, isFilled, lineThick) {
	ctx.beginPath();
	ctx.moveTo(poly[0].x + ctrx, poly[0].y + ctry);
	for(var p = 1; p < poly.length; p++) {
		ctx.lineTo(poly[p].x + ctrx, poly[p].y + ctry);
	}
	ctx.closePath();
	if(isFilled) {
		ctx.fillStyle = style;
		ctx.fill();
	}
	else {
		ctx.lineWidth = (lineThick === undefined ? outlineThickness : lineThick);
		ctx.strokeStyle = style;
		ctx.stroke();
		ctx.lineWidth = DEFAULT_LINE_THICKNESS;
	}
}

function drawPolygon(ctx, poly, ctrx, ctry, style, lineThick) {
	renderPolygon(ctx, poly, ctrx, ctry, style, false, lineThick);
}

function fillPolygon(ctx, poly, ctrx, ctry, style) {
	renderPolygon(ctx, poly, ctrx, ctry, style, true);
}

function drawComponent(ctx, orix, oriy, pptx, ppty, ctyp, matr) {
//	if(matr == MATERIAL_EMPTY) { return; }
	ctx.fillStyle = matr;
	var rotation = ctyp % 10;
	var basetyp = ctyp - (ctyp % 10);
	switch(basetyp) {
		case HEX_EMPTY:
		case HEX_WHOLE:
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(pptx[9], ppty[9]),
				new Point(pptx[13], ppty[13]),
				new Point(pptx[17], ppty[17]),
				new Point(pptx[21], ppty[21])
			];
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_FHALF:
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(pptx[9], ppty[9]),
				new Point(pptx[21], ppty[21])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_PHALF:
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[3], ppty[3]),
				new Point(pptx[15], ppty[15]),
				new Point(pptx[17], ppty[17]),
				new Point(pptx[21], ppty[21])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_APEX:
			var poly = [
				new Point(pptx[21], ppty[21]),
				new Point(pptx[1], ppty[1]),
				new Point(pptx[17], ppty[17])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_BULK:
			var poly = [
				new Point(pptx[21], ppty[21]),
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(pptx[13], ppty[13]),
				new Point(pptx[17], ppty[17])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_TWALL:
			var poly = [
				new Point(pptx[24], ppty[24]),
				new Point(pptx[6], ppty[6]),
				new Point(pptx[8], ppty[8]),
				new Point(pptx[22], ppty[22])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_TEDGE:
			var poly = [
				new Point(pptx[6], ppty[6]),
				new Point(pptx[8], ppty[8]),
				new Point(pptx[10], ppty[10]),
				new Point(pptx[12], ppty[12])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_TCORN:
			var poly = [];
			if(mapGrain == GRAIN_VERTICAL) {
				if(rotation == 1) {
					poly = [
						new Point(pptx[10], ppty[10]),
						new Point(pptx[12], ppty[12]),
						new Point(pptx[12], ppty[10])
					];
				}
				else if(rotation == 2) {
					poly = [
						new Point(pptx[14], ppty[14]),
						new Point(pptx[16], ppty[16]),
						new Point(pptx[15], ppty[12])
					];
				}
				else if(rotation == 3) {
					poly = [
						new Point(pptx[18], ppty[18]),
						new Point(pptx[20], ppty[20]),
						new Point(pptx[18], ppty[20])
					];
				}
				else if(rotation == 4) {
					poly = [
						new Point(pptx[22], ppty[22]),
						new Point(pptx[24], ppty[24]),
						new Point(pptx[24], ppty[22])
					];
				}
				else if(rotation == 5) {
					poly = [
						new Point(pptx[2], ppty[2]),
						new Point(pptx[4], ppty[4]),
						new Point(pptx[3], ppty[6])
					];
				}
				else {
					poly = [
						new Point(pptx[6], ppty[6]),
						new Point(pptx[8], ppty[8]),
						new Point(pptx[6], ppty[8])
					];
				}
			}
			else {
				if(rotation == 1) {
					poly = [
						new Point(pptx[14], ppty[14]),
						new Point(pptx[16], ppty[16]),
						new Point(pptx[16], ppty[14])
					];
				}
				else if(rotation == 2) {
					poly = [
						new Point(pptx[18], ppty[18]),
						new Point(pptx[20], ppty[20]),
						new Point(pptx[18], ppty[20])
					];
				}
				else if(rotation == 3) {
					poly = [
						new Point(pptx[22], ppty[22]),
						new Point(pptx[24], ppty[24]),
						new Point(pptx[20], ppty[23])
					];
				}
				else if(rotation == 4) {
					poly = [
						new Point(pptx[2], ppty[2]),
						new Point(pptx[4], ppty[4]),
						new Point(pptx[4], ppty[2])
					];
				}
				else if(rotation == 5) {
					poly = [
						new Point(pptx[6], ppty[6]),
						new Point(pptx[8], ppty[8]),
						new Point(pptx[6], ppty[8])
					];
				}
				else {
					poly = [
						new Point(pptx[10], ppty[10]),
						new Point(pptx[12], ppty[12]),
						new Point(pptx[8], ppty[11])
					];
				}
			}
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_CHANL:
			var poly = [
				new Point(pptx[2], ppty[2]),
				new Point(pptx[4], ppty[4]),
				new Point(pptx[14], ppty[14]),
				new Point(pptx[16], ppty[16])
			];
			poly = rotatePolygon(poly, rotation * 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_JUNCT:
			ctx.beginPath();
			ctx.arc(orix, oriy, (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[2]) : (ppty[11] - ppty[10])), 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			var poly = [
				new Point(pptx[4], ppty[4]),
				new Point(pptx[2], ppty[2]),
				new Point(((pptx[2] + pptx[16]) / 2), ((ppty[2] + ppty[16]) / 2)),
				new Point(((pptx[14] + pptx[4]) / 2), ((ppty[14] + ppty[4]) / 2))
			];
			poly = rotatePolygon(poly, rotation * 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_SLICE:
			var poly = [
				new Point(0, 0),
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_WIDE:
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(pptx[13], ppty[13]),
				new Point(pptx[17], ppty[17])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 0 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_JOIN:
			ctx.beginPath();
			ctx.arc(orix, oriy, (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[1]) : (ppty[11] - ppty[9])), 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(((pptx[5] + pptx[13]) / 2), ((ppty[5] + ppty[13]) / 2)),
				new Point(((pptx[1] + pptx[17]) / 2), ((ppty[1] + ppty[17]) / 2))
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 0 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_CORNR:
			ctx.beginPath();
			if(rotation == 1) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[7] + orix, ppty[7] + oriy);
					ctx.lineTo(orix, oriy);
					ctx.lineTo(pptx[23] + orix, ppty[23] + oriy);
				}
				else {
					ctx.moveTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[13] + oriy);
				}
			}
			else if(rotation == 2) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(Math.round((pptx[1] + pptx[17]) / 2) + orix, Math.round((ppty[1] + ppty[17]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[7] + orix, ppty[7] + oriy);
					ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[15] + orix, ppty[15] + oriy);
					ctx.lineTo(orix, oriy);
				}
			}
			else if(rotation == 3) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(Math.round((pptx[1] + pptx[17]) / 2) + orix, Math.round((ppty[1] + ppty[17]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[9] + oriy);
				}
			}
			else if(rotation == 4) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[11] + orix, ppty[11] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[19] + orix, ppty[19] + oriy);
					ctx.lineTo(orix, oriy);
				}
				else {
					ctx.moveTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
					ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[1] + oriy);
				}
			}
			else if(rotation == 5) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(Math.round((pptx[5] + pptx[13]) / 2) + orix, Math.round((ppty[5] + ppty[13]) / 2) + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				}
				else {
					ctx.moveTo(pptx[19] + orix, ppty[19] + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
					ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[3] + orix, ppty[3] + oriy);
					ctx.lineTo(orix, oriy);
				}
			}
			else {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(Math.round((pptx[5] + pptx[13]) / 2) + orix, Math.round((ppty[5] + ppty[13]) / 2) + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				}
				else {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[21] + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				}
			}
			ctx.closePath();
			ctx.fill();
			break;
		case HEX_FINSH:
			ctx.beginPath();
			if(rotation == 1) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(Math.round((pptx[3] + pptx[0]) / 2) + orix, Math.round((ppty[3] + ppty[0]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[9] + oriy);
				}
			}
			else if(rotation == 2) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(Math.round((pptx[5] + pptx[13]) / 2) + orix, Math.round((ppty[5] + ppty[13]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(Math.round((pptx[11] + pptx[0]) / 2) + orix, ppty[11] + oriy);
				}
			}
			else if(rotation == 3) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(Math.round((pptx[5] + pptx[13]) / 2) + orix, Math.round((ppty[5] + ppty[13]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[13] + oriy);
				}
			}
			else if(rotation == 4) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(Math.round((pptx[15] + pptx[0]) / 2) + orix, Math.round((ppty[15] + ppty[0]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[21] + orix, ppty[21] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[21] + oriy);
				}
			}
			else if(rotation == 5) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
					ctx.lineTo(Math.round((pptx[1] + pptx[17]) / 2) + orix, Math.round((ppty[1] + ppty[17]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[21] + orix, ppty[21] + oriy);
					ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(Math.round((pptx[23] + pptx[0]) / 2) + orix, ppty[23] + oriy);
				}
			}
			else {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(Math.round((pptx[1] + pptx[17]) / 2) + orix, Math.round((ppty[1] + ppty[17]) / 2) + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				}
				else {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[1] + oriy);
				}
			}
			ctx.closePath();
			ctx.fill();
			break;
		case HEX_BORDR:
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(pptx[6], ppty[6]),
				new Point(pptx[24], ppty[24])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_LRSID:
			if(rotation == 0 || rotation == 3) {
				ctx.beginPath();
				ctx.moveTo(((pptx[2] + pptx[3]) / 2) + orix, ((ppty[2] + ppty[3]) / 2) + oriy);
				ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, ((ppty[3] + ppty[4]) / 2) + oriy);
				ctx.lineTo(((pptx[14] + pptx[15]) / 2) + orix, ((ppty[14] + ppty[15]) / 2) + oriy);
				ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, ((ppty[15] + ppty[16]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1 || rotation == 4) {
				ctx.beginPath();
				ctx.moveTo(((pptx[6] + pptx[7]) / 2) + orix, ((ppty[6] + ppty[7]) / 2) + oriy);
				ctx.lineTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
				ctx.lineTo(((pptx[18] + pptx[19]) / 2) + orix, ((ppty[18] + ppty[19]) / 2) + oriy);
				ctx.lineTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2 || rotation == 5) {
				ctx.beginPath();
				ctx.moveTo(((pptx[10] + pptx[11]) / 2) + orix, ((ppty[10] + ppty[11]) / 2) + oriy);
				ctx.lineTo(((pptx[11] + pptx[12]) / 2) + orix, ((ppty[12] + ppty[11]) / 2) + oriy);
				ctx.lineTo(((pptx[22] + pptx[23]) / 2) + orix, ((ppty[22] + ppty[23]) / 2) + oriy);
				ctx.lineTo(((pptx[23] + pptx[24]) / 2) + orix, ((ppty[23] + ppty[24]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			break;
		case HEX_LRWAL:
			if(rotation == 0) {
				ctx.beginPath();
				ctx.moveTo(pptx[8] + orix, ppty[8] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[22] + orix, ppty[22] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1) {
				ctx.beginPath();
				ctx.moveTo(pptx[12] + orix, ppty[12] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[2] + orix, ppty[2] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2) {
				ctx.beginPath();
				ctx.moveTo(pptx[6] + orix, ppty[6] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[16] + orix, ppty[16] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 3) {
				ctx.beginPath();
				ctx.moveTo(pptx[10] + orix, ppty[10] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[20] + orix, ppty[20] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 4) {
				ctx.beginPath();
				ctx.moveTo(pptx[14] + orix, ppty[14] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[24] + orix, ppty[24] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 5) {
				ctx.beginPath();
				ctx.moveTo(pptx[18] + orix, ppty[18] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[4] + orix, ppty[4] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			break;
		case HEX_LRCNR:
			if(rotation == 0) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[8]) + orix, (ppty[8]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[8]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[6]) / 2) + orix, ((ppty[7] + ppty[6]) / 2) + oriy);
					ctx.lineTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
					ctx.lineTo((pptx[6]) + orix, ((ppty[10] + ppty[11]) / 2) + oriy);
					ctx.lineTo((pptx[12]) + orix, (ppty[12]) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.lineTo(((pptx[4] + pptx[5]) / 2) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 1) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (pptx[3] - pptx[2]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[6]) / 2) + orix, ((ppty[7] + ppty[6]) / 2) + oriy);
					ctx.lineTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[8] + pptx[18]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[8] + ppty[18]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[20] + pptx[6]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[11] + pptx[10]) / 2) + orix, ((ppty[11] + ppty[10]) / 2) + oriy);
					ctx.lineTo(((pptx[11] + pptx[12]) / 2) + orix, ((ppty[11] + ppty[12]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (ppty[11] - ppty[10]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[11] + pptx[10]) / 2) + orix, ((ppty[11] + ppty[10]) / 2) + oriy);
					ctx.lineTo(((pptx[11] + pptx[12]) / 2) + orix, ((ppty[11] + ppty[12]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[16] + pptx[15]) / 2) + orix, ((ppty[16] + ppty[15]) / 2) + oriy);
					ctx.lineTo(((pptx[14] + pptx[15]) / 2) + orix, ((ppty[14] + ppty[15]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[14] + pptx[4]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[14] + ppty[4]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[16] + pptx[2]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[16] + ppty[2]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 2) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[10]) + orix, (ppty[10]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[10]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[18]) / 2) + orix, ((ppty[19] + ppty[18]) / 2) + oriy);
					ctx.lineTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
					ctx.lineTo(((pptx[4] + pptx[5]) / 2) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.lineTo((pptx[14]) + orix, (ppty[14]) + oriy);
					ctx.lineTo((pptx[17]) + orix, (ppty[12]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 3) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[20]) + orix, (ppty[20]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[20]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
					ctx.lineTo(((pptx[19] + pptx[18]) / 2) + orix, ((ppty[19] + ppty[18]) / 2) + oriy);
					ctx.lineTo(((pptx[17] + pptx[16]) / 2) + orix, (Math.round((ppty[0] + ((ppty[11] + ppty[12]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.lineTo((pptx[24]) + orix, (ppty[24]) + oriy);
					ctx.lineTo((pptx[4]) + orix, ((ppty[11] + ppty[12]) / 2) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 4) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (pptx[3] - pptx[2]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[18]) / 2) + orix, ((ppty[19] + ppty[18]) / 2) + oriy);
					ctx.lineTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[20] + pptx[6]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[8] + pptx[18]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[8] + ppty[18]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[23] + pptx[22]) / 2) + orix, ((ppty[23] + ppty[22]) / 2) + oriy);
					ctx.lineTo(((pptx[23] + pptx[24]) / 2) + orix, ((ppty[23] + ppty[24]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (ppty[11] - ppty[10]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[4]) / 2) + orix, ((ppty[3] + ppty[4]) / 2) + oriy);
					ctx.lineTo(((pptx[3] + pptx[2]) / 2) + orix, ((ppty[3] + ppty[2]) / 2) + oriy);
					ctx.lineTo((Math.round((((pptx[3] + pptx[2]) / 2) + ((pptx[16] + pptx[15]) / 2)) / 2)) + orix, (Math.round((((ppty[3] + ppty[2]) / 2) + ((ppty[16] + ppty[15]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((((pptx[3] + pptx[4]) / 2) + ((pptx[14] + pptx[15]) / 2)) / 2)) + orix, (Math.round((((ppty[3] + ppty[4]) / 2) + ((ppty[14] + ppty[15]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[23] + pptx[22]) / 2) + orix, ((ppty[23] + ppty[22]) / 2) + oriy);
					ctx.lineTo(((pptx[23] + pptx[24]) / 2) + orix, ((ppty[23] + ppty[24]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 5) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[22]) + orix, (ppty[22]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[22]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[6]) / 2) + orix, ((ppty[7] + ppty[6]) / 2) + oriy);
					ctx.lineTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
					ctx.lineTo(((pptx[5] + pptx[6]) / 2) + orix, (Math.round((ppty[0] + ((ppty[11] + ppty[12]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.lineTo((pptx[2]) + orix, (ppty[2]) + oriy);
					ctx.lineTo((pptx[5]) + orix, (ppty[10]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			break;
		case HEX_SRSID:
			if(rotation == 0) {
				ctx.beginPath();
				ctx.moveTo(((pptx[2] + pptx[3]) / 2) + orix, ((ppty[2] + ppty[3]) / 2) + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, ((ppty[15] + ppty[16]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1) {
				ctx.beginPath();
				ctx.moveTo(((pptx[6] + pptx[7]) / 2) + orix, ((ppty[6] + ppty[7]) / 2) + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2) {
				ctx.beginPath();
				ctx.moveTo(((pptx[10] + pptx[11]) / 2) + orix, ((ppty[10] + ppty[11]) / 2) + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(((pptx[23] + pptx[24]) / 2) + orix, ((ppty[23] + ppty[24]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 3) {
				ctx.beginPath();
				ctx.moveTo(((pptx[14] + pptx[15]) / 2) + orix, ((ppty[14] + ppty[15]) / 2) + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, ((ppty[3] + ppty[4]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 4) {
				ctx.beginPath();
				ctx.moveTo(((pptx[18] + pptx[19]) / 2) + orix, ((ppty[18] + ppty[19]) / 2) + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 5) {
				ctx.beginPath();
				ctx.moveTo(((pptx[22] + pptx[23]) / 2) + orix, ((ppty[22] + ppty[23]) / 2) + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(((pptx[11] + pptx[12]) / 2) + orix, ((ppty[12] + ppty[11]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			break;
		case HEX_SRCNR:
			if(rotation == 0) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[5]) + orix, (ppty[5]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[6]) / 2) + orix, ((ppty[7] + ppty[6]) / 2) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.lineTo(((pptx[4] + pptx[5]) / 2) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 1) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (pptx[3] - pptx[2]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[6]) / 2) + orix, ((ppty[7] + ppty[6]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[20] + pptx[6]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.lineTo(((pptx[11] + pptx[12]) / 2) + orix, ((ppty[11] + ppty[12]) / 2) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (ppty[11] - ppty[10]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[11] + pptx[10]) / 2) + orix, ((ppty[11] + ppty[10]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[16] + pptx[2]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[16] + ppty[2]) / 2)) / 2)) + oriy);
					ctx.lineTo(((pptx[16] + pptx[15]) / 2) + orix, ((ppty[16] + ppty[15]) / 2) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 2) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
					ctx.lineTo(((pptx[4] + pptx[5]) / 2) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.lineTo((pptx[17]) + orix, (ppty[17]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 3) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[17]) + orix, (ppty[17]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[18]) / 2) + orix, ((ppty[19] + ppty[18]) / 2) + oriy);
					ctx.lineTo(((pptx[17] + pptx[16]) / 2) + orix, (Math.round((ppty[0] + ((ppty[11] + ppty[12]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 4) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (pptx[3] - pptx[2]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[18]) / 2) + orix, ((ppty[19] + ppty[18]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[8] + pptx[18]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[8] + ppty[18]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.lineTo(((pptx[23] + pptx[24]) / 2) + orix, ((ppty[23] + ppty[24]) / 2) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (ppty[11] - ppty[10]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[4]) / 2) + orix, ((ppty[3] + ppty[4]) / 2) + oriy);
					ctx.lineTo((Math.round((((pptx[3] + pptx[4]) / 2) + ((pptx[14] + pptx[15]) / 2)) / 2)) + orix, (Math.round((((ppty[3] + ppty[4]) / 2) + ((ppty[14] + ppty[15]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.lineTo(((pptx[23] + pptx[22]) / 2) + orix, ((ppty[23] + ppty[22]) / 2) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 5) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
					ctx.lineTo(((pptx[5] + pptx[6]) / 2) + orix, (Math.round((ppty[0] + ((ppty[11] + ppty[12]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.lineTo((pptx[5]) + orix, (ppty[5]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			break;
		case HEX_PILAR:
			ctx.beginPath();
			ctx.arc(orix, oriy, (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[1]) : (ppty[11] - ppty[9])), 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			break;
		case HEX_PSEUD:
			if(rotation == 0) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 3) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 4) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 5) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			break;
		default:
			break;
	}
}

function drawFeature(ctx, pptx, ppty, scl, ctyp, matr) {
	var orix = pptx[0];
	var oriy = ppty[0];
	var rotation  = ctyp % 10;
	var symbology = ctyp - rotation;
	ctx.fillStyle = matr;
	ctx.lineWidth = DEFAULT_LINE_THICKNESS;
	switch(symbology) {
		case FEATURE_DOOR:
			var thick = Math.max(Math.round(scl / 4), 1);
			var diffX = (pptx[17] - pptx[1]) * 1.0;
			var diffY = (ppty[17] - ppty[1]) * 1.0;
			var segStepX = (Math.abs(diffX) > Math.abs(diffY) ? 1.0 * Math.sign(diffX) : diffX / diffY);
			var segStepY = (Math.abs(diffY) > Math.abs(diffX) ? 1.0 * Math.sign(diffY) : diffY / diffX);
			var poly = [
				new Point(pptx[1] - orix, (ppty[1]) - oriy),
				new Point(pptx[5] - orix, (ppty[5]) - oriy),
				new Point((pptx[5] + Math.round(segStepX * thick)) - orix, (ppty[5] + Math.round(segStepY * thick)) - oriy),
				new Point((pptx[1] + Math.round(segStepX * thick)) - orix, (ppty[1] + Math.round(segStepY * thick)) - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_STAIRS:
			var diffX = (pptx[15] - pptx[3]) / 13.0;
			var diffY = (ppty[15] - ppty[3]) / 13.0;
			var poly1 = [
				new Point((pptx[2] - orix) + (diffX * 1.0), (ppty[2] - oriy) + (diffY * 1.0)),
				new Point((pptx[4] - orix) + (diffX * 1.0), (ppty[4] - oriy) + (diffY * 1.0)),
				new Point((pptx[4] - orix) + (diffX * 3.0), (ppty[4] - oriy) + (diffY * 3.0)),
				new Point((pptx[2] - orix) + (diffX * 3.0), (ppty[2] - oriy) + (diffY * 3.0))
			];
			var poly2 = [
				new Point((pptx[2] - orix) + (diffX * 4.0), (ppty[2] - oriy) + (diffY * 4.0)),
				new Point((pptx[4] - orix) + (diffX * 4.0), (ppty[4] - oriy) + (diffY * 4.0)),
				new Point((pptx[4] - orix) + (diffX * 6.0), (ppty[4] - oriy) + (diffY * 6.0)),
				new Point((pptx[2] - orix) + (diffX * 6.0), (ppty[2] - oriy) + (diffY * 6.0))
			];
			var poly3 = [
				new Point((pptx[2] - orix) + (diffX * 7.0), (ppty[2] - oriy) + (diffY * 7.0)),
				new Point((pptx[4] - orix) + (diffX * 7.0), (ppty[4] - oriy) + (diffY * 7.0)),
				new Point((pptx[4] - orix) + (diffX * 9.0), (ppty[4] - oriy) + (diffY * 9.0)),
				new Point((pptx[2] - orix) + (diffX * 9.0), (ppty[2] - oriy) + (diffY * 9.0))
			];
			var poly4 = [
				new Point((pptx[2] - orix) + (diffX * 10.0), (ppty[2] - oriy) + (diffY * 10.0)),
				new Point((pptx[4] - orix) + (diffX * 10.0), (ppty[4] - oriy) + (diffY * 10.0)),
				new Point((pptx[4] - orix) + (diffX * 12.0), (ppty[4] - oriy) + (diffY * 12.0)),
				new Point((pptx[2] - orix) + (diffX * 12.0), (ppty[2] - oriy) + (diffY * 12.0))
			];
			poly1 = rotatePolygon(poly1, (rotation * 60));
			poly2 = rotatePolygon(poly2, (rotation * 60));
			poly3 = rotatePolygon(poly3, (rotation * 60));
			poly4 = rotatePolygon(poly4, (rotation * 60));
			fillPolygon(ctx, poly1, orix, oriy, matr);
			fillPolygon(ctx, poly2, orix, oriy, matr);
			fillPolygon(ctx, poly3, orix, oriy, matr);
			fillPolygon(ctx, poly4, orix, oriy, matr);
			break;
		case FEATURE_STAIR:
			var diffX = (pptx[15] - pptx[3]) / 13.0;
			var diffY = (ppty[15] - ppty[3]) / 13.0;
			var poly1 = [
				new Point((pptx[2] - orix) + (diffX * 1.0), (ppty[2] - oriy) + (diffY * 1.0)),
				new Point((pptx[4] - orix) + (diffX * 1.0), (ppty[4] - oriy) + (diffY * 1.0)),
				new Point((pptx[4] - orix) + (diffX * 3.0), (ppty[4] - oriy) + (diffY * 3.0)),
				new Point((pptx[2] - orix) + (diffX * 3.0), (ppty[2] - oriy) + (diffY * 3.0))
			];
			var poly2 = [
				new Point((pptx[2] - orix) + (diffX * 4.0), (ppty[2] - oriy) + (diffY * 4.0)),
				new Point((pptx[4] - orix) + (diffX * 4.0), (ppty[4] - oriy) + (diffY * 4.0)),
				new Point((pptx[4] - orix) + (diffX * 6.0), (ppty[4] - oriy) + (diffY * 6.0)),
				new Point((pptx[2] - orix) + (diffX * 6.0), (ppty[2] - oriy) + (diffY * 6.0))
			];
			poly1 = rotatePolygon(poly1, (rotation * 60));
			poly2 = rotatePolygon(poly2, (rotation * 60));
			fillPolygon(ctx, poly1, orix, oriy, matr);
			fillPolygon(ctx, poly2, orix, oriy, matr);
			break;
		case FEATURE_SHAFT:
			ctx.beginPath();
			ctx.arc(orix, oriy, Math.round(scl / 2), 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			ctx.lineWidth = outlineThickness;
			ctx.strokeStyle = hexOutline;
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case FEATURE_UNDERS:
			var diffY = (ppty[15] - ppty[3]) / 9.0;
			var symbRadius = Math.max(scl / 10, 1);
			var symbCenter = new Point(0.0, -(diffY * 3.0));
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			symbCenter.x = 0;
			symbCenter.y = -Math.round(diffY);
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			symbCenter.x = 0;
			symbCenter.y = Math.round(diffY);
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			symbCenter.x = 0;
			symbCenter.y = Math.round(diffY * 3.0);
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			break;
		case FEATURE_UNDER:
			var diffY = (ppty[15] - ppty[3]) / 9.0;
			var symbRadius = Math.max(scl / 10, 1);
			var symbCenter = new Point(0.0, -(diffY * 3.0));
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			symbCenter.x = 0;
			symbCenter.y = -Math.round(diffY);
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			break;
		case FEATURE_BANDS:
			if(rotation == 0 || rotation == 3) {
				ctx.beginPath();
				ctx.moveTo((pptx[2] + pptx[3]) / 2, (ppty[2] + ppty[3]) / 2);
				ctx.lineTo((pptx[3] + pptx[4]) / 2, (ppty[3] + ppty[4]) / 2);
				ctx.lineTo((pptx[14] + pptx[15]) / 2, (ppty[14] + ppty[15]) / 2);
				ctx.lineTo((pptx[15] + pptx[16]) / 2, (ppty[15] + ppty[16]) / 2);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1 || rotation == 4) {
				ctx.beginPath();
				ctx.moveTo((pptx[6] + pptx[7]) / 2, (ppty[6] + ppty[7]) / 2);
				ctx.lineTo((pptx[7] + pptx[8]) / 2, (ppty[7] + ppty[8]) / 2);
				ctx.lineTo((pptx[18] + pptx[19]) / 2, (ppty[18] + ppty[19]) / 2);
				ctx.lineTo((pptx[19] + pptx[20]) / 2, (ppty[19] + ppty[20]) / 2);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2 || rotation == 5) {
				ctx.beginPath();
				ctx.moveTo((pptx[10] + pptx[11]) / 2, (ppty[10] + ppty[11]) / 2);
				ctx.lineTo((pptx[11] + pptx[12]) / 2, (ppty[12] + ppty[11]) / 2);
				ctx.lineTo((pptx[22] + pptx[23]) / 2, (ppty[22] + ppty[23]) / 2);
				ctx.lineTo((pptx[23] + pptx[24]) / 2, (ppty[23] + ppty[24]) / 2);
				ctx.closePath();
				ctx.fill();
			}
			break;
		case FEATURE_BAND:
			ctx.beginPath();
			ctx.arc(pptx[0], ppty[0], (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[2]) : (ppty[11] - ppty[10])) / 2, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			if(rotation == 0) {
				ctx.beginPath();
				ctx.moveTo((pptx[3] + pptx[4]) / 2, (ppty[3] + ppty[4]) / 2);
				ctx.lineTo((pptx[3] + pptx[2]) / 2, (ppty[3] + ppty[2]) / 2);
				ctx.lineTo(Math.round((((pptx[3] + pptx[2]) / 2) + ((pptx[16] + pptx[15]) / 2)) / 2), Math.round((((ppty[3] + ppty[2]) / 2) + ((ppty[16] + ppty[15]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[3] + pptx[4]) / 2) + ((pptx[14] + pptx[15]) / 2)) / 2), Math.round((((ppty[3] + ppty[4]) / 2) + ((ppty[14] + ppty[15]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1) {
				ctx.beginPath();
				ctx.moveTo((pptx[7] + pptx[6]) / 2, (ppty[7] + ppty[6]) / 2);
				ctx.lineTo((pptx[7] + pptx[8]) / 2, (ppty[7] + ppty[8]) / 2);
				ctx.lineTo(Math.round((pptx[0] + ((pptx[8] + pptx[18]) / 2)) / 2), Math.round((ppty[0] + ((ppty[8] + ppty[18]) / 2)) / 2));
				ctx.lineTo(Math.round((pptx[0] + ((pptx[20] + pptx[6]) / 2)) / 2), Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2) {
				ctx.beginPath();
				ctx.moveTo((pptx[11] + pptx[10]) / 2, (ppty[11] + ppty[10]) / 2);
				ctx.lineTo((pptx[11] + pptx[12]) / 2, (ppty[11] + ppty[12]) / 2);
				ctx.lineTo(Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2), Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2));
				ctx.lineTo(Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2), Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 3) {
				ctx.beginPath();
				ctx.moveTo((pptx[16] + pptx[15]) / 2, (ppty[16] + ppty[15]) / 2);
				ctx.lineTo((pptx[14] + pptx[15]) / 2, (ppty[14] + ppty[15]) / 2);
				ctx.lineTo(Math.round((pptx[0] + ((pptx[14] + pptx[4]) / 2)) / 2), Math.round((ppty[0] + ((ppty[14] + ppty[4]) / 2)) / 2));
				ctx.lineTo(Math.round((pptx[0] + ((pptx[16] + pptx[2]) / 2)) / 2), Math.round((ppty[0] + ((ppty[16] + ppty[2]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 4) {
				ctx.beginPath();
				ctx.moveTo((pptx[19] + pptx[18]) / 2, (ppty[19] + ppty[18]) / 2);
				ctx.lineTo((pptx[19] + pptx[20]) / 2, (ppty[19] + ppty[20]) / 2);
				ctx.lineTo(Math.round((pptx[0] + ((pptx[20] + pptx[6]) / 2)) / 2), Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2));
				ctx.lineTo(Math.round((pptx[0] + ((pptx[8] + pptx[18]) / 2)) / 2), Math.round((ppty[0] + ((ppty[8] + ppty[18]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 5) {
				ctx.beginPath();
				ctx.moveTo((pptx[23] + pptx[22]) / 2, (ppty[23] + ppty[22]) / 2);
				ctx.lineTo((pptx[23] + pptx[24]) / 2, (ppty[23] + ppty[24]) / 2);
				ctx.lineTo(Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2), Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2));
				ctx.lineTo(Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2), Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			break;
		case FEATURE_AXONS:
/*
			if(rotation == 0 || rotation == 3) {
				ctx.beginPath();
				ctx.moveTo((pptx[24] + pptx[1]) / 2, (ppty[24] + ppty[1]) / 2);
				ctx.lineTo(pptx[1], ppty[1]);
				ctx.lineTo((pptx[1] + pptx[2]) / 2, (ppty[1] + ppty[2]) / 2);
				ctx.lineTo((pptx[12] + pptx[13]) / 2, (ppty[12] + ppty[13]) / 2);
				ctx.lineTo(pptx[13], ppty[13]);
				ctx.lineTo((pptx[13] + pptx[14]) / 2, (ppty[13] + ppty[14]) / 2);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1 || rotation == 4) {
				ctx.beginPath();
				ctx.moveTo((pptx[4] + pptx[5]) / 2, (ppty[4] + ppty[5]) / 2);
				ctx.lineTo(pptx[5], ppty[5]);
				ctx.lineTo((pptx[5] + pptx[6]) / 2, (ppty[5] + ppty[6]) / 2);
				ctx.lineTo((pptx[16] + pptx[17]) / 2, (ppty[16] + ppty[17]) / 2);
				ctx.lineTo(pptx[17], ppty[17]);
				ctx.lineTo((pptx[17] + pptx[18]) / 2, (ppty[17] + ppty[18]) / 2);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2 || rotation == 5) {
				ctx.beginPath();
				ctx.moveTo((pptx[8] + pptx[9]) / 2, (ppty[8] + ppty[9]) / 2);
				ctx.lineTo(pptx[9], ppty[9]);
				ctx.lineTo((pptx[9] + pptx[10]) / 2, (ppty[9] + ppty[10]) / 2);
				ctx.lineTo((pptx[20] + pptx[21]) / 2, (ppty[20] + ppty[21]) / 2);
				ctx.lineTo(pptx[21], ppty[21]);
				ctx.lineTo((pptx[21] + pptx[22]) / 2, (ppty[21] + ppty[22]) / 2);
				ctx.closePath();
				ctx.fill();
			}
*/
			var poly = [
				new Point(((pptx[24] + pptx[1]) / 2) - orix, ((ppty[24] + ppty[1]) / 2) - oriy),
				new Point(pptx[1] - orix, ppty[1] - oriy),
				new Point(((pptx[1] + pptx[2]) / 2) - orix, ((ppty[1] + ppty[2]) / 2) - oriy),
				new Point(((pptx[12] + pptx[13]) / 2) - orix, ((ppty[12] + ppty[13]) / 2) - oriy),
				new Point(pptx[13] - orix, ppty[13] - oriy),
				new Point(((pptx[13] + pptx[14]) / 2) - orix, ((ppty[13] + ppty[14]) / 2) - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_AXON:
			ctx.beginPath();
			ctx.arc(pptx[0], ppty[0], (mapGrain == GRAIN_VERTICAL ? (ppty[24] - ppty[1]) : (pptx[5] - pptx[4])) / 2, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			if(rotation == 0) {
				ctx.beginPath();
				ctx.moveTo((pptx[24] + pptx[1]) / 2, (ppty[24] + ppty[1]) / 2);
				ctx.lineTo(pptx[1], ppty[1]);
				ctx.lineTo((pptx[1] + pptx[2]) / 2, (ppty[1] + ppty[2]) / 2);
				ctx.lineTo(Math.round((((pptx[1] + pptx[2]) / 2) + ((pptx[13] + pptx[12]) / 2)) / 2), Math.round((((ppty[1] + ppty[2]) / 2) + ((ppty[13] + ppty[12]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[1] + pptx[24]) / 2) + ((pptx[13] + pptx[14]) / 2)) / 2), Math.round((((ppty[1] + ppty[24]) / 2) + ((ppty[13] + ppty[14]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1) {
				ctx.beginPath();
				ctx.moveTo((pptx[4] + pptx[5]) / 2, (ppty[4] + ppty[5]) / 2);
				ctx.lineTo(pptx[5], ppty[5]);
				ctx.lineTo((pptx[5] + pptx[6]) / 2, (ppty[5] + ppty[6]) / 2);
				ctx.lineTo(Math.round((((pptx[5] + pptx[6]) / 2) + ((pptx[17] + pptx[16]) / 2)) / 2), Math.round((((ppty[5] + ppty[6]) / 2) + ((ppty[17] + ppty[16]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[5] + pptx[4]) / 2) + ((pptx[17] + pptx[18]) / 2)) / 2), Math.round((((ppty[5] + ppty[4]) / 2) + ((ppty[17] + ppty[18]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2) {
				ctx.beginPath();
				ctx.moveTo((pptx[8] + pptx[9]) / 2, (ppty[8] + ppty[9]) / 2);
				ctx.lineTo(pptx[9], ppty[9]);
				ctx.lineTo((pptx[9] + pptx[10]) / 2, (ppty[9] + ppty[10]) / 2);
				ctx.lineTo(Math.round((((pptx[9] + pptx[10]) / 2) + ((pptx[21] + pptx[20]) / 2)) / 2), Math.round((((ppty[9] + ppty[10]) / 2) + ((ppty[21] + ppty[20]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[9] + pptx[8]) / 2) + ((pptx[21] + pptx[22]) / 2)) / 2), Math.round((((ppty[9] + ppty[8]) / 2) + ((ppty[21] + ppty[22]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 3) {
				ctx.beginPath();
				ctx.moveTo((pptx[12] + pptx[13]) / 2, (ppty[12] + ppty[13]) / 2);
				ctx.lineTo(pptx[13], ppty[13]);
				ctx.lineTo((pptx[13] + pptx[14]) / 2, (ppty[13] + ppty[14]) / 2);
				ctx.lineTo(Math.round((((pptx[1] + pptx[24]) / 2) + ((pptx[13] + pptx[14]) / 2)) / 2), Math.round((((ppty[1] + ppty[24]) / 2) + ((ppty[13] + ppty[14]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[1] + pptx[2]) / 2) + ((pptx[13] + pptx[12]) / 2)) / 2), Math.round((((ppty[1] + ppty[2]) / 2) + ((ppty[13] + ppty[12]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 4) {
				ctx.beginPath();
				ctx.moveTo((pptx[16] + pptx[17]) / 2, (ppty[16] + ppty[17]) / 2);
				ctx.lineTo(pptx[17], ppty[17]);
				ctx.lineTo((pptx[17] + pptx[18]) / 2, (ppty[17] + ppty[18]) / 2);
				ctx.lineTo(Math.round((((pptx[5] + pptx[4]) / 2) + ((pptx[17] + pptx[18]) / 2)) / 2), Math.round((((ppty[5] + ppty[4]) / 2) + ((ppty[17] + ppty[18]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[5] + pptx[6]) / 2) + ((pptx[17] + pptx[16]) / 2)) / 2), Math.round((((ppty[5] + ppty[6]) / 2) + ((ppty[17] + ppty[16]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 5) {
				ctx.beginPath();
				ctx.moveTo((pptx[20] + pptx[21]) / 2, (ppty[20] + ppty[21]) / 2);
				ctx.lineTo(pptx[21], ppty[21]);
				ctx.lineTo((pptx[21] + pptx[22]) / 2, (ppty[21] + ppty[22]) / 2);
				ctx.lineTo(Math.round((((pptx[9] + pptx[8]) / 2) + ((pptx[21] + pptx[22]) / 2)) / 2), Math.round((((ppty[9] + ppty[8]) / 2) + ((ppty[21] + ppty[22]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[9] + pptx[10]) / 2) + ((pptx[21] + pptx[20]) / 2)) / 2), Math.round((((ppty[9] + ppty[10]) / 2) + ((ppty[21] + ppty[20]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			break;
		case FEATURE_DNDOOR:
			var thick = Math.max((scl / 8), 1);
			var dunit = scl / 4;
			var yoff  = -(scl - (dunit / 2));
			var poly = [
				new Point(dunit * -2.0, yoff),
				new Point(dunit * -1.0, yoff),
				new Point(dunit * -1.0, yoff + thick),
				new Point(dunit *  1.0, yoff + thick),
				new Point(dunit *  1.0, yoff),
				new Point(dunit *  2.0, yoff),
				new Point(dunit *  2.0, yoff + thick),
				new Point(dunit *  1.0 + thick, yoff + thick),
				new Point(dunit *  1.0 + thick, yoff + thick * 2.0),
				new Point(dunit * -1.0 - thick, yoff + thick * 2.0),
				new Point(dunit * -1.0 - thick, yoff + thick),
				new Point(dunit * -2.0, yoff + thick)
			];
/*
// NOTE: To work in either grain, it would be ideal to start with the Vertical grain hex coordinates as reference

			var poly = [
				new Point(pptx[1] - orix, ppty[1] - oriy),
//				new Point(((pptx[2] - orix) + (pptx[1] - orix)) / 2, ppty[1] - oriy),
//				new Point(((pptx[2] - orix) + (pptx[1] - orix)) / 2, ppty[24] - oriy),
//				new Point(((pptx[5] - orix) + (pptx[4] - orix)) / 2, ppty[24] - oriy),
//				new Point(((pptx[5] - orix) + (pptx[4] - orix)) / 2, ppty[1] - oriy),

				new Point(pptx[2] - orix, ppty[1] - oriy),
				new Point(pptx[2] - orix, ppty[24] - oriy),
				new Point(pptx[4] - orix, ppty[24] - oriy),
				new Point(pptx[4] - orix, ppty[1] - oriy),


				new Point(pptx[5] - orix, ppty[5] - oriy),
				new Point(pptx[6] - orix, ppty[6] - oriy),
				new Point(pptx[5] - orix, ppty[6] - oriy),
				new Point(pptx[5] - orix, ppty[7] - oriy),
				new Point(pptx[1] - orix, ppty[7] - oriy),
				new Point(pptx[1] - orix, ppty[6] - oriy),
				new Point(pptx[24] - orix, ppty[24] - oriy)
			];
*/
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_SPIRAL:
			var minrad = scl * 0.2;
			var maxrad = scl * 0.6;
			ctx.lineWidth = outlineThickness;
			ctx.strokeStyle = matr;
			ctx.beginPath();
			ctx.arc(pptx[0], ppty[0], minrad, 0, Math.PI*2);
			ctx.closePath();
			ctx.stroke();
			ctx.beginPath();
			ctx.arc(pptx[0], ppty[0], maxrad, 0, Math.PI*2);
			ctx.closePath();
			ctx.stroke();
			for(var sstp = 0; sstp < 360; sstp += 45) {
				ctx.moveTo(pptx[0] + (minrad * Math.cos(sstp * ANGLE_STEP)), ppty[0]  + (minrad * Math.sin(sstp * ANGLE_STEP)));
				ctx.lineTo(pptx[0] + (maxrad * Math.cos(sstp * ANGLE_STEP)), ppty[0]  + (maxrad * Math.sin(sstp * ANGLE_STEP)));
				ctx.stroke();
			}
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case FEATURE_WINDOW:
			var thick = Math.max((scl / 16), 1);
			var dunit = (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[2]) : (ppty[11] - ppty[10]));
			var yoff  = -(scl - (dunit / 2));
			var poly = [
				new Point(dunit *  2.0, yoff),
				new Point(dunit *  2.0, yoff + thick * 4.0),
				new Point(dunit *  1.0 + thick, yoff + thick * 4.0),
				new Point(dunit *  1.0 + thick, yoff + thick),
				new Point(dunit *  0.0 + thick, yoff + thick),
				new Point(dunit *  0.0 + thick, yoff + thick * 3.0),
				new Point(dunit *  0.0 - thick, yoff + thick * 3.0),
				new Point(dunit *  0.0 - thick, yoff + thick),
				new Point(dunit * -1.0 - thick, yoff + thick),
				new Point(dunit * -1.0 - thick, yoff + thick * 3.0),
				new Point(dunit *  1.0 + thick, yoff + thick * 3.0),
				new Point(dunit *  1.0 + thick, yoff + thick * 4.0),
				new Point(dunit * -2.0, yoff + thick * 4.0),
				new Point(dunit * -2.0, yoff)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_MIDOOR:
			var thick = Math.max((scl / 8), 1);
			var dunit = (mapGrain == GRAIN_VERTICAL ? (ppty[15] - ppty[3]) : (pptx[12] - pptx[23])) / 4;
			var poly = [
				new Point(dunit *  2.0, 0),
				new Point(dunit *  2.0, thick),
				new Point(dunit *  1.0 + thick, thick),
				new Point(dunit *  1.0 + thick, thick * 2.0),
				new Point(dunit * -1.0 - thick, thick * 2.0),
				new Point(dunit * -1.0 - thick, thick),
				new Point(dunit * -2.0, thick),
				new Point(dunit * -2.0, 0),
				new Point(dunit * -1.0, 0),
				new Point(dunit * -1.0, thick),
				new Point(dunit *  1.0, thick),
				new Point(dunit *  1.0, 0),
				new Point(dunit *  1.0, -thick),
				new Point(dunit * -1.0, -thick),
				new Point(dunit * -1.0, 0),
				new Point(dunit * -2.0, 0),
				new Point(dunit * -2.0, -thick),
				new Point(dunit * -1.0 - thick, -thick),
				new Point(dunit * -1.0 - thick, -thick * 2.0),
				new Point(dunit *  1.0 + thick, -thick * 2.0),
				new Point(dunit *  1.0 + thick, -thick),
				new Point(dunit *  2.0, -thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_VERTICAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_MIWIND:
			var thick = Math.max((scl / 16), 1);
			var dunit = (mapGrain == GRAIN_VERTICAL ? (ppty[15] - ppty[3]) : (pptx[12] - pptx[23])) / 4;
			var poly = [
				new Point(dunit *  2.0, -thick * 2.0),
				new Point(dunit *  2.0,  thick * 2.0),
				new Point(dunit *  1.0 + thick,  thick * 2.0),
				new Point(dunit *  1.0 + thick, -thick),
				new Point(dunit *  0.0 + thick, -thick),
				new Point(dunit *  0.0 + thick,  thick),
				new Point(dunit *  0.0 - thick,  thick),
				new Point(dunit *  0.0 - thick, -thick),
				new Point(dunit * -1.0 - thick, -thick),
				new Point(dunit * -1.0 - thick, thick),
				new Point(dunit *  1.0 + thick, thick),
				new Point(dunit *  1.0 + thick, thick * 2.0),
				new Point(dunit * -2.0, thick * 2.0),
				new Point(dunit * -2.0, -thick * 2.0)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_VERTICAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_TDOOR:
			hexpts = getNorthernHexPoints(orix, oriy, scl);
			pptx = hexpts[0];
			ppty = hexpts[1];
			var thick = Math.max((scl / 16), 1);
			var poly = [
				new Point(pptx[5] - orix, ppty[6] - oriy),
				new Point(pptx[6] - orix, ppty[6] - oriy),
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[5] - orix, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[24] - orix, ppty[24] - oriy),
				new Point(pptx[1] - orix, ppty[24] - oriy),
				new Point(pptx[1] - orix, ppty[22] - oriy),
				new Point(pptx[22] - orix, ppty[22] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[1] - orix, ppty[24] - oriy),
				new Point(pptx[5] - orix, ppty[6] - oriy),
				new Point(pptx[5] - orix, (ppty[6] - oriy) + thick),
				new Point(pptx[1] - orix, (ppty[24] - oriy) + thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[1] - orix, ppty[22] - oriy),
				new Point(pptx[5] - orix, ppty[8] - oriy),
				new Point(pptx[5] - orix, (ppty[8] - oriy) - thick),
				new Point(pptx[1] - orix, (ppty[22] - oriy) - thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_TWIND:
			hexpts = getNorthernHexPoints(orix, oriy, scl);
			pptx = hexpts[0];
			ppty = hexpts[1];
			var thick = Math.max((scl / 16), 1);
			var poly = [
				new Point(pptx[5] - orix, ppty[6] - oriy),
				new Point(pptx[6] - orix, ppty[6] - oriy),
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[5] - orix, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[24] - orix, ppty[24] - oriy),
				new Point(pptx[1] - orix, ppty[24] - oriy),
				new Point(pptx[1] - orix, ppty[22] - oriy),
				new Point(pptx[22] - orix, ppty[22] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[1] - orix, ppty[24] - oriy),
				new Point(pptx[5] - orix, ppty[6] - oriy),
				new Point(pptx[5] - orix, (ppty[6] - oriy) + thick),
				new Point(pptx[1] - orix, (ppty[24] - oriy) + thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[1] - orix, ppty[22] - oriy),
				new Point(pptx[5] - orix, ppty[8] - oriy),
				new Point(pptx[5] - orix, (ppty[8] - oriy) - thick),
				new Point(pptx[1] - orix, (ppty[22] - oriy) - thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point((pptx[3] - orix) - thick, ppty[6] - oriy),
				new Point((pptx[3] - orix) + thick, ppty[6] - oriy),
				new Point((pptx[3] - orix) + thick, ppty[8] - oriy),
				new Point((pptx[3] - orix) - thick, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_EDOOR:
			hexpts = getNorthernHexPoints(orix, oriy, scl);
			pptx = hexpts[0];
			ppty = hexpts[1];
			var thick = Math.max((scl / 16), 1);
			var poly = [
				new Point(pptx[6] - orix, ppty[6] - oriy),
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[6] - orix, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[10] - orix, ppty[10] - oriy),
				new Point(pptx[12] - orix, ppty[12] - oriy),
				new Point(pptx[12] - orix, ppty[10] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[8] - orix, ppty[10] - oriy),
				new Point((pptx[8] - orix) - thick, ppty[10] - oriy),
				new Point((pptx[8] - orix) - thick, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[6] - orix, ppty[8] - oriy),
				new Point(pptx[6] - orix, ppty[10] - oriy),
				new Point((pptx[6] - orix) + thick, ppty[10] - oriy),
				new Point((pptx[6] - orix) + thick, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_EWIND:
			hexpts = getNorthernHexPoints(orix, oriy, scl);
			pptx = hexpts[0];
			ppty = hexpts[1];
			var thick = Math.max((scl / 16), 1);
			var poly = [
				new Point(pptx[6] - orix, ppty[6] - oriy),
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[6] - orix, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[10] - orix, ppty[10] - oriy),
				new Point(pptx[12] - orix, ppty[12] - oriy),
				new Point(pptx[12] - orix, ppty[10] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[8] - orix, ppty[10] - oriy),
				new Point((pptx[8] - orix) - thick, ppty[10] - oriy),
				new Point((pptx[8] - orix) - thick, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[6] - orix, ppty[8] - oriy),
				new Point(pptx[6] - orix, ppty[10] - oriy),
				new Point((pptx[6] - orix) + thick, ppty[10] - oriy),
				new Point((pptx[6] - orix) + thick, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[6] - orix, (ppty[9] - oriy) - thick),
				new Point(pptx[6] - orix, (ppty[9] - oriy) + thick),
				new Point(pptx[8] - orix, (ppty[9] - oriy) + thick),
				new Point(pptx[8] - orix, (ppty[9] - oriy) - thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_ADOOR:
			var thick = Math.max((scl / 8), 1);
			var dunit = (mapGrain == GRAIN_VERTICAL ? (ppty[15] - ppty[3]) : (pptx[12] - pptx[23])) / 4;
			var poly = [
				new Point(dunit *  2.0, 0),
				new Point(dunit *  2.0, thick),
				new Point(dunit *  1.0 + thick, thick),
				new Point(dunit *  1.0 + thick, thick * 2.0),
				new Point(dunit * -1.0 - thick, thick * 2.0),
				new Point(dunit * -1.0 - thick, thick),
				new Point(dunit * -2.0, thick),
				new Point(dunit * -2.0, 0),
				new Point(dunit * -1.0, 0),
				new Point(dunit * -1.0, thick),
				new Point(dunit *  1.0, thick),
				new Point(dunit *  1.0, 0),
				new Point(dunit *  1.0, -thick),
				new Point(dunit * -1.0, -thick),
				new Point(dunit * -1.0, 0),
				new Point(dunit * -2.0, 0),
				new Point(dunit * -2.0, -thick),
				new Point(dunit * -1.0 - thick, -thick),
				new Point(dunit * -1.0 - thick, -thick * 2.0),
				new Point(dunit *  1.0 + thick, -thick * 2.0),
				new Point(dunit *  1.0 + thick, -thick),
				new Point(dunit *  2.0, -thick)
			];
			poly = rotatePolygon(poly, (mapGrain == GRAIN_VERTICAL ? 60 : 30));
			poly[21] = new Point(((pptx[12] + pptx[13]) / 2) - orix, ((ppty[12] + ppty[13]) / 2) - oriy);
			poly[0] = new Point(pptx[13] - orix, ppty[13] - oriy);
			poly[1] = new Point(((pptx[13] + pptx[14]) / 2) - orix, ((ppty[13] + ppty[14]) / 2) - oriy);
			poly[6] = new Point(((pptx[24] + pptx[1]) / 2) - orix, ((ppty[24] + ppty[1]) / 2) - oriy);
			poly[7] = new Point(pptx[1] - orix, ppty[1] - oriy);
			poly[15] = new Point(pptx[1] - orix, ppty[1] - oriy);
			poly[16] = new Point(((pptx[1] + pptx[2]) / 2) - orix, ((ppty[1] + ppty[2]) / 2) - oriy);
			poly = rotatePolygon(poly, (rotation * 60));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_AWIND:
			var thick = Math.max((scl / 8), 1);
			var halfthick = Math.max((thick / 2), 1);
			var dunit = (mapGrain == GRAIN_VERTICAL ? (ppty[15] - ppty[3]) : (pptx[12] - pptx[23])) / 4;
			var poly = [
				new Point(dunit *  2.0, 0),
				new Point(dunit *  2.0, thick),
				new Point(dunit *  1.0 + thick, thick),
				new Point(dunit *  1.0 + thick, -halfthick),
				new Point(halfthick, -halfthick),
				new Point(halfthick, halfthick),
				new Point(-halfthick, halfthick),
				new Point(-halfthick, -halfthick),
				new Point(dunit * -1.0 - thick, -halfthick),
				new Point(dunit * -1.0 - thick, halfthick),
				new Point(dunit *  1.0 + thick, halfthick),
				new Point(dunit *  1.0 + thick, thick),
				new Point(dunit * -1.0 - thick, thick),
				new Point(dunit * -2.0, thick),
				new Point(dunit * -2.0, 0),
				new Point(dunit * -2.0, -thick),
				new Point(dunit * -1.0 - thick, -thick),
				new Point(dunit *  1.0 + thick, -thick),
				new Point(dunit *  2.0, -thick)
			];
			poly = rotatePolygon(poly, (mapGrain == GRAIN_VERTICAL ? 60 : 30));
			poly[18] = new Point(((pptx[12] + pptx[13]) / 2) - orix, ((ppty[12] + ppty[13]) / 2) - oriy);
			poly[0] = new Point(pptx[13] - orix, ppty[13] - oriy);
			poly[1] = new Point(((pptx[13] + pptx[14]) / 2) - orix, ((ppty[13] + ppty[14]) / 2) - oriy);
			poly[13] = new Point(((pptx[24] + pptx[1]) / 2) - orix, ((ppty[24] + ppty[1]) / 2) - oriy);
			poly[14] = new Point(pptx[1] - orix, ppty[1] - oriy);
			poly[15] = new Point(((pptx[1] + pptx[2]) / 2) - orix, ((ppty[1] + ppty[2]) / 2) - oriy);
			poly = rotatePolygon(poly, (rotation * 60));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_PILAR:
			ctx.beginPath();
			ctx.arc(orix, oriy, (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[1]) : (ppty[11] - ppty[9])), 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			break;
		case TERRAIN_BOUNDER:
			var poly = [
				new Point( 0,             -0.75 * scl),
				new Point( 0.1875 * scl,  -0.1875 * scl),
				new Point( 0.75 * scl,    -0.1875 * scl),
				new Point( 0.28125 * scl,  0.140625 * scl),
				new Point( 0.46875 * scl,  0.703125 * scl),
				new Point( 0,              0.328125 * scl),
				new Point(-0.46875 * scl,  0.703125 * scl),
				new Point(-0.28125 * scl,  0.140625 * scl),
				new Point(-0.75 * scl,    -0.1875 * scl),
				new Point(-0.1875 * scl,  -0.1875 * scl)
			];
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case TERRAIN_WATER:
			ctx.beginPath();
			if(mapGrain == GRAIN_HORIZONTAL) {
				var pt = getLineIntersection(new Point(pptx[9], ppty[7]), new Point(pptx[8], ppty[9]), new Point(pptx[9], ppty[9]), new Point(pptx[5], ppty[5]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[8], ppty[9]);
				ctx.lineTo(pptx[10], ppty[10]);
				ctx.lineTo(pptx[8], ppty[11]);
				ctx.lineTo(pptx[12], ppty[12]);
				ctx.lineTo(pptx[8], ppty[13]);
				pt = getLineIntersection(new Point(pptx[8], ppty[13]), new Point(pptx[9], ppty[15]), new Point(pptx[13], ppty[13]), new Point(pptx[17], ppty[17]));
				ctx.lineTo(pt.x, pt.y);
				pt = getLineIntersection(new Point(pptx[8], ppty[7]), new Point(pptx[7], ppty[9]), new Point(pptx[9], ppty[9]), new Point(pptx[5], ppty[5]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[7], ppty[9]);
				ctx.lineTo(pptx[8], ppty[10]);
				ctx.lineTo(pptx[7], ppty[11]);
				ctx.lineTo(pptx[8], ppty[12]);
				ctx.lineTo(pptx[7], ppty[13]);
				pt = getLineIntersection(new Point(pptx[7], ppty[13]), new Point(pptx[14], ppty[15]), new Point(pptx[13], ppty[13]), new Point(pptx[17], ppty[17]));
				ctx.lineTo(pt.x, pt.y);
				ctx.moveTo(pptx[7], ppty[7]);
				ctx.lineTo(pptx[6], ppty[9]);
				ctx.lineTo(pptx[7], ppty[10]);
				ctx.lineTo(pptx[6], ppty[11]);
				ctx.lineTo(pptx[7], ppty[12]);
				ctx.lineTo(pptx[6], ppty[13]);
				ctx.lineTo(pptx[15], ppty[15]);
				ctx.moveTo(pptx[5], ppty[5]);
				ctx.lineTo(pptx[6], ppty[7]);
				ctx.lineTo(pptx[5], ppty[9]);
				ctx.lineTo(pptx[6], ppty[10]);
				ctx.lineTo(pptx[5], ppty[11]);
				ctx.lineTo(pptx[6], ppty[12]);
				ctx.lineTo(pptx[5], ppty[13]);
				ctx.lineTo(pptx[6], ppty[15]);
				ctx.lineTo(pptx[17], ppty[17]);
				pt = getLineIntersection(new Point(pptx[4], ppty[5]), new Point(pptx[5], ppty[3]), new Point(pptx[1], ppty[1]), new Point(pptx[5], ppty[5]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[5], ppty[3]);
				ctx.lineTo(pptx[4], ppty[9]);
				ctx.lineTo(pptx[5], ppty[10]);
				ctx.lineTo(pptx[4], ppty[11]);
				ctx.lineTo(pptx[5], ppty[12]);
				ctx.lineTo(pptx[4], ppty[13]);
				ctx.lineTo(pptx[5], ppty[19]);
				pt = getLineIntersection(new Point(pptx[5], ppty[19]), new Point(pptx[4], ppty[17]), new Point(pptx[21], ppty[21]), new Point(pptx[17], ppty[17]));
				ctx.lineTo(pt.x, pt.y);
				pt = getLineIntersection(new Point(pptx[3], ppty[5]), new Point(pptx[4], ppty[3]), new Point(pptx[1], ppty[1]), new Point(pptx[5], ppty[5]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[4], ppty[3]);
				ctx.lineTo(pptx[3], ppty[9]);
				ctx.lineTo(pptx[4], ppty[10]);
				ctx.lineTo(pptx[3], ppty[11]);
				ctx.lineTo(pptx[4], ppty[12]);
				ctx.lineTo(pptx[3], ppty[13]);
				ctx.lineTo(pptx[4], ppty[19]);
				pt = getLineIntersection(new Point(pptx[4], ppty[19]), new Point(pptx[19], ppty[17]), new Point(pptx[21], ppty[21]), new Point(pptx[17], ppty[17]));
				ctx.lineTo(pt.x, pt.y);
				ctx.moveTo(pptx[3], ppty[3]);
				ctx.lineTo(pptx[2], ppty[9]);
				ctx.lineTo(pptx[3], ppty[10]);
				ctx.lineTo(pptx[2], ppty[11]);
				ctx.lineTo(pptx[3], ppty[12]);
				ctx.lineTo(pptx[2], ppty[13]);
				ctx.lineTo(pptx[19], ppty[19]);
				ctx.moveTo(pptx[1], ppty[1]);
				ctx.lineTo(pptx[2], ppty[10]);
				ctx.lineTo(pptx[1], ppty[11]);
				ctx.lineTo(pptx[2], ppty[12]);
				ctx.lineTo(pptx[21], ppty[21]);
			}
			else {
				var pt = getLineIntersection(new Point(pptx[23], ppty[1]), new Point(pptx[1], ppty[24]), new Point(pptx[1], ppty[1]), new Point(pptx[21], ppty[21]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[1], ppty[24]);
				ctx.lineTo(pptx[2], ppty[2]);
				ctx.lineTo(pptx[3], ppty[24]);
				ctx.lineTo(pptx[4], ppty[4]);
				ctx.lineTo(pptx[5], ppty[24]);
				pt = getLineIntersection(new Point(pptx[5], ppty[24]), new Point(pptx[7], ppty[1]), new Point(pptx[5], ppty[5]), new Point(pptx[9], ppty[9]));
				ctx.lineTo(pt.x, pt.y);
				pt = getLineIntersection(new Point(pptx[23], ppty[24]), new Point(pptx[1], ppty[23]), new Point(pptx[1], ppty[1]), new Point(pptx[21], ppty[21]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[1], ppty[23]);
				ctx.lineTo(pptx[2], ppty[24]);
				ctx.lineTo(pptx[3], ppty[23]);
				ctx.lineTo(pptx[4], ppty[24]);
				ctx.lineTo(pptx[5], ppty[23]);
				pt = getLineIntersection(new Point(pptx[5], ppty[23]), new Point(pptx[7], ppty[6]), new Point(pptx[5], ppty[5]), new Point(pptx[9], ppty[9]));
				ctx.lineTo(pt.x, pt.y);
				ctx.moveTo(pptx[23], ppty[23]);
				ctx.lineTo(pptx[1],  ppty[22]);
				ctx.lineTo(pptx[2],  ppty[23]);
				ctx.lineTo(pptx[3],  ppty[22]);
				ctx.lineTo(pptx[4],  ppty[23]);
				ctx.lineTo(pptx[5],  ppty[22]);
				ctx.lineTo(pptx[7],  ppty[7]);
				ctx.moveTo(pptx[21], ppty[21]);
				ctx.lineTo(pptx[23], ppty[22]);
				ctx.lineTo(pptx[1],  ppty[21]);
				ctx.lineTo(pptx[2],  ppty[22]);
				ctx.lineTo(pptx[3],  ppty[21]);
				ctx.lineTo(pptx[4],  ppty[22]);
				ctx.lineTo(pptx[5],  ppty[21]);
				ctx.lineTo(pptx[7],  ppty[22]);
				ctx.lineTo(pptx[9],  ppty[9]);
				pt = getLineIntersection(new Point(pptx[21], ppty[20]), new Point(pptx[19], ppty[21]), new Point(pptx[17], ppty[17]), new Point(pptx[21], ppty[21]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[19], ppty[21]);
				ctx.lineTo(pptx[1],  ppty[20]);
				ctx.lineTo(pptx[2],  ppty[21]);
				ctx.lineTo(pptx[3],  ppty[20]);
				ctx.lineTo(pptx[4],  ppty[21]);
				ctx.lineTo(pptx[5],  ppty[20]);
				ctx.lineTo(pptx[11], ppty[21]);
				pt = getLineIntersection(new Point(pptx[11], ppty[21]), new Point(pptx[9], ppty[20]), new Point(pptx[13], ppty[13]), new Point(pptx[9], ppty[9]));
				ctx.lineTo(pt.x, pt.y);
				pt = getLineIntersection(new Point(pptx[21], ppty[19]), new Point(pptx[19], ppty[20]), new Point(pptx[17], ppty[17]), new Point(pptx[21], ppty[21]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[19], ppty[20]);
				ctx.lineTo(pptx[1],  ppty[19]);
				ctx.lineTo(pptx[2],  ppty[20]);
				ctx.lineTo(pptx[3],  ppty[19]);
				ctx.lineTo(pptx[4],  ppty[20]);
				ctx.lineTo(pptx[5],  ppty[19]);
				ctx.lineTo(pptx[11], ppty[20]);
				pt = getLineIntersection(new Point(pptx[11], ppty[20]), new Point(pptx[9], ppty[11]), new Point(pptx[13], ppty[13]), new Point(pptx[9], ppty[9]));
				ctx.lineTo(pt.x, pt.y);
				ctx.moveTo(pptx[19], ppty[19]);
				ctx.lineTo(pptx[1],  ppty[18]);
				ctx.lineTo(pptx[2],  ppty[19]);
				ctx.lineTo(pptx[3],  ppty[18]);
				ctx.lineTo(pptx[4],  ppty[19]);
				ctx.lineTo(pptx[5],  ppty[18]);
				ctx.lineTo(pptx[11], ppty[11]);
				ctx.moveTo(pptx[17], ppty[17]);
				ctx.lineTo(pptx[2],  ppty[18]);
				ctx.lineTo(pptx[3],  ppty[17]);
				ctx.lineTo(pptx[4],  ppty[18]);
				ctx.lineTo(pptx[13], ppty[13]);
			}
			ctx.strokeStyle = matr;
			ctx.lineWidth = outlineThickness;
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case TERRAIN_MOUNTOUT:
		case TERRAIN_MOUNTFIL:
			var xunit = scl * 0.125;
			var yunit = scl * 0.2;
			var radunit = xunit * 0.9;
			ctx.lineWidth = outlineThickness;
			ctx.beginPath();
			ctx.ellipse(pptx[0], ppty[0], radunit * 4, radunit * 2, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (xunit * 3), ppty[0] - (yunit * 2), radunit * 2, radunit, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (xunit * 5), ppty[0], radunit, radunit / 2, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (xunit * 2), ppty[0] + (yunit * 2), radunit * 2, radunit, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (xunit * 2), ppty[0] - (yunit * 3), radunit, radunit / 2, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (xunit * 5), ppty[0] - yunit, radunit, radunit / 2, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (xunit * 2), ppty[0] + (yunit * 3), radunit, radunit / 2, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case TERRAIN_TREES:
			var unitx = (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[2]) : (ppty[11] - ppty[10]));
			var unity = (mapGrain == GRAIN_VERTICAL ? (ppty[7] - ppty[6]) : (ppty[11] - ppty[10]) * CALCTHIN);
			var arcunitx = unitx / 2.5;
			var arcunity = unity / 2.5;
			ctx.fillStyle = matr;
			ctx.beginPath();
			ctx.ellipse(pptx[0], ppty[0], arcunitx * 4.0, arcunity * 4.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - unitx, ppty[0], arcunitx * 3.0, arcunity * 3.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0], ppty[0] - (unity * 2.0), arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (unitx * 2.0), ppty[0] - unity, arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (unitx * 2.0), ppty[0] - (unity * 0.5), arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (unitx * 2.0), ppty[0] + unity, arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (unitx * 2.0), ppty[0] + unity, arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + unitx, ppty[0] + (unity * 2.0), arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0], ppty[0] + (unity * 2.5), arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - unitx, ppty[0] - (unity * 2.0), arcunitx * 1.5, arcunity * 1.5, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + unitx, ppty[0] - (unity * 2.0), arcunitx, arcunity, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (unitx * 2.0), ppty[0] - (unity * 1.5), arcunitx * 1.5, arcunity * 1.5, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (unitx * 2.5), ppty[0], arcunitx, arcunity, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - unitx, ppty[0] + (unity * 2.0), arcunitx, arcunity, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (unitx * 2.0), ppty[0] + (unity * 2.0), arcunitx, arcunity, 0, 0, Math.PI*2);
			ctx.fill();
			break;
		case TERRAIN_SWAMP:
			var mxunit = scl * 0.125;
			var myunit = scl * 0.1;
			ctx.lineWidth = outlineThickness;
			ctx.strokeStyle = matr;
/*
			// this version recreates the original pattern
			ctx.beginPath();
			ctx.moveTo(pptx[0], ppty[0] - (myunit * 4));
			ctx.lineTo(pptx[0], ppty[0] - (myunit * 6));
			ctx.moveTo(pptx[0] - mxunit, ppty[0] - (myunit * 4));
			ctx.lineTo(pptx[0] - (mxunit * 2), ppty[0] - (myunit * 6));
			ctx.moveTo(pptx[0] + mxunit, ppty[0] - (myunit * 4));
			ctx.lineTo(pptx[0] + (mxunit * 2), ppty[0] - (myunit * 6));
			ctx.stroke();
			ctx.beginPath();
			ctx.moveTo(pptx[0] - (mxunit * 2), ppty[0]);
			ctx.lineTo(pptx[0] - (mxunit * 3), ppty[0] + (myunit * 2));
			ctx.moveTo(pptx[0] - (mxunit * 4), ppty[0]);
			ctx.lineTo(pptx[0] - (mxunit * 4), ppty[0] + (myunit * 2));
			ctx.moveTo(pptx[0] - (mxunit * 6), ppty[0]);
			ctx.lineTo(pptx[0] - (mxunit * 5), ppty[0] + (myunit * 2));
			ctx.stroke();
			ctx.beginPath();
			ctx.moveTo(pptx[0] + mxunit, ppty[0] + (myunit * 2));
			ctx.lineTo(pptx[0] + (mxunit * 2), ppty[0] + (myunit * 4));
			ctx.moveTo(pptx[0] + (mxunit * 3), ppty[0] + (myunit * 2));
			ctx.lineTo(pptx[0] + (mxunit * 3), ppty[0] + (myunit * 4));
			ctx.moveTo(pptx[0] + (mxunit * 5), ppty[0] + (myunit * 2));
			ctx.lineTo(pptx[0] + (mxunit * 4), ppty[0] + (myunit * 4));
			ctx.stroke();
*/
			// this version is more compatible with both grains
			ctx.beginPath();
			ctx.moveTo(pptx[0] + mxunit, ppty[0] - (myunit * 4));
			ctx.lineTo(pptx[0] + mxunit, ppty[0] - (myunit * 6));
			ctx.moveTo(pptx[0], ppty[0] - (myunit * 4));
			ctx.lineTo(pptx[0] - mxunit, ppty[0] - (myunit * 6));
			ctx.moveTo(pptx[0] + (mxunit * 2), ppty[0] - (myunit * 4));
			ctx.lineTo(pptx[0] + (mxunit * 3), ppty[0] - (myunit * 6));
			ctx.stroke();
			ctx.beginPath();
			ctx.moveTo(pptx[0] - (mxunit * 1), ppty[0]);
			ctx.lineTo(pptx[0] - (mxunit * 2), ppty[0] + (myunit * 2));
			ctx.moveTo(pptx[0] - (mxunit * 3), ppty[0]);
			ctx.lineTo(pptx[0] - (mxunit * 3), ppty[0] + (myunit * 2));
			ctx.moveTo(pptx[0] - (mxunit * 5), ppty[0]);
			ctx.lineTo(pptx[0] - (mxunit * 4), ppty[0] + (myunit * 2));
			ctx.stroke();
			ctx.beginPath();
			ctx.moveTo(pptx[0] + mxunit, ppty[0] + (myunit * 2));
			ctx.lineTo(pptx[0] + (mxunit * 2), ppty[0] + (myunit * 4));
			ctx.moveTo(pptx[0] + (mxunit * 3), ppty[0] + (myunit * 2));
			ctx.lineTo(pptx[0] + (mxunit * 3), ppty[0] + (myunit * 4));
			ctx.moveTo(pptx[0] + (mxunit * 5), ppty[0] + (myunit * 2));
			ctx.lineTo(pptx[0] + (mxunit * 4), ppty[0] + (myunit * 4));
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case TERRAIN_PLAINS:
			ctx.lineWidth = outlineThickness;
			ctx.beginPath();
			ctx.moveTo(pptx[7], ppty[7]);
			ctx.lineTo(pptx[7] - ((pptx[7] - pptx[19]) * 0.25), ppty[7] + ((ppty[19] - ppty[7])  * 0.25));
			ctx.moveTo(pptx[0], ppty[0]);
			ctx.lineTo(pptx[7] - ((pptx[7] - pptx[19]) * 0.75), ppty[7] + ((ppty[19] - ppty[7])  * 0.75));
			ctx.moveTo(pptx[5] - ((pptx[5] - pptx[21]) * 0.25), ppty[5] + ((ppty[21] - ppty[5])  * 0.25));
			ctx.lineTo(pptx[5] - ((pptx[5] - pptx[21]) * 0.5), ppty[5] + ((ppty[21] - ppty[5])  * 0.5));
			ctx.moveTo(pptx[5] - ((pptx[5] - pptx[21]) * 0.75), ppty[5] + ((ppty[21] - ppty[5])  * 0.75));
			ctx.lineTo(pptx[21], ppty[21]);
			ctx.moveTo(pptx[9] - ((pptx[9] - pptx[17]) * 0.25), ppty[9] + ((ppty[17] - ppty[9])  * 0.25));
			ctx.lineTo(pptx[9] - ((pptx[9] - pptx[17]) * 0.5), ppty[9] + ((ppty[17] - ppty[9])  * 0.5));
			ctx.moveTo(pptx[9] - ((pptx[9] - pptx[17]) * 0.75), ppty[9] + ((ppty[17] - ppty[9])  * 0.75));
			ctx.lineTo(pptx[17], ppty[17]);
			ctx.strokeStyle = matr;
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case TERRAIN_DESERT:
			var unitx = (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[2]) : (ppty[11] - ppty[10]));
			var unity = (mapGrain == GRAIN_VERTICAL ? (ppty[7] - ppty[6]) : (ppty[11] - ppty[10]) * CALCTHIN);
			var sandsize = Math.max(scl * 0.125, 1);
			var grainsize = Math.max(sandsize * 0.66, 1);
			ctx.fillStyle = matr;
			ctx.beginPath();
			ctx.rect(pptx[0] - (unitx * 2.0), ppty[0] - (unity * 3.0), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + unitx, ppty[0] - (unity * 2.5), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - unitx, ppty[0] - (unity * 1.3), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - (unitx * 3.15), ppty[0] - unity, sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 2.0), ppty[0], sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0], ppty[0] + (unity * 0.7), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - (unitx * 1.25), ppty[0] + (unity * 1.6), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 0.5), ppty[0] + (unity * 2.2), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 0.5), ppty[0] - (unity * 2.8), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - unitx, ppty[0] - (unity * 2.75), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 2.25), ppty[0] - (unity * 2.25), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 2.75), ppty[0] - (unity * 1.25), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 1.8), ppty[0] - (unity * 1.5), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 0.5), ppty[0] - (unitx * 0.5), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - unitx, ppty[0], grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - (unitx * 2.85), ppty[0] + unity, grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - (unitx * 2.0), ppty[0] + (unity * 2.1), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 2.0), ppty[0] + (unity * 2.25), grainsize, grainsize);
			ctx.fill();
			break;
		case TERRAIN_CAPITOL:
			var csize = scl * 0.75;
			var poly = [
				new Point( 0,               -0.75     * csize),
				new Point( 0.1875  * csize, -0.1875   * csize),
				new Point( 0.75    * csize, -0.1875   * csize),
				new Point( 0.28125 * csize,  0.140625 * csize),
				new Point( 0.46875 * csize,  0.703125 * csize),
				new Point( 0,                0.328125 * csize),
				new Point(-0.46875 * csize,  0.703125 * csize),
				new Point(-0.28125 * csize,  0.140625 * csize),
				new Point(-0.75    * csize, -0.1875   * csize),
				new Point(-0.1875  * csize, -0.1875   * csize)
			];
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case TERRAIN_CITY:
			var cunit = scl * 0.45;
			ctx.beginPath();
			ctx.rect(pptx[0] - cunit, ppty[0] - cunit, cunit * 2, cunit * 2);
			ctx.fillStyle = matr;
			ctx.fill();
			break;
		case TERRAIN_VILLAGE:
			var vunit = scl * 0.45;
			ctx.beginPath();
			ctx.arc(pptx[0], ppty[0], vunit, 0, Math.PI*2);
			ctx.fillStyle = matr;
			ctx.fill();
			break;
		case TERRAIN_FORT:
			var cunit = scl * 0.45;
			var qunit = cunit * 0.5;
			var poly = [
				new Point(-cunit, -cunit),
				new Point(-cunit + qunit, -cunit),
				new Point(-cunit + qunit, -cunit + qunit),
				new Point( cunit - qunit, -cunit + qunit),
				new Point( cunit - qunit, -cunit),
				new Point( cunit, -cunit),
				new Point( cunit,  cunit),
				new Point(-cunit,  cunit)
			];
			drawPolygon(ctx, poly, orix, oriy, matr);
			break;
		case TERRAIN_MINE:
			var mxunit = scl * 0.125;
			var myunit = scl * 0.2;
			var poly = [
				new Point(-mxunit * 0.75, -myunit * 2.5),
				new Point( mxunit * 0.75, -myunit * 2.5),
				new Point( mxunit * 0.75, myunit * 0.5),
				new Point( mxunit * 2.0, 0),
				new Point( mxunit * 2.0, myunit),
				new Point( mxunit,       myunit * 2.0),
				new Point( 0,            myunit * 2.5),
				new Point(-mxunit,       myunit * 2.0),
				new Point(-mxunit * 2.0, myunit),
				new Point(-mxunit * 2.0, 0),
				new Point(-mxunit * 0.75, myunit * 0.5)
			];
			drawPolygon(ctx, poly, orix, oriy, matr);
			break;
		case TERRAIN_RUIN:
			var vunit = scl * 0.45;
			var poly = [
				new Point(0, -vunit),
				new Point(-vunit, vunit * 0.72),
				new Point(vunit, vunit * 0.72)
			];
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case TERRAIN_INTEREST:
			var isize = scl * 0.75;
			var poly = [
				new Point( 0,               -0.75     * isize),
				new Point( 0.1875  * isize, -0.1875   * isize),
				new Point( 0.75    * isize, -0.1875   * isize),
				new Point( 0.28125 * isize,  0.140625 * isize),
				new Point( 0.46875 * isize,  0.703125 * isize),
				new Point( 0,                0.328125 * isize),
				new Point(-0.46875 * isize,  0.703125 * isize),
				new Point(-0.28125 * isize,  0.140625 * isize),
				new Point(-0.75    * isize, -0.1875   * isize),
				new Point(-0.1875  * isize, -0.1875   * isize)
			];
			drawPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_ARROWEL:
			var dunit = scl / 4;
			var poly = [
				new Point(0, -(dunit * 2.5)),
				new Point(dunit * 2.5, 0),
				new Point(dunit, 0),
				new Point(dunit, (dunit * 2.5)),
				new Point(-dunit, (dunit * 2.5)),
				new Point(-dunit, 0),
				new Point(-(dunit * 2.5), 0)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_ARROWES:
			var dunit = scl / 4;
			var vunit = scl / 3;
			var poly = [
				new Point(0, -(dunit * 1.5) - vunit),
				new Point(dunit * 1.5, - vunit),
				new Point(dunit * 0.5, - vunit),
				new Point(dunit * 0.5, (dunit * 1.5) - vunit),
				new Point(-dunit * 0.5, (dunit * 1.5) - vunit),
				new Point(-dunit * 0.5, - vunit),
				new Point(-(dunit * 1.5), - vunit)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_ARROWVL:
			var dunit = scl / 4;
			var poly = [
				new Point(0, -(dunit * 2.5)),
				new Point(dunit * 2.5, 0),
				new Point(dunit, 0),
				new Point(dunit, (dunit * 2.5)),
				new Point(-dunit, (dunit * 2.5)),
				new Point(-dunit, 0),
				new Point(-(dunit * 2.5), 0)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0) + 30);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_ARROWVS:
			var dunit = scl / 4;
			var vunit = scl / 3;
			var poly = [
				new Point(0, -(dunit * 1.5) - vunit),
				new Point(dunit * 1.5, - vunit),
				new Point(dunit * 0.5, - vunit),
				new Point(dunit * 0.5, (dunit * 1.5) - vunit),
				new Point(-dunit * 0.5, (dunit * 1.5) - vunit),
				new Point(-dunit * 0.5, - vunit),
				new Point(-(dunit * 1.5), - vunit)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0) + 30);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_ACCENTL:
			var rotangle = rotation * 60;
			if(mapGrain == GRAIN_VERTICAL) {
				var poly = [
					new Point(pptx[1] - orix, ppty[6] - oriy),
					new Point(pptx[5] - orix, ppty[6] - oriy),
					new Point(pptx[4] - orix, ppty[7] - oriy),
					new Point(pptx[2] - orix, ppty[7] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr);
			}
			else {
				var poly = [
					new Point(pptx[8] - orix, ppty[9] - oriy),
					new Point(pptx[8] - orix, ppty[13] - oriy),
					new Point(pptx[7] - orix, ppty[12] - oriy),
					new Point(pptx[7] - orix, ppty[10] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr);
			}
			break;
		case SYMBOL_ACCENTC:
			var vunit = scl * 0.7;
			var cunit = scl * 0.15;
			ctx.beginPath();
			var dotpoint = rotatePoint(new Point(0, -vunit), (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0) + 30);
			ctx.arc(orix + dotpoint.x, oriy + dotpoint.y, cunit, 0, Math.PI*2);
			ctx.fillStyle = matr;
			ctx.fill();
			break;
		case SYMBOL_ACCENTT:
			var dunit = scl / 4;
			var poly = [
				new Point(0, -dunit * 3.25),
				new Point(dunit, -dunit * 2),
				new Point(-dunit, -dunit * 2)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0) + 30);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_ACCENTD:
			var dunit = scl / 4;
			var poly = [
				new Point(0, -dunit * 3.25),
				new Point(dunit * 0.5, -dunit * 2.5),
				new Point(0, -dunit * 1.75),
				new Point(-dunit * 0.5, -dunit * 2.5)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0) + 30);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_PROHIBL:
			var dunit = scl / 2;
			var thick = scl / 12;
			ctx.beginPath();
			ctx.arc(orix, oriy, dunit, 0, Math.PI*2);
			ctx.closePath();
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			var poly = [
				new Point(-dunit, 0),
				new Point(dunit, 0)
			];
			poly = rotatePolygon(poly, (rotation * 60) + 60);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			break;
		case SYMBOL_PROHIBS:
			var dunit = scl / 4;
			var vunit = scl / 2.5;
			var thick = scl / 16;
			var ctrpoint = rotatePoint(new Point(0, -vunit), (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + ctrpoint.x, oriy + ctrpoint.y, dunit, 0, Math.PI*2);
			ctx.closePath();
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			var poly = [
				new Point(-dunit, 0),
				new Point(dunit, 0)
			];
			poly = rotatePolygon(poly, 60);
			drawPolygon(ctx, poly, orix + ctrpoint.x, oriy + ctrpoint.y, matr, thick);
			break;
		case SYMBOL_MARKER:
			var poly = [
				new Point( 0,             -0.75 * scl),
				new Point( 0.1875 * scl,  -0.1875 * scl),
				new Point( 0.75 * scl,    -0.1875 * scl),
				new Point( 0.28125 * scl,  0.140625 * scl),
				new Point( 0.46875 * scl,  0.703125 * scl),
				new Point( 0,              0.328125 * scl),
				new Point(-0.46875 * scl,  0.703125 * scl),
				new Point(-0.28125 * scl,  0.140625 * scl),
				new Point(-0.75 * scl,    -0.1875 * scl),
				new Point(-0.1875 * scl,  -0.1875 * scl)
			];
			if(mapGrain == GRAIN_HORIZONTAL) { poly = rotatePolygon(poly, 35); }
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_EVENT:
			hexpts = getNorthernHexPoints(orix, oriy, scl);
			pptx = hexpts[0];
			ppty = hexpts[1];
			var thick = Math.max((scl / 16), 1);
			var poly = [
				new Point(pptx[4] - orix, (ppty[7] - oriy)),
				new Point((pptx[3] - orix) - (thick / 2), (ppty[7] - oriy)),
				new Point((pptx[3] - orix) - (thick / 2), (ppty[6] - oriy)),
				new Point((pptx[3] - orix) + (thick / 2), (ppty[6] - oriy)),
				new Point((pptx[3] - orix) + (thick / 2), (ppty[7] - oriy)),
				new Point(pptx[2] - orix, (ppty[23] - oriy)),
				new Point(pptx[2] - orix, (ppty[23] - oriy) + thick),
				new Point(pptx[4] - orix, (ppty[7] - oriy) + thick)
			];
			poly = rotatePolygon(poly, (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = rotatePolygon(poly, 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = rotatePolygon(poly, 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = rotatePolygon(poly, 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = rotatePolygon(poly, 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = rotatePolygon(poly, 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case DECOR_TRAP:
			var thick = Math.max((scl / 12), 1);
			var tunit = scl / 3;
			var qunit = scl / 6;
			var ounit = scl / 10;
			var sunit = scl / 16;
			var rotangle = rotation * 60;
			var poly = [
				new Point(-tunit, -tunit),
				new Point(tunit, -tunit),
				new Point(tunit, tunit),
				new Point(-tunit, tunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-qunit, -qunit),
				new Point(qunit, qunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-qunit, qunit),
				new Point(qunit, -qunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			break;
		case DECOR_TRAPDOOR:
			var thick = Math.max((scl / 12), 1);
			var tunit = scl / 3;
			var qunit = scl / 4;
			var ounit = scl / 10;
			var sunit = scl / 16;
			var rotangle = rotation * 60;
			var poly = [
				new Point(-tunit, -tunit),
				new Point(tunit, -tunit),
				new Point(tunit, tunit),
				new Point(-tunit, tunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			ctx.beginPath();
			var dotpoint = rotatePoint(new Point(0, ounit), rotangle);
			ctx.arc(orix + dotpoint.x, oriy + dotpoint.y, ounit, 0, Math.PI*2);
			ctx.fillStyle = matr;
			ctx.fill();
			break;
		case DECOR_FLOORPIT:
			var thick = Math.max((scl / 12), 1);
			var tunit = scl / 3;
			var qunit = scl / 4;
			var ounit = scl / 6;
			var sunit = scl / 16;
			var rotangle = rotation * 60;
			var poly = [
				new Point(-tunit, -tunit),
				new Point(tunit, -tunit),
				new Point(tunit, tunit),
				new Point(-tunit, tunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-tunit, -tunit),
				new Point(tunit, tunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-ounit, -ounit),
				new Point(tunit, -ounit),
				new Point(tunit, tunit),
				new Point(-ounit, tunit)
			];
			poly = rotatePolygon(poly, rotangle);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case DECOR_TRIPWIRE:
			var thick = Math.max((scl / 12), 1);
			var funit = scl / 9;
			var rotangle = (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0);
			var poly = [
				new Point(-(funit * 9), 0),
				new Point(-(funit * 7), 0)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-(funit * 5), 0),
				new Point(-(funit * 3), 0)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-(funit * 1), 0),
				new Point((funit * 1), 0)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point((funit * 3), 0),
				new Point((funit * 5), 0)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point((funit * 7), 0),
				new Point((funit * 9), 0)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			break;
		case DECOR_CHESTL:
			var thick = Math.max((scl / 12), 1);
			var rotangle = rotation * 60;
			if(mapGrain == GRAIN_VERTICAL) {
				var poly = [
					new Point(pptx[1] - orix, ppty[7] - oriy),
					new Point(pptx[5] - orix, ppty[7] - oriy),
					new Point(pptx[5] - orix, ppty[11] - oriy),
					new Point(pptx[1] - orix, ppty[11] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[2] - orix, ppty[7] - oriy),
					new Point(pptx[2] - orix, ppty[11] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] - orix, ppty[7] - oriy),
					new Point(pptx[4] - orix, ppty[11] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[2] + ((pptx[3] - pptx[2]) / 2) - orix, ppty[10] - oriy),
					new Point(pptx[3] + ((pptx[4] - pptx[3]) / 2) - orix, ppty[10] - oriy),
					new Point(pptx[3] + ((pptx[4] - pptx[3]) / 2) - orix, ppty[11] - oriy),
					new Point(pptx[2] + ((pptx[3] - pptx[2]) / 2) - orix, ppty[11] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr, thick);
			}
			else {
				var poly = [
					new Point(pptx[3] - orix, ppty[9] - oriy),
					new Point(pptx[7] - orix, ppty[9] - oriy),
					new Point(pptx[7] - orix, ppty[13] - oriy),
					new Point(pptx[3] - orix, ppty[13] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] - orix, ppty[9] - oriy),
					new Point(pptx[4] - orix, ppty[13] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[6] - orix, ppty[9] - oriy),
					new Point(pptx[6] - orix, ppty[13] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] + ((pptx[5] - pptx[4]) / 2) - orix, ppty[12] - oriy),
					new Point(pptx[5] + ((pptx[6] - pptx[5]) / 2) - orix, ppty[12] - oriy),
					new Point(pptx[5] + ((pptx[6] - pptx[5]) / 2) - orix, ppty[13] - oriy),
					new Point(pptx[4] + ((pptx[5] - pptx[4]) / 2) - orix, ppty[13] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr, thick);
			}
			break;
		case DECOR_CHESTS:
			var thick = Math.max((scl / 12), 1);
			var rotangle = rotation * 60;
			if(mapGrain == GRAIN_VERTICAL) {
				var poly = [
					new Point(pptx[1] - orix, ppty[8] - oriy),
					new Point(pptx[5] - orix, ppty[8] - oriy),
					new Point(pptx[5] - orix, ppty[10] - oriy),
					new Point(pptx[1] - orix, ppty[10] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[2] - orix, ppty[8] - oriy),
					new Point(pptx[2] - orix, ppty[10] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] - orix, ppty[8] - oriy),
					new Point(pptx[4] - orix, ppty[10] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[2] + ((pptx[3] - pptx[2]) / 2) - orix, ppty[9] - oriy),
					new Point(pptx[3] + ((pptx[4] - pptx[3]) / 2) - orix, ppty[9] - oriy),
					new Point(pptx[3] + ((pptx[4] - pptx[3]) / 2) - orix, ppty[10] - oriy),
					new Point(pptx[2] + ((pptx[3] - pptx[2]) / 2) - orix, ppty[10] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr, thick);
			}
			else {
				var poly = [
					new Point(pptx[3] - orix, ppty[10] - oriy),
					new Point(pptx[7] - orix, ppty[10] - oriy),
					new Point(pptx[7] - orix, ppty[12] - oriy),
					new Point(pptx[3] - orix, ppty[12] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] - orix, ppty[10] - oriy),
					new Point(pptx[4] - orix, ppty[12] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[6] - orix, ppty[10] - oriy),
					new Point(pptx[6] - orix, ppty[12] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] + ((pptx[5] - pptx[4]) / 2) - orix, ppty[11] - oriy),
					new Point(pptx[5] + ((pptx[6] - pptx[5]) / 2) - orix, ppty[11] - oriy),
					new Point(pptx[5] + ((pptx[6] - pptx[5]) / 2) - orix, ppty[12] - oriy),
					new Point(pptx[4] + ((pptx[5] - pptx[4]) / 2) - orix, ppty[12] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr, thick);
			}
			break;
		case DECOR_KEY:
			var thick = Math.max((scl / 12), 1);
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			var rotangle = rotation * 60;
			if(mapGrain == GRAIN_VERTICAL) {
				var ringwidth = (pptx[2] - pptx[24]) / 2;
				var centerpoint = rotatePoint(new Point((pptx[24] + ringwidth) - orix, ppty[21] - oriy), rotangle);
				ctx.beginPath();
				ctx.arc(centerpoint.x + orix, centerpoint.y + oriy, ringwidth, 0, Math.PI*2);
				ctx.closePath();
				ctx.stroke();
				var linepoint1 = rotatePoint(new Point(pptx[2] - orix, ppty[9] - oriy), rotangle);
				var linepoint2 = rotatePoint(new Point((pptx[5] - orix) + Math.max(thick / 2, 1), ppty[9] - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				var toothbase = ppty[10] + ((ppty[11] - ppty[10]) / 2);
				linepoint1 = rotatePoint(new Point(pptx[5] - orix, ppty[9] - oriy), rotangle);
				linepoint2 = rotatePoint(new Point(pptx[5] - orix, toothbase - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				linepoint1 = rotatePoint(new Point(pptx[4] - orix, ppty[9] - oriy), rotangle);
				linepoint2 = rotatePoint(new Point(pptx[4] - orix, toothbase - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			}
			else {
				var ringwidth = (ppty[24] - ppty[2]) / 2;
				var centerpoint = rotatePoint(new Point((pptx[2] + ringwidth) - orix, ppty[23] - oriy), rotangle);
				ctx.beginPath();
				ctx.arc(centerpoint.x + orix, centerpoint.y + oriy, ringwidth, 0, Math.PI*2);
				ctx.closePath();
				ctx.stroke();
				var linepoint1 = rotatePoint(new Point((pptx[2] - orix) + (ringwidth * 2), ppty[11] - oriy), rotangle);
				var linepoint2 = rotatePoint(new Point(((pptx[8] - ((pptx[8] - pptx[7]) / 2)) - orix) + Math.max(thick / 2, 1), ppty[11] - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				var toothbase = ppty[12] + ((ppty[13] - ppty[12]) / 2);
				linepoint1 = rotatePoint(new Point((pptx[8] - ((pptx[8] - pptx[7]) / 2)) - orix, ppty[11] - oriy), rotangle);
				linepoint2 = rotatePoint(new Point((pptx[8] - ((pptx[8] - pptx[7]) / 2)) - orix, toothbase - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				linepoint1 = rotatePoint(new Point((pptx[7] - ((pptx[7] - pptx[6]) / 2)) - orix, ppty[11] - oriy), rotangle);
				linepoint2 = rotatePoint(new Point((pptx[7] - ((pptx[7] - pptx[6]) / 2)) - orix, toothbase - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			}
			break;
		case DECOR_PENTAGRM:
			var dunit = scl / 2;
			var thick = scl / 16;
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			ctx.beginPath();
			ctx.arc(orix, oriy, dunit, 0, Math.PI*2);
			ctx.closePath();
			ctx.stroke();
			var currLineCap = ctx.lineCap;
			ctx.lineCap = "round";
			var starrot = (rotation * 60);
			var starpoint = rotatePoint(new Point(pptx[0] - orix, (ppty[0] - oriy) - dunit), starrot);
			for(var pointloop = 0; pointloop < 5; pointloop++) {
				ctx.beginPath();
				ctx.moveTo(orix + starpoint.x, oriy + starpoint.y);
				starrot = starrot + 144;
				starpoint = rotatePoint(new Point(pptx[0] - orix, (ppty[0] - oriy) - dunit), starrot);
				ctx.lineTo(orix + starpoint.x, oriy + starpoint.y);
				ctx.closePath();
				ctx.stroke();
			}
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			ctx.lineCap = ctx.lineCap;
			break;
		case DECOR_GATE:
			var xunit = scl / 2;
			var yunit = scl / 2;
			var mxunit = scl / 9;
			var myunit = scl / 12;
			var thick = scl / 32;
			var rotradians = (rotation * 30) * (Math.PI / 180);
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			for(var el = 0; el < 5; el++) {
				ctx.beginPath();
				ctx.ellipse(orix, oriy, xunit, yunit, rotradians, 0, Math.PI*2);
				ctx.closePath();
				ctx.stroke();
				xunit = xunit - mxunit;
				yunit = yunit - myunit;
			}
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case DECOR_LIGHT:
			var dunit = scl / 6;
			var sunit = dunit * 2;
			var eunit = dunit * 3;
			var thick = scl / 16;
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			ctx.beginPath();
			ctx.arc(orix, oriy, dunit, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			var currLineCap = ctx.lineCap;
			ctx.lineCap = "round";
			var baserot = (rotation * 10) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0);
			var spoint = null;
			var epoint = null;
			for(var pointloop = 0; pointloop < 6; pointloop++) {
				spoint = rotatePoint(new Point(pptx[0] - orix, (ppty[0] - oriy) - sunit), baserot);
				epoint = rotatePoint(new Point(pptx[0] - orix, (ppty[0] - oriy) - eunit), baserot);
				ctx.beginPath();
				ctx.moveTo(orix + spoint.x, oriy + spoint.y);
				ctx.lineTo(orix + epoint.x, oriy + epoint.y);
				ctx.closePath();
				ctx.stroke();
				baserot = baserot + 60;
			}
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			ctx.lineCap = ctx.lineCap;
			break;
		case DECOR_SCONCE:
			var dunit = scl / 6;
			var sunit = dunit * 2;
			var eunit = dunit * 3;
			var thick = scl / 16;
			var rotangle = (rotation * 60);
			var rotradians = (rotangle - (mapGrain == GRAIN_HORIZONTAL ? 30 : 0)) * (Math.PI / 180);
			var mountpoint = rotatePoint(new Point(pptx[3] - orix, ppty[3] - oriy), rotangle);
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			ctx.beginPath();
			ctx.ellipse(mountpoint.x + orix, mountpoint.y + oriy, dunit, dunit, rotradians, 0, Math.PI);
			ctx.closePath();
			ctx.fill();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		default:
			break;
	}
}

function drawAnnotation(ctx, pptx, ppty, scl, ctyp, matr) {
	var fontStr = "bold " + scl + "px RobotoMono";
	var orix = pptx[0];
	var oriy = ppty[0];
	ctx.font = fontStr;
	ctx.fillStyle = matr;
	ctx.textAlign = "center";
	ctx.textBaseline = "middle";
	ctx.fillText(String.fromCharCode(ctyp - ANNOTATION_BASE), orix, oriy);
}

function drawAnnotations(ctx, pptx, ppty, scl, annots) {
	var fontStr = "bold " + scl + "px RobotoMono";
	if(annots.length > 1) {
		// draw first character of annotation pair
		var ctyp = annots[0].ctype;
		var matr = hexBase[annots[0].material];
		var orix = pptx[0] - (scl / 3);
		var oriy = ppty[0];
		ctx.font = fontStr;
		ctx.fillStyle = matr;
		ctx.textAlign = "center";
		ctx.textBaseline = "middle";
		ctx.fillText(String.fromCharCode(ctyp - ANNOTATION_BASE), orix, oriy);
		// draw second character of annotation pair
		ctyp = annots[1].ctype;
		matr = hexBase[annots[1].material];
		orix = pptx[0] + (scl / 3);
		oriy = ppty[0];
		ctx.font = fontStr;
		ctx.fillStyle = matr;
		ctx.textAlign = "center";
		ctx.textBaseline = "middle";
		ctx.fillText(String.fromCharCode(ctyp - ANNOTATION_BASE), orix, oriy);
	}
	else {
		// draw single annotation character
		var ctyp = annots[0].ctype;
		var matr = hexBase[annots[0].material];
		var orix = pptx[0];
		var oriy = ppty[0];
		ctx.font = fontStr;
		ctx.fillStyle = matr;
		ctx.textAlign = "center";
		ctx.textBaseline = "middle";
		ctx.fillText(String.fromCharCode(ctyp - ANNOTATION_BASE), orix, oriy);
	}
}

function drawAnnotationEraser(ctx, pptx, ppty, scl) {
	var cunit = scl * 0.5;
	var xs = pptx[0] - cunit;
	var ys = ppty[0] - cunit;
	var xe = pptx[0] + cunit;
	var ye = ppty[0] + cunit;
	ctx.beginPath();
	for(yv = ys; yv < ye; yv = yv + 3.0) {
		ctx.moveTo(xs, yv);
		ctx.lineTo(xe, yv);
	}
	ctx.moveTo(xe, ye);
	ctx.closePath();
	ctx.strokeStyle = hexBase[3];
	ctx.stroke();
}

function drawHexOutline(ctx, pptx, ppty, clr) {
	if(typeof clr != "undefined") {} else { clr = hexOutline; }
	ctx.beginPath();
	ctx.moveTo(pptx[1], ppty[1]);
	ctx.lineTo(pptx[5], ppty[5]);
	ctx.lineTo(pptx[9], ppty[9]);
	ctx.lineTo(pptx[13], ppty[13]);
	ctx.lineTo(pptx[17], ppty[17]);
	ctx.lineTo(pptx[21], ppty[21]);
	ctx.closePath();
	ctx.strokeStyle = clr;
	ctx.lineWidth = outlineThickness;
	ctx.stroke();
}

function drawMegahex(ctx, col, row, clr, thick, dotwt) {
	var hexpts;
	ctx.fillStyle = clr;
	ctx.strokeStyle = clr;
	ctx.lineWidth = thick;
	// draw center hex
	var hcol = col;
	var hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
	}
	// draw hex above
	hcol = col - (mapGrain == GRAIN_HORIZONTAL ? ((row - 1) % 2) : 0);
	hrow = row - 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.stroke();
	}
	// draw hex upper-left (need to calculate based on row/column)
	hcol = col - 1;
	hrow = row - (mapGrain == GRAIN_HORIZONTAL ? 0 : ((col - 1) % 2));
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.stroke();
	}
	// draw hex upper-right (need to calculate based on row/column)
	hcol = col + (mapGrain == GRAIN_HORIZONTAL ? (row % 2) : 1);
	hrow = row - (mapGrain == GRAIN_HORIZONTAL ? 1 : ((col + 1) % 2));
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.stroke();
	}
	// draw hex below
	hcol = col + (mapGrain == GRAIN_HORIZONTAL ? Math.abs(row % 2) : 0);
	hrow = row + 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.stroke();
	}
	// draw hex lower-left
	hcol = col - (mapGrain == GRAIN_HORIZONTAL ? ((row + 1) % 2) : 1);
	hrow = row + 1 - (mapGrain == GRAIN_HORIZONTAL ? 0 : ((col - 1) % 2));
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.stroke();
	}
	// draw hex upper-right
	hcol = col + 1;
	hrow = row + (mapGrain == GRAIN_HORIZONTAL ? 0 : (1 - ((col + 1) % 2)));
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.stroke();
	}
	ctx.lineWidth = DEFAULT_LINE_THICKNESS;
}

function drawSquareMegahexVertical(ctx, col, row, clr, thick, dotwt) {
	var hexpts;
	ctx.fillStyle = clr;
	ctx.strokeStyle = clr;
	ctx.lineWidth = thick;
	// draw center upper hex
	var hcol = col;
	var hrow = row + ((col + 1) % 2);
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.stroke();
	}
	// draw center lower hex
	hcol = col;
	hrow = row + 1 + ((col + 1) % 2);
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.stroke();
	}
	// draw upper left hex
	hcol = col - 1;
	hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.stroke();
	}
	// draw center left hex
	hcol = col - 1;
	hrow = row + 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.stroke();
	}
	// draw lower left hex
	hcol = col - 1;
	hrow = row + 2;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.stroke();
	}
	// draw upper right hex
	hcol = col + 1;
	hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.stroke();
	}
	// draw center right hex
	hcol = col + 1;
	hrow = row + 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.stroke();
	}
	// draw lower right hex
	hcol = col + 1;
	hrow = row + 2;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.stroke();
	}
	ctx.lineWidth = DEFAULT_LINE_THICKNESS;
}

function drawSquareMegahexHorizontal(ctx, col, row, clr, thick, dotwt) {
	var hexpts;
	ctx.fillStyle = clr;
	ctx.strokeStyle = clr;
	ctx.lineWidth = thick;
	// draw center left hex
	var hcol = col - ((row + 1) % 2);
	var hrow = row + 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.stroke();
	}
	// draw center right hex
	hcol = col + (row % 2);
	hrow = row + 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.stroke();
	}
	// draw upper center hex
	hcol = col;
	hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.stroke();
	}
	// draw lower center hex
	hcol = col;
	hrow = row + 2;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.stroke();
	}
	// draw upper left hex
	hcol = col - 1;
	hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.stroke();
	}
	// draw lower left hex
	hcol = col - 1;
	hrow = row + 2;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.stroke();
	}
	// draw upper right hex
	hcol = col + 1;
	hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.stroke();
	}
	// draw lower right hex
	hcol = col + 1;
	hrow = row + 2;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.stroke();
	}
	ctx.lineWidth = DEFAULT_LINE_THICKNESS;
}

function drawMegahexGrid(grain, startX, startY, offsetFactor) {
	var plotX = startX;
	var plotY = startY;
	if(grain == GRAIN_HORIZONTAL) {
		var xOffset = 0;
		var yOffset = 0;
		var nextIncr = (startY == -1);
		while(plotY < gridy + 2) {
			while(plotX < gridx + 2) {
				drawMegahex(mhx, plotX, plotY + yOffset, mhexOutline, megahexWeight, megahexDotWeight);
				if(yOffset == 0) {
					plotX += (2 + Math.abs(plotY % 2));
					yOffset = 1;
				}
				else if(yOffset == 1) {
					plotX += 2;
					yOffset = -1;
				}
				else {
					plotX += (3 - Math.abs(plotY % 2));
					yOffset = 0;
				}
			}
			plotY += 3;
			if(nextIncr) { xOffset++; }
			if(xOffset > 1) {
				xOffset = -5;
				nextIncr = true;
			}
			nextIncr = !nextIncr;
			yOffset = 0;
			plotX = startX + xOffset;
		}
	}
	else {
		var prestep = offsetFactor;
		while(plotY < gridy + 5) {
			var step = 0;
			var stepBias = prestep;
			var loopRise = 0;
			var colpos = plotX;
			while(colpos < gridx + 2) {
				drawMegahex(mhx, colpos, plotY - Math.floor((step + stepBias) / 2) + loopRise, mhexOutline, megahexWeight, megahexDotWeight);
				colpos += 3;
				step++;
				if(step > 5) { step = 0; colpos--; stepBias = Math.abs(stepBias - 1); if(stepBias == 0) { loopRise--; } }
			}
			if(plotX == -1) {
				plotX = 1;
				plotY += 2;
				prestep = 0;
			}
			else if(plotX == 1) {
				plotX = 0;
				plotY += 3;
				prestep = 1;
			}
			else if(plotX == 0) {
				plotX = -1;
				plotY += 2;
				prestep = 0;
			}
		}
	}
}

function drawSquareMegahexRow(colpos, rowpos, isoddrow) {
	var drawSquare = isoddrow;
	while(colpos < gridx + 2) {
		if(drawSquare) {
			drawSquareMegahexVertical(mhx, colpos, rowpos, mhexOutline, megahexWeight, megahexDotWeight);
		}
		else {
			drawMegahex(mhx, colpos, rowpos + 1, mhexOutline, megahexWeight, megahexDotWeight);
		}
		drawSquare = !drawSquare;
		colpos += 3;
	}
}

function drawSquareMegahexCol(colpos, rowpos, isoddrow) {
	var drawSquare = isoddrow;
	while(rowpos < gridy + 2) {
		if(drawSquare) {
			drawSquareMegahexHorizontal(mhx, colpos, rowpos, mhexOutline, megahexWeight, megahexDotWeight);
		}
		else {
			drawMegahex(mhx, colpos, rowpos + 1, mhexOutline, megahexWeight, megahexDotWeight);
		}
		drawSquare = !drawSquare;
		rowpos += 3;
	}
}

function correctMegahexOffsetsForGrain() {
	if(mapGrain == GRAIN_HORIZONTAL) {
		if(mhexOverlayY < -2) {
			mhexOverlayY = 0;
		}
		else if(mhexOverlayY > 0) {
			mhexOverlayY = -2;
		}
		if(mhexOverlayX < -6) {
			mhexOverlayX = 0;
		}
		else if(mhexOverlayX > 0) {
			mhexOverlayX = -6;
		}
	}
	else {
		if(mhexOverlayY < -6) {
			mhexOverlayY = 0;
		}
		else if(mhexOverlayY > 0) {
			mhexOverlayY = -6;
		}
		if(mhexOverlayX < -1) {
			mhexOverlayX = 1;
		}
		else if(mhexOverlayX > 1) {
			mhexOverlayX = -1;
		}
	}
}

function shiftMegahexes(dir) {
	if(mapGrain == GRAIN_HORIZONTAL) {
		switch(dir) {
			case CARDINAL_N:
				mhexOverlayY -= 1;
				break;
			case CARDINAL_S:
				mhexOverlayY += 1;
				break;
			case CARDINAL_W:
				mhexOverlayX -= 1;
				break;
			case CARDINAL_E:
				mhexOverlayX += 1;
				break;
			default:
				break;
		}
	}
	else {
		switch(dir) {
			case CARDINAL_N:
				mhexOverlayY -= 1;
				break;
			case CARDINAL_S:
				mhexOverlayY += 1;
				break;
			case CARDINAL_W:
				mhexOverlayX -= 1;
				break;
			case CARDINAL_E:
				mhexOverlayX += 1;
				break;
			default:
				break;
		}
	}
	correctMegahexOffsetsForGrain();
	mhexOverlayS = 0;
	if(mhexOverlayX == 0) { mhexOverlayS = 1; }
	if(mhexOverlayY == 0 && mhexOverlayX != 0) { mhexOverlayS = 0; }
	overlayMegahexes();
}

function overlayMegahexes() {
	mhx.clearRect(0, 0, mhexes.width, mhexes.height);
	if(mhexType == MEGAHEX_TYPE_SQUARE) {
		if(mapGrain == GRAIN_HORIZONTAL) {
			var colnum = 1;
			var oddrow = true;
			while(colnum < gridx) {
				drawSquareMegahexCol(colnum, 0, oddrow);
				colnum += 2 + (oddrow ? 0 : 1);
				oddrow = !oddrow;
			}
		}
		else {
			var rownum = 0;
			var oddrow = true;
			while(rownum < gridy) {
				drawSquareMegahexRow(1, rownum, oddrow);
				rownum += 2 + (oddrow ? 0 : 1);
				oddrow = !oddrow;
			}
		}
	}
	else {
		drawMegahexGrid(mapGrain, mhexOverlayX, mhexOverlayY, mhexOverlayS);
	}
	doubleBuffer();
}

function toggleMegahexes(show) {
	megahexesOn = show;
	drawMap();
	if(show) {
		correctMegahexOffsetsForGrain();
		overlayMegahexes();
		document.getElementById('mhexhide').className = "palettetool";
		document.getElementById('mhexshow').className = "palettetool" + (mhexType == MEGAHEX_TYPE_STANDARD ? " toolactive" : "");
		document.getElementById('mhexsqr').className = "palettetool" + (mhexType == MEGAHEX_TYPE_SQUARE ? " toolactive" : "");
	}
	else {
		document.getElementById('mhexshow').className = "palettetool";
		document.getElementById('mhexsqr').className = "palettetool";
		document.getElementById('mhexhide').className = "palettetool toolactive";
	}
	cnv.clearRect(0, 0, canvas.width, canvas.height);
	doubleBuffer();
}

function setMegahexType(mhtype) {
	mhexType = mhtype;
}

function standardMegahexes() {
	setMegahexType(MEGAHEX_TYPE_STANDARD);
	toggleMegahexes(true);
}

function squareMegahexes() {
	setMegahexType(MEGAHEX_TYPE_SQUARE);
	toggleMegahexes(true);
}

function plotHex(ctx, x, y, isCursor) {
	var hexcor = getHexCoords(drawscale);
	var hexptx = getHexX(y, x);
	var hexpty = getHexY(y, x);
	var hexfin = getHexFineCoords(x, y);
	var hexptr = hexarray[x][y];

	/* draw base components */
	if(hexptr.bases.length > 0) {
		for(var i = 0; i < hexptr.bases.length; i++) {
//			if(hexptr.bases[i].ctype > HEX_EMPTY) {
				drawComponent(ctx, hexptx, hexpty, hexcor[0], hexcor[1], hexptr.bases[i].ctype, hexBase[hexptr.bases[i].material]);
//			}
		}
	}
	if(isCursor && activeLayer == LAYER_BASE) {
		if(translucencyEnabled) { ctx.globalAlpha = TRANSLUCE; }
		drawComponent(ctx, hexptx, hexpty, hexcor[0], hexcor[1], activeComponent, (activeComponent == HEX_EMPTY ? hexBase[MATERIAL_EMPTY] : hexBase[activeMaterial]));
		if(translucencyEnabled) { ctx.globalAlpha = OPAQUE; }
	}

	/* draw overlay components */
	if(hexptr.overlays.length > 0) {
		for(var i = 0; i < hexptr.overlays.length; i++) {
			drawComponent(ctx, hexptx, hexpty, hexcor[0], hexcor[1], hexptr.overlays[i].ctype, hexOverlay[hexptr.overlays[i].material]);
		}
	}
	if(isCursor && activeLayer == LAYER_OVERLAY) {
		if(translucencyEnabled) { ctx.globalAlpha = TRANSLUCE; }
		drawComponent(ctx, hexptx, hexpty, hexcor[0], hexcor[1], activeComponent, (activeComponent == HEX_EMPTY ? hexBase[MATERIAL_EMPTY] : hexOverlay[activeMaterial]));
		if(translucencyEnabled) { ctx.globalAlpha = OPAQUE; }
	}

	/* draw outline */
	drawHexOutline(ctx, hexfin[0], hexfin[1]);

	/* draw feature components */
	if(hexptr.features.length > 0) {
		for(var i = 0; i < hexptr.features.length; i++) {
			drawFeature(ctx, hexfin[0], hexfin[1], drawscale, hexptr.features[i].ctype, hexBase[hexptr.features[i].material]);
		}
	}

	/* draw annotation components */
	if(hexptr.annotations.length > 0) {
		drawAnnotations(ctx, hexfin[0], hexfin[1], drawscale, hexptr.annotations);
	}

	/* if this is the active cursor brush, draw it on top of the other elements */
	if(isCursor) {
		if(translucencyEnabled) { ctx.globalAlpha = TRANSLUCE; }
		if(activeLayer == LAYER_FEATURE) {
			drawFeature(ctx, hexfin[0], hexfin[1], drawscale, activeFeature, (activeFeature == FEATURE_EMPTY ? hexBase[MATERIAL_EMPTY] : hexBase[activeMaterial]));
		}
		else if(activeLayer == LAYER_TERRAIN) {
			drawFeature(ctx, hexfin[0], hexfin[1], drawscale, activeTerrain, (activeTerrain == TERRAIN_EMPTY ? hexBase[MATERIAL_EMPTY] : hexBase[activeMaterial]));
		}
		else if(activeLayer == LAYER_SYMBOLS) {
			drawFeature(ctx, hexfin[0], hexfin[1], drawscale, activeSymbol, (activeSymbol == SYMBOL_EMPTY ? hexBase[MATERIAL_EMPTY] : hexBase[activeMaterial]));
		}
		else if(activeLayer == LAYER_DECOR) {
			drawFeature(ctx, hexfin[0], hexfin[1], drawscale, activeDecor, (activeDecor == DECOR_EMPTY ? hexBase[MATERIAL_EMPTY] : hexBase[activeMaterial]));
		}
		else if(activeLayer == LAYER_ANNOTATE) {
			if(activeAnnotation == ANNOTATION_BASE) {
				drawAnnotationEraser(ctx, hexfin[0], hexfin[1], drawscale);
			}
			else {
				drawAnnotation(ctx, hexfin[0], hexfin[1], drawscale, activeAnnotation, hexBase[activeMaterial]);
			}
		}
		if(translucencyEnabled) { ctx.globalAlpha = OPAQUE; }
	}
}

function plotHexes(ctx) {
	for(x = 0; x < hexarray.length; x++) {
		for(y = 0; y < hexarray[x].length; y++) {
			plotHex(ctx, x, y, false);
		}
	}
}

function plotTool(ctx, x, y) {
	plotHex(ctx, x, y, true);
	if(modeStartX > -1 && modeStartY > -1) {
		if(activeMode == MODE_LINE) {
			var slope = getSlope(modeStartX, modeStartY, x, y);
			if(mapGrain == GRAIN_VERTICAL) {
				if(slope <= 0.25) { y = modeStartY; }
				else if(slope >= 0.75) { x = modeStartX; }
				else { y = modeStartY + (Math.round(Math.abs(modeStartX - x) * 0.5) * (y < modeStartY ? -1 : 1)) + (y < modeStartY ? ((modeStartX % 2) * Math.abs((x % 2) - 1)) : -(x % 2) + Math.abs((modeStartX % 2) * (x % 2))); }
			}
			else {
				if(slope <= 1.0) { y = modeStartY; }
				else if(slope >= 3.0) { x = modeStartX; }
				else { x = modeStartX + (Math.round(Math.abs(modeStartY - y) * 0.5) * (x < modeStartX ? -1 : 1)) + (x < modeStartX ? ((modeStartY % 2) * Math.abs((y % 2) - 1)) : -(y % 2) + Math.abs((modeStartY % 2) * (y % 2))); }
			}
			ctx.beginPath();
			ctx.moveTo(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX));
			ctx.lineTo(getHexX(y, x), getHexY(y, x));
			ctx.closePath();
			ctx.strokeStyle = toolOutline;
			ctx.stroke();
		}
		else if(activeMode == MODE_RECT) {
			ctx.strokeStyle = toolOutline;
			ctx.rect(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX), getHexX(y, x) - getHexX(modeStartY, modeStartX), getHexY(y, x) - getHexY(modeStartY, modeStartX));
			ctx.stroke();
		}
		else if(activeMode == MODE_AREA) {
			ctx.fillStyle = toolOutline;
			ctx.fillRect(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX), getHexX(y, x) - getHexX(modeStartY, modeStartX), getHexY(y, x) - getHexY(modeStartY, modeStartX));
		}
		else if(activeMode == MODE_SKIP) {
			if(mapGrain == GRAIN_VERTICAL) {
				ctx.beginPath();
				ctx.moveTo(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX));
				ctx.lineTo(getHexX(y, x), getHexY(modeStartY, modeStartX));
				ctx.closePath();
				ctx.strokeStyle = toolOutline;
				ctx.stroke();
			}
			else {
				ctx.beginPath();
				ctx.moveTo(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX));
				ctx.lineTo(getHexX(modeStartY, modeStartX), getHexY(y, x));
				ctx.closePath();
				ctx.strokeStyle = toolOutline;
				ctx.stroke();
			}
		}
		else if(activeMode == MODE_SMART) {
			ctx.beginPath();
			ctx.moveTo(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX));
			ctx.lineTo(getHexX(y, x), getHexY(modeStartY, modeStartX));
			ctx.closePath();
			ctx.strokeStyle = toolOutline;
			ctx.stroke();
		}
		else if(activeMode == MODE_LROOM || activeMode == MODE_SROOM) {
			if(Math.abs(modeStartX - x) % 2 > 0) {
				if(x < modeStartX) {
					if(x > 0) {
						x = x - 1;
					}
					else {
						x = x + 1;
					}
				}
				else {
					if(x < (gridx - 1)) {
						x = x + 1;
					}
					else {
						x = x - 1;
					}
				}
			}
			ctx.strokeStyle = toolOutline;
			ctx.rect(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX), getHexX(y, x) - getHexX(modeStartY, modeStartX), getHexY(y, x) - getHexY(modeStartY, modeStartX));
			ctx.stroke();
		}
	}
}

function hideCursor(evt) {
	doubleBuffer();
}

function drawMap() {
	buf.clearRect(0, 0, buffer.width, buffer.height);
	plotHexes(buf);
	doubleBuffer();
}

function setPaintMode(pmode) {
	if(pmode != activeMode) {
		modeButton = document.getElementById("mode" + pmode);
		if(true == modeButton.disabled) {
			return;
		}
		document.getElementById("mode" + activeMode).className = "palettetool";
		activeMode = pmode;
		modeButton.className = "palettetool toolactive";
		if(pmode == MODE_RECT || pmode == MODE_AREA) {
			setComponent(HEX_WHOLE);
		}
		else if(pmode == MODE_LROOM) {
			if(activeLayer != LAYER_ALL) {
				setLayer(LAYER_BASE);
				setComponent(HEX_WHOLE);
				setPaintMode(MODE_LROOM);
			}
		}
		else if(pmode == MODE_SROOM) {
			if(activeLayer != LAYER_ALL) {
				setLayer(LAYER_BASE);
				setComponent(HEX_WHOLE);
				setPaintMode(MODE_SROOM);
			}
		}
		endModeDraw();
	}
}

function setLayer(typ) {
	if(typ != activeLayer) {
		document.getElementById("layer" + activeLayer).className = "palettetool";
		activeLayer = typ;
		document.getElementById("layer" + typ).className = "palettetool toolactive";
		if(activeMode == MODE_LROOM || activeMode == MODE_SROOM) { setPaintMode(MODE_PAINT); }
		if(typ == LAYER_BASE || typ == LAYER_OVERLAY) {
			document.getElementById("pblockMaterial").style.display = "block";
			document.getElementById("pblockComponent").style.display = "block";
			document.getElementById("pblockFeature").style.display = "none";
			document.getElementById("pblockTerrain").style.display = "none";
			document.getElementById("pblockSymbol").style.display = "none";
			document.getElementById("pblockDecor").style.display = "none";
			document.getElementById("pblockMegahex").style.display = "none";
			document.getElementById("pblockAnnotate").style.display = "none";
			updatePreview();
		}
		else if(typ == LAYER_FEATURE) {
			document.getElementById("pblockMaterial").style.display = "block";
			document.getElementById("pblockComponent").style.display = "none";
			document.getElementById("pblockFeature").style.display = "block";
			document.getElementById("pblockTerrain").style.display = "none";
			document.getElementById("pblockSymbol").style.display = "none";
			document.getElementById("pblockDecor").style.display = "none";
			document.getElementById("pblockMegahex").style.display = "none";
			document.getElementById("pblockAnnotate").style.display = "none";
			updateFeaturePreview();
		}
		else if(typ == LAYER_TERRAIN) {
			document.getElementById("pblockMaterial").style.display = "block";
			document.getElementById("pblockComponent").style.display = "none";
			document.getElementById("pblockFeature").style.display = "none";
			document.getElementById("pblockTerrain").style.display = "block";
			document.getElementById("pblockSymbol").style.display = "none";
			document.getElementById("pblockDecor").style.display = "none";
			document.getElementById("pblockMegahex").style.display = "none";
			document.getElementById("pblockAnnotate").style.display = "none";
			updateTerrainPreview();
		}
		else if(typ == LAYER_DECOR) {
			document.getElementById("pblockMaterial").style.display = "block";
			document.getElementById("pblockComponent").style.display = "none";
			document.getElementById("pblockFeature").style.display = "none";
			document.getElementById("pblockTerrain").style.display = "none";
			document.getElementById("pblockSymbol").style.display = "none";
			document.getElementById("pblockDecor").style.display = "block";
			document.getElementById("pblockMegahex").style.display = "none";
			document.getElementById("pblockAnnotate").style.display = "none";
			updateDecorPreview();
		}
		else if(typ == LAYER_SYMBOLS) {
			document.getElementById("pblockMaterial").style.display = "block";
			document.getElementById("pblockComponent").style.display = "none";
			document.getElementById("pblockFeature").style.display = "none";
			document.getElementById("pblockTerrain").style.display = "none";
			document.getElementById("pblockSymbol").style.display = "block";
			document.getElementById("pblockDecor").style.display = "none";
			document.getElementById("pblockMegahex").style.display = "none";
			document.getElementById("pblockAnnotate").style.display = "none";
			updateSymbolPreview();
		}
		else if(typ == LAYER_ANNOTATE) {
			document.getElementById("pblockMaterial").style.display = "block";
			document.getElementById("pblockComponent").style.display = "none";
			document.getElementById("pblockFeature").style.display = "none";
			document.getElementById("pblockTerrain").style.display = "none";
			document.getElementById("pblockSymbol").style.display = "none";
			document.getElementById("pblockDecor").style.display = "none";
			document.getElementById("pblockMegahex").style.display = "none";
			document.getElementById("pblockAnnotate").style.display = "block";
		}
		else if(typ == LAYER_MEGAHEX) {
			document.getElementById("pblockMaterial").style.display = "none";
			document.getElementById("pblockComponent").style.display = "none";
			document.getElementById("pblockFeature").style.display = "none";
			document.getElementById("pblockTerrain").style.display = "none";
			document.getElementById("pblockSymbol").style.display = "none";
			document.getElementById("pblockDecor").style.display = "none";
			document.getElementById("pblockMegahex").style.display = "block";
			document.getElementById("pblockAnnotate").style.display = "none";
		}
		else {
			document.getElementById("pblockMaterial").style.display = "block";
			updatePreview();
		}
	}
}

function setMaterial(typ) {
	if(typ != activeMaterial) {
		document.getElementById("material" + activeMaterial).className = "palettetool";
		activeMaterial = typ;
		document.getElementById("material" + typ).className = "palettetool toolactive";
		if(activeLayer == LAYER_FEATURE) {
			updateFeaturePreview();
		}
		else if(activeLayer == LAYER_TERRAIN) {
			updateTerrainPreview();
		}
		else if(activeLayer == LAYER_DECOR) {
			updateDecorPreview();
		}
		else if(activeLayer == LAYER_SYMBOLS) {
			updateSymbolPreview();
		}
		else {
			updatePreview();
		}
	}
}

function setComponent(typ) {
	if(typ != getBaseGlyph(activeComponent)) {
		document.getElementById("tool" + getBaseGlyph(activeComponent)).className = "palettetool";
		activeComponent = typ;
		document.getElementById("tool" + typ).className = "palettetool toolactive";
		updatePreview();
		if(activeMode == MODE_LROOM || activeMode == MODE_SROOM) { setPaintMode(MODE_PAINT); }
	}
}

function setFeature(typ) {
	if(typ != getBaseGlyph(activeFeature)) {
		document.getElementById("feature" + getBaseGlyph(activeFeature)).className = "palettetool";
		activeFeature = typ;
		document.getElementById("feature" + typ).className = "palettetool toolactive";
		updateFeaturePreview();
		if(activeMode == MODE_LROOM || activeMode == MODE_SROOM) { setPaintMode(MODE_PAINT); }
	}
}

function setTerrain(typ) {
	if(typ != activeTerrain) {
		document.getElementById("terrain" + activeTerrain).className = "palettetool";
		activeTerrain = typ;
		document.getElementById("terrain" + typ).className = "palettetool toolactive";
		updateTerrainPreview();
		if(activeMode == MODE_LROOM || activeMode == MODE_SROOM) { setPaintMode(MODE_PAINT); }
	}
}

function setSymbol(typ) {
	if(typ != getBaseGlyph(activeSymbol)) {
		document.getElementById("symbol" + getBaseGlyph(activeSymbol)).className = "palettetool";
		activeSymbol = typ;
		document.getElementById("symbol" + typ).className = "palettetool toolactive";
		updateSymbolPreview();
		if(activeMode == MODE_LROOM || activeMode == MODE_SROOM) { setPaintMode(MODE_PAINT); }
	}
}

function setDecor(typ) {
	if(typ != getBaseGlyph(activeDecor)) {
		document.getElementById("decor" + getBaseGlyph(activeDecor)).className = "palettetool";
		activeDecor = typ;
		document.getElementById("decor" + typ).className = "palettetool toolactive";
		updateDecorPreview();
		if(activeMode == MODE_LROOM || activeMode == MODE_SROOM) { setPaintMode(MODE_PAINT); }
	}
}

function updatePreview() {
	scratchTemp = outlineThickness;
	outlineThickness = DEFAULT_LINE_THICKNESS;
	var hexraw = getHexRawCoords(PREVIEW_SCALE, PREVIEW_SIZEX, PREVIEW_SIZEY);
	var hexcor = getHexCoords(PREVIEW_SCALE);
	prv.clearRect(0, 0, PREVIEW_SIZEX * 2, PREVIEW_SIZEY * 2);
	drawComponent(prv, PREVIEW_SIZEX, 24, hexcor[0], hexcor[1], activeComponent, (activeLayer == LAYER_OVERLAY ? hexOverlay[activeMaterial] : hexBase[activeMaterial]));
	drawHexOutline(prv, hexraw[0], hexraw[1]);
	outlineThickness = scratchTemp;
}

function updateFeaturePreview() {
	scratchTemp = outlineThickness;
	outlineThickness = DEFAULT_LINE_THICKNESS;
	var hexraw = getHexRawCoords(PREVIEW_SCALE, PREVIEW_SIZEX, PREVIEW_SIZEY);
	fpr.clearRect(0, 0, PREVIEW_SIZEX * 2, PREVIEW_SIZEY * 2);
	drawFeature(fpr, hexraw[0], hexraw[1], PREVIEW_SCALE, activeFeature, hexOverlay[activeMaterial]);
	drawHexOutline(fpr, hexraw[0], hexraw[1]);
	outlineThickness = scratchTemp;
}

function updateTerrainPreview() {
	scratchTemp = outlineThickness;
	outlineThickness = DEFAULT_LINE_THICKNESS;
	var hexraw = getHexRawCoords(PREVIEW_SCALE, PREVIEW_SIZEX, PREVIEW_SIZEY);
	tpr.clearRect(0, 0, PREVIEW_SIZEX * 2, PREVIEW_SIZEY * 2);
	drawFeature(tpr, hexraw[0], hexraw[1], PREVIEW_SCALE, activeTerrain, hexOverlay[activeMaterial]);
	drawHexOutline(tpr, hexraw[0], hexraw[1]);
	outlineThickness = scratchTemp;
}

function updateSymbolPreview() {
	scratchTemp = outlineThickness;
	outlineThickness = DEFAULT_LINE_THICKNESS;
	var hexraw = getHexRawCoords(PREVIEW_SCALE, PREVIEW_SIZEX, PREVIEW_SIZEY);
	spr.clearRect(0, 0, PREVIEW_SIZEX * 2, PREVIEW_SIZEY * 2);
	drawFeature(spr, hexraw[0], hexraw[1], PREVIEW_SCALE, activeSymbol, hexOverlay[activeMaterial]);
	drawHexOutline(spr, hexraw[0], hexraw[1]);
	outlineThickness = scratchTemp;
}

function updateDecorPreview() {
	scratchTemp = outlineThickness;
	outlineThickness = DEFAULT_LINE_THICKNESS;
	var hexraw = getHexRawCoords(PREVIEW_SCALE, PREVIEW_SIZEX, PREVIEW_SIZEY);
	dpr.clearRect(0, 0, PREVIEW_SIZEX * 2, PREVIEW_SIZEY * 2);
	drawFeature(dpr, hexraw[0], hexraw[1], PREVIEW_SCALE, activeDecor, hexOverlay[activeMaterial]);
	drawHexOutline(dpr, hexraw[0], hexraw[1]);
	outlineThickness = scratchTemp;
}

function getBaseGlyph(currGlyph) {
	return (currGlyph < 10 ? currGlyph : (currGlyph - (currGlyph % 10)));
}

function rotateComponent(clockwise) {
	var baseComponent = getBaseGlyph(activeComponent);
	if(baseComponent >= 10) {
		if(typeof clockwise != "undefined") {} else { clockwise = true; }
		var currRotation = activeComponent % 10;
		if(clockwise) {
			currRotation++;
		}
		else {
			currRotation--;
		}
		while(currRotation < 0) {
			currRotation += 6;
		}
		while(currRotation >= 6) {
			currRotation -= 6;
		}
		activeComponent = baseComponent + currRotation;
		updatePreview();
	}
}

function flipComponent() {
	if(activeComponent >= 10) {
		var currRotation = activeComponent % 10;
		var baseComponent = activeComponent - currRotation;
		currRotation = currRotation + 3;
		while(currRotation < 0) {
			currRotation += 6;
		}
		while(currRotation >= 6) {
			currRotation -= 6;
		}
		activeComponent = baseComponent + currRotation;
		updatePreview();
	}
}

function rotateFeature(clockwise) {
	var baseFeature = getBaseGlyph(activeFeature);
	if(typeof clockwise != "undefined") {} else { clockwise = true; }
	var currRotation = activeFeature % 10;
	if(clockwise) {
		currRotation++;
	}
	else {
		currRotation--;
	}
	while(currRotation < 0) {
		currRotation += 6;
	}
	while(currRotation >= 6) {
		currRotation -= 6;
	}
	activeFeature = baseFeature + currRotation;
	updateFeaturePreview();
}

function flipFeature() {
	var currRotation = activeFeature % 10;
	var baseFeature = activeFeature - currRotation;
	currRotation = currRotation + 3;
	while(currRotation < 0) {
		currRotation += 6;
	}
	while(currRotation >= 6) {
		currRotation -= 6;
	}
	activeFeature = baseFeature + currRotation;
	updateFeaturePreview();
}

function rotateSymbol(clockwise) {
	var baseSymbol = getBaseGlyph(activeSymbol);
	if(typeof clockwise != "undefined") {} else { clockwise = true; }
	var currRotation = activeSymbol % 10;
	if(clockwise) {
		currRotation++;
	}
	else {
		currRotation--;
	}
	while(currRotation < 0) {
		currRotation += 6;
	}
	while(currRotation >= 6) {
		currRotation -= 6;
	}
	activeSymbol = baseSymbol + currRotation;
	updateSymbolPreview();
}

function flipSymbol() {
	var currRotation = activeSymbol % 10;
	var baseSymbol = activeSymbol - currRotation;
	currRotation = currRotation + 3;
	while(currRotation < 0) {
		currRotation += 6;
	}
	while(currRotation >= 6) {
		currRotation -= 6;
	}
	activeSymbol = baseSymbol + currRotation;
	updateSymbolPreview();
}

function rotateDecor(clockwise) {
	var baseDecor = getBaseGlyph(activeDecor);
	if(typeof clockwise != "undefined") {} else { clockwise = true; }
	var currRotation = activeDecor % 10;
	if(clockwise) {
		currRotation++;
	}
	else {
		currRotation--;
	}
	while(currRotation < 0) {
		currRotation += 6;
	}
	while(currRotation >= 6) {
		currRotation -= 6;
	}
	activeDecor = baseDecor + currRotation;
	updateDecorPreview();
}

function flipDecor() {
	var currRotation = activeDecor % 10;
	var baseDecor = activeDecor - currRotation;
	currRotation = currRotation + 3;
	while(currRotation < 0) {
		currRotation += 6;
	}
	while(currRotation >= 6) {
		currRotation -= 6;
	}
	activeDecor = baseDecor + currRotation;
	updateDecorPreview();
}

function setHex(col, row) {
	if(activeLayer == LAYER_MEGAHEX) { return; }
	if(col >= 0 && col < hexarray.length && row >= 0  && row < hexarray[0].length) {
		if(activeLayer == LAYER_ALL || activeLayer == LAYER_BASE) {
			if(activeComponent == HEX_EMPTY || activeMaterial == MATERIAL_EMPTY) {
				hexarray[col][row].clearBases();
			}
			else {
				hexarray[col][row].addBase(activeComponent, activeMaterial);
			}
		}
		if(activeLayer == LAYER_ALL || activeLayer == LAYER_OVERLAY) {
			if(activeComponent == HEX_EMPTY || activeMaterial == MATERIAL_EMPTY) {
				hexarray[col][row].clearOverlays();
			}
			else {
				hexarray[col][row].addOverlay(activeComponent, activeMaterial);
			}
		}
		if(activeLayer == LAYER_ALL || activeLayer == LAYER_FEATURE) {
			if(activeFeature == FEATURE_EMPTY || activeMaterial == MATERIAL_EMPTY) {
				hexarray[col][row].clearFeatures();
			}
			else {
				hexarray[col][row].addFeature(activeFeature, activeMaterial);
			}
		}
		if(activeLayer == LAYER_ALL || activeLayer == LAYER_TERRAIN) {
			if(activeTerrain == TERRAIN_EMPTY || activeMaterial == MATERIAL_EMPTY) {
				hexarray[col][row].clearFeatures();
			}
			else {
				hexarray[col][row].addFeature(activeTerrain, activeMaterial);
			}
		}
		if(activeLayer == LAYER_ALL || activeLayer == LAYER_SYMBOLS) {
			if(activeSymbol == SYMBOL_EMPTY || activeMaterial == MATERIAL_EMPTY) {
				hexarray[col][row].clearFeatures();
			}
			else {
				hexarray[col][row].addFeature(activeSymbol, activeMaterial);
			}
		}
		if(activeLayer == LAYER_ALL || activeLayer == LAYER_DECOR) {
			if(activeDecor == DECOR_EMPTY || activeMaterial == MATERIAL_EMPTY) {
				hexarray[col][row].clearFeatures();
			}
			else {
				hexarray[col][row].addFeature(activeDecor, activeMaterial);
			}
		}
		if(activeLayer == LAYER_ALL || activeLayer == LAYER_ANNOTATE) {
			if(activeAnnotation == ANNOTATION_BASE || activeMaterial == MATERIAL_EMPTY) {
				hexarray[col][row].clearAnnotations();
			}
			else {
				hexarray[col][row].addAnnotation(activeAnnotation, activeMaterial);
			}
		}
		plotHex(buf, col, row, false);
	}
}

function assignHexComponent(col, row, component, rotation) {
	var currLayer = activeLayer;
	var currComponent = activeComponent;
	activeLayer = LAYER_BASE;
	if(component > HEX_WHOLE) {
		activeComponent = component + rotation;
	}
	else {
		activeComponent = component;
	}
	setHex(col, row);
	activeLayer = currLayer;
	activeComponent = currComponent;
}

function assignHexFeature(col, row, feature, rotation) {
	var currLayer = activeLayer;
	var currFeature = activeFeature;
	activeLayer = LAYER_FEATURE;
	activeFeature = feature + rotation;
	setHex(col, row);
	activeLayer = currLayer;
	activeFeature = currFeature;
}

function processMouseMove(evt) {
	if(activeLayer == LAYER_MEGAHEX) { return; }
	var mh = getMouseHex(evt.clientX, evt.clientY);
	var col = mh.x;
	var row = mh.y;
	if(col >= 0 && col < hexarray.length && row >= 0  && row < hexarray[0].length) {
		if(col != lastHexCol || row != lastHexRow) {
			lastHexCol = col;
			lastHexRow = row;
			document.getElementById('mousetracker').innerHTML = '<span>' + col + ' / ' + row + '</span>';
			if(paintOn) {
				setHex(col, row);
			}
			doubleBuffer();
			plotTool(cnv, col, row);
		}
	}
}

function togglePaintMode(evt, toggleval) {
	paintOn = toggleval;
	if(paintOn) {
		processMousePaint(evt);
	}
}

function processMousePaint(evt) {
	if(activeLayer == LAYER_MEGAHEX) { return; }
	var mh = getMouseHex(evt.clientX, evt.clientY);
	var col = mh.x;
	var row = mh.y;
	if(col >= 0 && col < hexarray.length && row >= 0  && row < hexarray[0].length) {
		if(activeMode != MODE_PAINT) {
			if(modeStartX < 0 || modeStartY < 0) {
				paintOn = false;
				modeStartX = col;
				modeStartY = row;
			}
			else {
				storeUndo();
				paintOn = true;
				if(activeMode == MODE_LINE) {
					drawLine(modeStartX, modeStartY, col, row);
				}
				else if(activeMode == MODE_RECT) {
					drawRect(modeStartX, modeStartY, col, row);
				}
				else if(activeMode == MODE_AREA) {
					fillRect(modeStartX, modeStartY, col, row);
				}
				else if(activeMode == MODE_SKIP) {
					if(mapGrain == GRAIN_VERTICAL) {
						drawSkipLine(modeStartX, modeStartY, col, modeStartY);
					}
					else {
						drawSkipLine(modeStartX, modeStartY, modeStartX, row);
					}
				}
				else if(activeMode == MODE_LROOM || activeMode == MODE_SROOM) {
					if(Math.abs(modeStartX - col) % 2 > 0) {
						if(col < modeStartX) {
							if(col > 0) {
								col = col - 1;
							}
							else {
								col = col + 1;
							}
						}
						else {
							if(col < (gridx - 1)) {
								col = col + 1;
							}
							else {
								col = col - 1;
							}
						}
					}
					if((Math.abs(modeStartX - col) > 0) && (Math.abs(modeStartY - row) > 0)) {
						drawRoom(modeStartX, modeStartY, col, row, (activeMode == MODE_SROOM));
					}
				}
				modeStartX = -1;
				modeStartY = -1;
				doubleBuffer();
			}
		}
		else {
			storeUndo();
			setHex(col, row);
			doubleBuffer();
		}
	}
}

function storeUndo() {
	hexundo.length = 0;
	for(var x = 0; x < hexarray.length; x++) {
		hexundo[x] = [];
		for(var y = 0; y < hexarray[x].length; y++) {
			hexundo[x].push(hexarray[x][y].clone());
		}
	}
	haveUndo = true;
}

function recallUndo() {
	hexarray.length = 0;
	for(var x = 0; x < hexundo.length; x++) {
		hexarray[x] = [];
		for(var y = 0; y < hexundo[x].length; y++) {
			hexarray[x].push(hexundo[x][y].clone());
		}
	}
	haveUndo = false;
}

function undo() {
	if(haveUndo) {
		recallUndo();
		drawMap();
	}
}

function recalculate() {
	drawscale = drawsize;
	halfscale = drawscale / 2;
	xoffset = drawscale;
	yoffset = (drawscale * CALCTHIN);
	outlineThickness = getOutlineThickness();
}

function resizeCanvas() {
	var newx = (mapGrain == GRAIN_VERTICAL
					? Math.ceil(((drawsize * 1.5) * gridx) + (leftmargin * 2) + (drawsize / 2))
					: Math.ceil(((drawsize * CALCSQRT) * gridx) + (leftmargin * 2) + (drawsize / 2) + (drawsize / 4))
				);
	var newy = (mapGrain == GRAIN_VERTICAL
					? Math.ceil(((drawsize * CALCSQRT) * gridy) + (topmargin * 2) + (drawsize / 2) + (drawsize / 4))
					: Math.ceil(((drawsize * 1.5) * gridy) + (topmargin * 2) + (drawsize / 2))
				);
	canvas.width = newx;
	canvas.height = newy;
	cnv = canvas.getContext('2d');
	buffer.width = newx;
	buffer.height = newy;
	buf = buffer.getContext('2d');
	buf.clearRect(0, 0, buffer.width, buffer.height);
	mhexes.width = newx;
	mhexes.height = newy;
	mhx = mhexes.getContext('2d');
	mhx.clearRect(0, 0, mhexes.width, mhexes.height);
	if(megahexesOn) { overlayMegahexes(); }
	doubleBuffer();
}

function zoomslide(src) {
	drawsize = parseInt(src.value);
	if(src.id == 'zoomlevel') {
		document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
		if(document.getElementById('prefZoomlevel') != null && document.getElementById('prefZoomlevel') != 'undefined') {
			document.getElementById('prefZoomlevel').value = drawsize;
			document.getElementById('showZoomlevelPref').innerHTML = drawsize + 'px';
		}
	}
	else if(src.id == 'prefZoomlevel') {
		document.getElementById('zoomlevel').value = drawsize;
		document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
		document.getElementById('showZoomlevelPref').innerHTML = drawsize + 'px';
	}
	recalculate();
	resizeCanvas();
	drawMap();
}

function zoomecho(src) {
	drawsize = parseInt(document.getElementById('zoomlevel').value);
	document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
}

function translide() {
	TRANSLUCE = parseFloat(document.getElementById('translevel').value);
	document.getElementById('showOpacity').innerHTML = TRANSLUCE;
	doubleBuffer();
}

function transecho() {
	TRANSLUCE = parseFloat(document.getElementById('translevel').value);
	document.getElementById('showOpacity').innerHTML = TRANSLUCE;
}

function setTranslucence(transval) {
	if(transval != null && transval >= 0.0 && transval <= 1.0) {
		TRANSLUCE = transval;
		document.getElementById('translevel').value = TRANSLUCE;
		document.getElementById('showOpacity').innerHTML = TRANSLUCE;
	}
}

function zoom(increase) {
	var changed = false;
	if(increase && drawsize < ZOOM_MAX) {
		drawsize += ZOOM_STEP;
		changed = true;
	}
	if(!increase && drawsize > ZOOM_MIN) {
		drawsize -= ZOOM_STEP;
		changed = true;
	}
	if(changed) {
		document.getElementById('zoomlevel').value = drawsize;
		document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
		recalculate();
		resizeCanvas();
		drawMap();
	}
}

function updateOutlineColor(src) {
	if(src != null && src.value != "undefined") {
		var greyscale = parseInt(src.value);
		if(src.id == 'hexshade') {
			document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((greyscale / 255) * 100)) + '%';
			if(document.getElementById('showShadeLevel') != null && document.getElementById('showShadeLevel') != "undefined") {
				document.getElementById('prefHexshade').value = greyscale;
				document.getElementById('showShadeLevelPref').innerHTML = (100 - Math.round((greyscale / 255) * 100)) + '%';
			}
		}
		else if(src.id == 'prefHexshade') {
			document.getElementById('showShadeLevelPref').innerHTML = (100 - Math.round((greyscale / 255) * 100)) + '%';
			document.getElementById('hexshade').value = greyscale;
			document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((greyscale / 255) * 100)) + '%';
		}
		defaultGridGrey = greyscale;
		defaultGridColor = "rgb(" + defaultGridGrey + "," + defaultGridGrey + "," + defaultGridGrey + ")";
		hexOutline = defaultGridColor;
		updatePreview();
		updateFeaturePreview();
		updateTerrainPreview();
		updateDecorPreview();
		updateSymbolPreview();
		drawMap();
	}
}

function outlineColorEcho() {
	var greyscale = parseInt(document.getElementById('hexshade').value);
	document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((greyscale / 255) * 100)) + '%';
}

function mhexWeightEcho() {
	document.getElementById('showMhexThick').innerHTML = megahexWeight + 'px';
}

function mhexslide(src) {
	megahexWeight = parseInt(src.value);
	mhexWeightEcho();
	overlayMegahexes();
}

function mhexWeightUpdate(refresh) {
	var mhsliderelem = document.getElementById("mhexthick");
	mhsliderelem.value = megahexWeight;
	mhexWeightEcho();
	if(refresh) { overlayMegahexes(); }
}

function mdotWeightEcho() {
	document.getElementById('showDotWeight').innerHTML = megahexDotWeight + 'px';
}

function mdotslide(src) {
	megahexDotWeight = parseInt(src.value);
	mdotWeightEcho();
	overlayMegahexes();
}

function mdotWeightUpdate(refresh) {
	var mdsliderelem = document.getElementById("mhexdotwt");
	mdsliderelem.value = megahexDotWeight;
	mdotWeightEcho();
	if(refresh) { overlayMegahexes(); }
}

function getSlope(startX, startY, endX, endY) {
	var slope = ((endY - startY) * 1.0) / ((endX - startX) * 1.0);
	if(slope < 0.0) { slope = slope * -1.0; }
	return slope;
}

function drawLine(startX, startY, endX, endY) {
	var slope = getSlope(startX, startY, endX, endY);
	if(mapGrain == GRAIN_VERTICAL) {
		if(slope <= 0.25) { endY = startY; }
		else if(slope >= 0.75) { endX = startX; }
	}
	else {
		if(slope <= 1.0) { endY = startY; }
		else if(slope >= 3.0) { endX = startX; }
	}
	var linelen = Math.max(Math.abs(startX - endX), Math.abs(startY - endY)) + 1;
	var linedir = 0;
	var lerror  = 0.0;
	if(linelen == 1) {
		setHex(startX, startY);
	}
	else if(startX == endX) {
		// vertical line
		linedir = (startY > endY ? -1 : 1);
		for(var i = startY; i != endY; i += linedir) {
			setHex(endX, i);
		}
		setHex(endX, endY);
	}
	else if(startY == endY) {
		// horizontal line
		linedir = (startX > endX ? -1 : 1);
		for(var i = startX; i != endX; i += linedir) {
			setHex(i, endY);
		}
		setHex(endX, endY);
	}
	else {
		// perfect diagonal
		if(mapGrain == GRAIN_VERTICAL) {
			endY = startY + (Math.round(Math.abs(startX - endX) * 0.5) * (endY < startY ? -1 : 1)) + (endY < startY ? ((startX % 2) * Math.abs((endX % 2) - 1)) : -(endX % 2) + Math.abs((startX % 2) * (endX % 2)));
		}
		else {
			endX = startX + (Math.round(Math.abs(startY - endY) * 0.5) * (endX < startX ? -1 : 1)) + (endX < startX ? ((startY % 2) * Math.abs((endY % 2) - 1)) : -(endY % 2) + Math.abs((startY % 2) * (endY % 2)));
		}
		if(startY > endY) {
			var tmpX = startX;
			var tmpY = startY;
			startX = endX;
			startY = endY;
			endX = tmpX;
			endY = tmpY;
		}
		linedir      = (startY > endY ? -1 : 1);
		var linedir2 = (startX > endX ? -1 : 1);
		var px       = startX;
		var py       = startY;
		var ullr     = (Math.sign(linedir) == Math.sign(linedir2) ? 0 : 1);
		for(var i = 0; i < linelen; i++) {
			setHex(px, py);
			if(mapGrain == GRAIN_VERTICAL) {
				px += linedir2;
				if(px % 2 == 0) {
					py += linedir;
				}
			}
			else {
				py += linedir;
				if(py % 2 == ullr) {
					px += linedir2;
				}
			}
		}
	}
}

function drawRect(upperX, upperY, lowerX, lowerY, val) {
	if(upperX > lowerX) {
		var tmp = upperX;
		upperX = lowerX;
		lowerX = tmp;
	}
	if(upperY > lowerY) {
		var tmp = upperY;
		upperY = lowerY;
		lowerY = tmp;
	}
	for(var x = upperX; x <= lowerX; x++) {
		setHex(x, upperY);
		setHex(x, lowerY);
	}
	for(var y = upperY; y <= lowerY; y++) {
		setHex(upperX, y);
		setHex(lowerX, y);
	}
}

function fillRect(upperX, upperY, lowerX, lowerY, val) {
	if(upperX > lowerX) {
		var tmp = upperX;
		upperX = lowerX;
		lowerX = tmp;
	}
	if(upperY > lowerY) {
		var tmp = upperY;
		upperY = lowerY;
		lowerY = tmp;
	}
	for(var x = upperX; x <= lowerX; x++) {
		for(var y = upperY; y <= lowerY; y++) {
			setHex(x, y);
		}
	}
}

function drawSkipLine(startX, startY, endX, endY) {
	var linelen = Math.max(Math.abs(startX - endX), Math.abs(startY - endY)) + 1;
	if(linelen == 1) {
		setHex(startX, startY);
	}
	else {
		if(mapGrain == GRAIN_VERTICAL) {
			if(startX > endX) {
				for(x = startX; x >= endX; x = x - 2) {
					setHex(x, startY);
				}
			}
			else {
				for(x = startX; x <= endX; x = x + 2) {
					setHex(x, startY);
				}
			}
		}
		else {
			if(startY > endY) {
				for(y = startY; y >= endY; y = y - 2) {
					setHex(startX, y);
				}
			}
			else {
				for(y = startY; y <= endY; y = y + 2) {
					setHex(startX, y);
				}
			}
		}
	}
}

function drawRoom(upperX, upperY, lowerX, lowerY, isSolid) {
	if(upperX > lowerX) {
		var tmp = upperX;
		upperX = lowerX;
		lowerX = tmp;
	}
	if(upperY > lowerY) {
		var tmp = upperY;
		upperY = lowerY;
		lowerY = tmp;
	}
	if(isSolid) {
		assignHexComponent(upperX, upperY, HEX_SRCNR, 2);
		assignHexComponent(lowerX, upperY, HEX_SRCNR, 3);
		assignHexComponent(upperX, lowerY, HEX_SRCNR, 0);
		assignHexComponent(lowerX, lowerY, HEX_SRCNR, 5);
	}
	else {
		assignHexComponent(upperX, upperY, HEX_LRCNR, 2);
		assignHexComponent(lowerX, upperY, HEX_LRCNR, 3);
		assignHexComponent(upperX, lowerY, HEX_LRCNR, 0);
		assignHexComponent(lowerX, lowerY, HEX_LRCNR, 5);
	}
	var borderYTop = upperY;
	var borderYBottom = lowerY - 1;
	if(upperX % 2 == 1) {
		borderYTop = upperY + 1;
		borderYBottom = lowerY;
	}
	var isBorderPhase = true;
	var rBorderComp = (isSolid ? HEX_WHOLE : HEX_BORDR);
	var rWallComp = (isSolid ? HEX_FHALF : HEX_LRWAL);
	var rSideComp = (isSolid ? HEX_SRSID : HEX_LRSID);
	for(var x = upperX + 1; x <= lowerX - 1; x++) {
		if(isBorderPhase) {
			assignHexComponent(x, borderYTop, rBorderComp, 0);
			assignHexComponent(x, borderYBottom, rBorderComp, 3);
		}
		else {
			assignHexComponent(x, upperY, rWallComp, 3);
			assignHexComponent(x, lowerY, rWallComp, 0);
		}
		isBorderPhase = !isBorderPhase;
	}
	for(var y = upperY + 1; y <= lowerY - 1; y++) {
		assignHexComponent(upperX, y, rSideComp, 0);
		assignHexComponent(lowerX, y, rSideComp, 3);
	}
	if(isSolid) {
		for(var y = upperY + 1; y <= lowerY - 1; y++) {
			for(var x = upperX + 1; x <= lowerX - 1; x++) {
				assignHexComponent(x, y, HEX_WHOLE, 0);
			}
		}
	}
}

function endModeDraw() {
	modeStartX = -1;
	modeStartY = -1;
}

function newMap() {
	if(confirm("Start new map? Any unsaved work will be lost.")) {
		initialise();
		resetBackdrop();
		document.getElementById('mapname').value = 'default';
		document.getElementById('mapauthor').value = defaultAuthorName;
	}
}

function clearMap() {
	if(confirm("Clear the map?")) {
		for(x = 0; x < hexarray.length; x++) {
			for(y = 0; y < hexarray[x].length; y++) {
				hexarray[x][y].clearAll();
			}
		}
		drawMap();
	}
}

function clearTotal() {
	if(confirm("Clear the map and backdrop?")) {
		for(x = 0; x < hexarray.length; x++) {
			for(y = 0; y < hexarray[x].length; y++) {
				hexarray[x][y].clearAll();
			}
		}
		resetBackdrop();
		drawMap();
		document.getElementById('mapname').value = 'default';
		document.getElementById('mapauthor').value = defaultAuthorName;
	}
}

function reinitialiseGrid() {
	var hexprev = hexarray.slice();
	hexarray = [];
	for(x = 0; x < gridx; x++) {
		hexarray[x] = [];
		for(y = 0; y < gridy; y++) {
			if(x < hexprev.length && y < hexprev[0].length) {
				hexarray[x][y] = hexprev[x][y];
			}
			else {
				var hex = new Hex(x, y);
				hex.addBase(HEX_EMPTY, MATERIAL_EMPTY);
				hexarray[x][y] = hex;
			}
		}
	}
	hexundo = hexarray.slice();
}

function resizeGrid(rows, cols, isExisting) {
	gridx = cols;
	gridy = rows;
	if(isExisting) {
		reinitialiseGrid();
	}
	else {
		initialiseGrid();
	}
	resizeCanvas();
}

function toggleRecalc(enable) {
	document.getElementById("btnupdategrid").disabled = !enable;
}

function recalcMap() {
	var newrows = parseInt(document.getElementById("maprows").value);
	var newcols = parseInt(document.getElementById("mapcols").value);
	if(isNaN(newrows) || isNaN(newcols) || newrows <= 0 || newcols <= 0 || newrows > MAX_GRID || newcols > MAX_GRID) {
		document.getElementById("maprows").value = gridy;
		document.getElementById("mapcols").value = gridx;
		toggleRecalc(false);
		window.alert("Row and Column values must be between 1 and " + MAX_GRID);
		return;
	}
	if((newcols >= gridx && newrows >= gridy) || confirm("Reducing the grid size will trim the map to fit.\r\nDo you wish to proceed?")) {
		resizeGrid(newrows, newcols, true);
		drawMap();
		toggleRecalc(false);
	}
}

function shiftMapElement(ocol, orow, ncol, nrow) {
	hexarray[ncol][nrow] = hexarray[ocol][orow].clone();
	hexarray[ocol][orow].clearAll();
//	var hexToMove = hexarray[ocol][orow];
//	hexarray[ocol].splice(orow, 1);
//	hexarray[ncol].splice(nrow, 0, hexToMove);
}

function shiftMap(shiftDir) {
	var xrange = hexarray.length;
	var yrange = hexarray[0].length;
	var shiftStep = 1;
	if(
		(mapGrain == GRAIN_VERTICAL && (shiftDir == SHIFTMAP_LEFT || shiftDir == SHIFTMAP_RIGHT || shiftDir == SHIFTMAP_RIGHTK))
		||
		(mapGrain == GRAIN_HORIZONTAL && (shiftDir == SHIFTMAP_UP || shiftDir == SHIFTMAP_DOWN || shiftDir == SHIFTMAP_DOWNK))
	) {
		shiftStep = 2;
	}
	if(shiftDir == SHIFTMAP_RIGHT) { xrange = xrange - shiftStep; }
	if(shiftDir == SHIFTMAP_DOWN) { yrange = yrange - shiftStep; }
	var addedHexes = false;
	if(shiftDir == SHIFTMAP_RIGHTK) {
		// add columns to the right
		gridx = hexarray.length + shiftStep;
		gridy = hexarray[0].length;
		addedHexes = true;
	}
	if(shiftDir == SHIFTMAP_DOWNK) {
		// add rows to the bottom
		gridx = hexarray.length;
		gridy = hexarray[0].length + shiftStep;
		addedHexes = true;
	}
	if(addedHexes) {
		resizeGrid(gridy, gridx, true);
		resizeCanvas();
		drawMap();
		document.getElementById("maprows").value = gridy;
		document.getElementById("mapcols").value = gridx;
	}
	if(shiftDir == SHIFTMAP_LEFT) {
		// shift map left
		for(x = shiftStep; x < xrange; x++) {
			for(y = 0; y < yrange; y++) {
				shiftMapElement(x, y, x - shiftStep, y);
			}
		}
	}
	else if(shiftDir == SHIFTMAP_RIGHT || shiftDir == SHIFTMAP_RIGHTK) {
		// shift map right
		for(x = xrange - 1; x >= 0; x--) {
			for(y = 0; y < yrange; y++) {
				shiftMapElement(x, y, x + shiftStep, y);
			}
		}
	}
	else if(shiftDir == SHIFTMAP_UP) {
		// shift map up
		for(x = 0; x < xrange; x++) {
			for(y = 1; y < yrange; y++) {
				shiftMapElement(x, y, x, y - shiftStep);
			}
		}
	}
	else if(shiftDir == SHIFTMAP_DOWN || shiftDir == SHIFTMAP_DOWNK) {
		// shift map down
		for(x = 0; x < xrange; x++) {
			for(y = yrange - 1; y >= 0; y--) {
				shiftMapElement(x, y, x, y + shiftStep);
			}
		}
	}
	drawMap();
}

function changeBackground(newBackground) {
	hexBackground = newBackground;
	hexBase[0] = hexBackground;
	hexOverlay[0] = hexBackground;
	document.getElementById("canvas").style.background = newBackground;
	document.getElementById("buffer").style.background = newBackground;
}

function changeTemplate(backgroundColor, outlineColor, rows, cols, scaling) {
	if(confirm('This will erase all unsaved map changes!')) {
		hexOutline = outlineColor;
		document.getElementById('hexshade').value = hexOutline;
		changeBackground(backgroundColor);
		resizeGrid(rows, cols, false);
		document.getElementById("maprows").value = rows;
		document.getElementById("mapcols").value = cols;
		drawsize = scaling;
		document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
		recalculate();
		resizeCanvas();
		drawMap();
	}
}

function getMapName() {
	var mapname = document.getElementById('mapname').value;
	if(mapname.length < 1) { mapname = "default"; }
	return mapname;
}

function getMapAuthor() {
	var mapauthor = document.getElementById('mapauthor').value;
	if(mapauthor.length < 1) { mapauthor = "default"; }
	return mapauthor;
}

function hideOverlay() {
	document.getElementById("loadmsg").innerHTML = "&nbsp;";
	document.getElementById("loadMapSelector").style.display = "none";
	document.getElementById("imageSaveGadget").style.display = "none";
	document.getElementById("randomMapTools").style.display = "none";
	document.getElementById("randomRoomsSettings").style.display = "none";
	document.getElementById("overlay").style.display = "none";
	document.getElementById("shader").style.display = "none";
}

function toggleRoomDrawMode() {
	if(mapGrain == GRAIN_VERTICAL) {
		document.getElementById("mode6").disabled = false;
		document.getElementById("mode6").className = "palettetool";
		document.getElementById("mode7").disabled = false;
		document.getElementById("mode7").className = "palettetool";
	}
	else {
		if(activeMode == 6 || activeMode == 7) {
			setPaintMode(0);
		}
		document.getElementById("mode6").disabled = true;
		document.getElementById("mode6").className = "palettetool tooldisabled";
		document.getElementById("mode7").disabled = true;
		document.getElementById("mode7").className = "palettetool tooldisabled";
	}
}

function toggleGrain(selector) {
	mapGrain = selector.options[selector.selectedIndex].value;
	toggleRoomDrawMode();
	correctMegahexOffsetsForGrain();
	resizeCanvas();
	drawMap();
	updatePreview();
	updateFeaturePreview();
	updateTerrainPreview();
	updateDecorPreview();
	updateSymbolPreview();
}

function changeLineWeight(selector) {
	lineWeight = selector.options[selector.selectedIndex].value;
	recalculate();
	drawMap();
}

function toggleSettings() {
	if(document.getElementById("hidabletoolblock").style.display == "none") {
		document.getElementById("hidabletoolblock").style.display = "block";
		document.getElementById("imgtoolsettingstoggle").src = "img/gadget_collapse.png";
	}
	else {
		document.getElementById("hidabletoolblock").style.display = "none";
		document.getElementById("imgtoolsettingstoggle").src = "img/gadget_expand.png";
	}
}

function nextMaterPalette() {
	if(maxMaterPalette > 0) {
		document.getElementById("materpalette" + currentMaterPalette).style.display = "none";
		currentMaterPalette = (currentMaterPalette == maxMaterPalette ? 0 : currentMaterPalette + 1);
		document.getElementById("materpalette" + currentMaterPalette).style.display = "inline-block";
	}
}

function nextCompPalette() {
	if(maxCompPalette > 0) {
		document.getElementById("comppalette" + currentCompPalette + "row1").style.display = "none";
		document.getElementById("comppalette" + currentCompPalette + "row2").style.display = "none";
		currentCompPalette = (currentCompPalette == maxCompPalette ? 0 : currentCompPalette + 1);
		document.getElementById("comppalette" + currentCompPalette + "row1").style.display = "inline-block";
		document.getElementById("comppalette" + currentCompPalette + "row2").style.display = "inline-block";
	}
}

function nextFeatPalette() {
	if(maxFeatPalette > 0) {
		document.getElementById("featpalette" + currentFeatPalette + "row1").style.display = "none";
		document.getElementById("featpalette" + currentFeatPalette + "row2").style.display = "none";
		currentFeatPalette = (currentFeatPalette == maxFeatPalette ? 0 : currentFeatPalette + 1);
		document.getElementById("featpalette" + currentFeatPalette + "row1").style.display = "inline-block";
		document.getElementById("featpalette" + currentFeatPalette + "row2").style.display = "inline-block";
	}
}

function nextSymbPalette() {
	if(maxSymbPalette > 0) {
		document.getElementById("symbpalette" + currentSymbPalette + "row1").style.display = "none";
		document.getElementById("symbpalette" + currentSymbPalette + "row2").style.display = "none";
		currentSymbPalette = (currentSymbPalette == maxSymbPalette ? 0 : currentSymbPalette + 1);
		document.getElementById("symbpalette" + currentSymbPalette + "row1").style.display = "inline-block";
		document.getElementById("symbpalette" + currentSymbPalette + "row2").style.display = "inline-block";
	}
}

function encodeBackdrop(img) {
	if(img != null) {
		return img.src;
	}
	else {
		return '';
	}
}

function changeBackdropLockState(newstate) {
	backdropLocked = (newstate == true || newstate == "true");
	document.getElementById("backdroplock").className = "palettetool" + (backdropLocked ? " toolactive" : "");
}

function toggleBackdropLockState() {
	changeBackdropLockState(!backdropLocked);
}

function warnBackdropLocked() {
	window.alert("Cannot modify backdrop while locked.\nToggle the backdrop lock to edit.");
}

function saveMapData() {
	/* increase file iteration */
	fileIteration = fileIteration + 1;

	/* build savefile */
	var csvfile = CSVPREAMBLE + FILEHEADER + NEWLINE;
	csvfile += FILEKEYVERS + VERSION_NUMBER + NEWLINE;
	csvfile += FILEKEYSIZE + hexarray.length + SEP_LIST + hexarray[0].length + NEWLINE;
	var mapname = getMapName();
	csvfile += FILEKEYNAME + mapname + NEWLINE;
	var mapauthor = getMapAuthor();
	csvfile += FILEKEYAUTH + mapauthor + NEWLINE;
	csvfile += FILEKEYITER + fileIteration + NEWLINE;
	csvfile += FILEKEYGRAN + mapGrain + NEWLINE;
	csvfile += FILEKEYLWGT + lineWeight + NEWLINE;
	csvfile += FILEKEYSHAD + document.getElementById('hexshade').value + NEWLINE;
	csvfile += FILEKEYZOOM + document.getElementById('zoomlevel').value + NEWLINE;
	csvfile += FILEKEYTRNO + translucencyEnabled + NEWLINE;
	csvfile += FILEKEYTRNA + TRANSLUCE + NEWLINE;
	if(megahexesOn) {
		csvfile += FILEKEYMHEX + mhexOverlayX + SEP_LIST + mhexOverlayY + SEP_LIST + mhexOverlayS + SEP_LIST + mhexType + SEP_LIST + NEWLINE;
	}
	csvfile += FILEKEYMHWT + megahexWeight + NEWLINE;
	csvfile += FILEKEYMHDT + megahexDotWeight + NEWLINE;
	if(backdropImage != null) {
		csvfile += BDROPHDR + NEWLINE;
		csvfile += BDROPDIMNW + backdropImage.naturalWidth + NEWLINE;
		csvfile += BDROPDIMNH + backdropImage.naturalHeight + NEWLINE;
		csvfile += BDROPDIMRW + backdropImage.width + NEWLINE;
		csvfile += BDROPDIMRH + backdropImage.height + NEWLINE;
		csvfile += BDROPOFFX + backdropOffset.x + NEWLINE;
		csvfile += BDROPOFFY + backdropOffset.y + NEWLINE;
		csvfile += BDROPDATA + encodeBackdrop(backdropImage) + NEWLINE;
	}
	csvfile += DATAHEADER + NEWLINE;
	for(var x = 0; x < hexarray.length; x++) {
		for(var y = 0; y < hexarray[x].length; y++) {
			var row = DATAHEX + x + SEP_DATA + y;
			var hex = hexarray[x][y];
			row += ",[";
			for(var b = 0; b < hex.bases.length; b++) {
				var bcomp = hex.bases[b];
				row += (b > 0 ? SEP_LIST : "") + bcomp.ctype + SEP_PAIR + bcomp.material;
			}
			row += "]";
			row += SEP_DATA;
			row += "{";
			for(var o = 0; o < hex.overlays.length; o++) {
				var ocomp = hex.overlays[o];
				row += (o > 0 ? SEP_LIST : "") + ocomp.ctype + SEP_PAIR + ocomp.material;
			}
			row += "}";
			row += SEP_DATA;
			row += "(";
			var elems = 0;
			for(var f = 0; f < hex.features.length; f++) {
				var feat = hex.features[f];
				row += (elems > 0 ? SEP_LIST : "") + feat.ctype + SEP_PAIR + feat.material;
				elems += 1;
			}
			for(var a = 0; a < hex.annotations.length; a++) {
				var anno = hex.annotations[a];
				row += (elems > 0 ? SEP_LIST : "") + anno.ctype + SEP_PAIR + anno.material;
				elems += 1;
			}
			row += ")";
			csvfile += row + NEWLINE;
		}
	}

	/* get URI for file */
	var csvUrl = encodeURI(csvfile);

	/* pseudolink for named download */
	var pseudoanchor = document.createElement("a");
	pseudoanchor.href = csvUrl;
	pseudoanchor.download = mapname + MAPFILE_EXTENSION;
	pseudoanchor.style.display = "none";
	document.body.appendChild(pseudoanchor);
	pseudoanchor.click();
	document.body.removeChild(pseudoanchor);
}

function loadMapData(cmp) {
	var loadFile = cmp.files[0];
	var lastPeriod = loadFile.name.lastIndexOf(".");
	var extn = (lastPeriod < 0 ? "null" : loadFile.name.substr(lastPeriod).toLowerCase());
	if(MAPFILE_EXTENSION != extn) {
		document.getElementById("loadmsg").innerHTML = "<span class='warning'><b>" + loadFile.name + "</b> does not have <b>" + MAPFILE_EXTENSION + "</b> extension</span>";
	}
	else {
		document.getElementById("loadmsg").innerHTML = "<span class='success'>Loading map <b>" + loadFile.name + "</b> ...</span>";
		backdropImage = null;
		var filereader = new FileReader();
		filereader.onload = function(lf) {
			parseMapData(lf.target.result.trim());
		}
		filereader.readAsText(loadFile);
		hideLoadOverlay();
		drawMap();
	}
}

function parseMapData(raw) {
	var lines = raw.split(NEWLINE);
	if(lines.length < 4) {
		document.getElementById("loadmsg").innerHTML = "<span class='warning'>Not a proper map file</span>";
		return;
	}
	var maprows = 0;
	var mapcols = 0;
	loadedMapName = "default";
	loadedAuthorName = "default";
	fileIteration = 1;
	mapGrain = GRAIN_VERTICAL;
	lineWeight = LINEWEIGHT_PROPORTIONAL;
	megahexWeight = MEGAHEX_WEIGHT_DEFAULT;
	megahexDotWeight = MEGADOT_WEIGHT_DEFAULT;
	megahexesOn = false;
	resetBackdrop();
	// process the map header
	var linenum = 0;
	var procline = '';
	while(lines[linenum] != DATAHEADER) {
		procline = lines[linenum].trim();
		if(procline == FILEHEADER) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Shadekeep Map</b> header</span>";
		}
		else if(procline.startsWith(FILEKEYVERS)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Shamat Version</b> data</span>";
			var fileVersionNumber = procline.substr(FILEKEYVERS.length);
			// TO DO: alert user if file version is newer than Shamat version
		}
		else if(procline.startsWith(FILEKEYSIZE)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Size</b> data</span>";
			var sizePair = procline.substr(FILEKEYSIZE.length);
			mapcols = parseInt(sizePair.substr(0, sizePair.indexOf(SEP_LIST)));
			maprows = parseInt(sizePair.substr(sizePair.indexOf(SEP_LIST) + 1));
			var gridcount = maprows * mapcols;
		}
		else if(procline.startsWith(FILEKEYNAME)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Name</b> data</span>";
			loadedMapName = procline.substr(FILEKEYNAME.length);
		}
		else if(procline.startsWith(FILEKEYAUTH)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Author</b> data</span>";
			loadedAuthorName = procline.substr(FILEKEYAUTH.length);
		}
		else if(procline.startsWith(FILEKEYITER)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Iteration</b> data</span>";
			try {
				fileIteration = parseInt(procline.substr(FILEKEYITER.length));
			}
			catch(err) {
				fileIteration = 1;
			}
		}
		else if(procline.startsWith(FILEKEYGRAN)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Map Grain</b> data</span>";
			var mapGrainVal = procline.substr(FILEKEYGRAN.length);
			mapGrain = parseInt(mapGrainVal);
			document.getElementById("mapgrain").selectedIndex = mapGrain;
		}
		else if(procline.startsWith(FILEKEYLWGT)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Line Weight</b> data</span>";
			var lineWeightVal = procline.substr(FILEKEYLWGT.length);
			lineWeight = parseInt(lineWeightVal);
			document.getElementById("lineweight").selectedIndex = lineWeight;
		}
		else if(procline.startsWith(FILEKEYSHAD)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Outline</b> data</span>";
			var greyscale = parseInt(procline.substr(FILEKEYSHAD.length));
			if(isNaN(greyscale)) {}
			else {
				document.getElementById('hexshade').value = greyscale;
				document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((greyscale / 255) * 100)) + '%';
				hexOutline = "rgb(" + greyscale + "," + greyscale + "," + greyscale + ")";
			}
		}
		else if(procline.startsWith(FILEKEYZOOM)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Zoom</b> data</span>";
			var zoomload = parseInt(procline.substr(FILEKEYZOOM.length));
			if(isNaN(zoomload)) {}
			else {
				drawsize = zoomload;
				document.getElementById('zoomlevel').value = drawsize;
				document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
				recalculate();
				resizeCanvas();
			}
		}
		else if(procline.startsWith(FILEKEYTRNO)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Translucent On</b> data</span>";
			var transload = procline.substr(FILEKEYTRNO.length);
			setTranslucency((transload == true || transload == 'true'));
		}
		else if(procline.startsWith(FILEKEYTRNA)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Translucent Amount</b> data</span>";
			var transamt = parseFloat(procline.substr(FILEKEYTRNA.length));
			if(isNaN(transamt)) {} else { setTranslucence(transamt); }
		}
		else if(procline.startsWith(FILEKEYMHEX)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Megahex</b> Definitions</span>";
			var mhargs = procline.substr(FILEKEYMHEX.length).split(SEP_LIST);
			mhexOverlayX = parseInt(mhargs[0]);
			mhexOverlayY = parseInt(mhargs[1]);
			mhexOverlayS = parseInt(mhargs[2]);
			mhexType = (typeof mhargs[3] === 'undefined' ? MEGAHEX_TYPE_STANDARD : parseInt(mhargs[3]));
			if(isNaN(mhexType)) { mhexType = MEGAHEX_TYPE_STANDARD; }
			megahexesOn = true;
		}
		else if(procline.startsWith(FILEKEYMHWT)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Megahex Line Thickness</b></span>";
			var mhexwght = parseInt(procline.substr(FILEKEYMHWT.length));
			if(isNaN(mhexwght)) {} else {
				megahexWeight = mhexwght;
				mhexWeightUpdate(false);
			}
		}
		else if(procline.startsWith(FILEKEYMHDT)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Megahex Dot Weight</b></span>";
			var mdotwght = parseInt(procline.substr(FILEKEYMHDT.length));
			if(isNaN(mdotwght)) {} else {
				megahexDotWeight = mdotwght;
				mdotWeightUpdate(false);
			}
		}
		else if(procline.startsWith(BDROPHDR)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Backdrop</b> header</span>";
			backdropImage = new Image();
		}
		else if(procline.startsWith(BDROPDIMNW)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Backdrop</b> natural width</span>";
			var valload = parseInt(procline.substr(BDROPDIMNW.length));
			if(isNaN(valload)) {} else { backdropImage.naturalWidth = valload; }
		}
		else if(procline.startsWith(BDROPDIMNH)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Backdrop</b> natural height</span>";
			var valload = parseInt(procline.substr(BDROPDIMNH.length));
			if(isNaN(valload)) {} else { backdropImage.naturalHeight = valload; }
		}
		else if(procline.startsWith(BDROPDIMRW)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Backdrop</b> width</span>";
			var valload = parseInt(procline.substr(BDROPDIMRW.length));
			if(isNaN(valload)) {} else { backdropImage.width = valload; }
		}
		else if(procline.startsWith(BDROPDIMRH)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Backdrop</b> height</span>";
			var valload = parseInt(procline.substr(BDROPDIMRH.length));
			if(isNaN(valload)) {} else { backdropImage.height = valload; }
		}
		else if(procline.startsWith(BDROPDIMRH)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Backdrop</b> height</span>";
			var valload = parseInt(procline.substr(BDROPDIMRH.length));
			if(isNaN(valload)) {} else { backdropImage.height = valload; }
		}
		else if(procline.startsWith(BDROPOFFX)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Backdrop Offset</b> X</span>";
			var valload = parseInt(procline.substr(BDROPOFFX.length));
			if(isNaN(valload)) {} else { backdropOffset.x = valload; }
		}
		else if(procline.startsWith(BDROPOFFY)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Backdrop Offset</b> Y</span>";
			var valload = parseInt(procline.substr(BDROPOFFY.length));
			if(isNaN(valload)) {} else { backdropOffset.y = valload; }
		}
		else if(procline.startsWith(BDROPDATA)) {
			document.getElementById("loadmsg").innerHTML = "<span class='success'>Obtained <b>Backdrop</b> encoded data</span>";
			var imgrawdata = procline.substr(BDROPDATA.length);
			if(imgrawdata != null && imgrawdata != '') {
				backdropImage.src = imgrawdata;
				backdropImage.naturalWidth = backdropImage.naturalWidth;
				backdropImage.naturalHeight = backdropImage.naturalHeight;
				backdropImage.width = backdropImage.width;
				backdropImage.height = backdropImage.height;
				updateBackdropWidgets();
			}
		}
		else {
			// unsupported header line
			debugMsg("UNKNOWN HEADER: " + procline);
		}
		linenum++;
	}
	// header looks good, reinitialise grid for incoming map data
	if(maprows > 0 && mapcols > 0) {
		resizeGrid(maprows, mapcols, false);
		document.getElementById("maprows").value = maprows;
		document.getElementById("mapcols").value = mapcols;
	}
	else {
		document.getElementById("loadmsg").innerHTML = "<span class='warning'>Map has zero length size setting</span>";
		return;
	}
	linenum++;
	// process the map data
	var dvals = [];
	var hexX = 0;
	var hexY = 0;
	var gstr = '';
	var grns = [];
	var wstr = '';
	var wals = [];
	var fstr = '';
	var ftrs = [];
	var comp = 0;
	var matr = 0;
	for(var l = linenum; l < lines.length; l++) {
		if(lines[l].substr(0, DATAHEX.length) == DATAHEX) {
			// hex data
			dvals = lines[l].substr(DATAHEX.length).split(SEP_DATA);
			hexX = parseInt(dvals[0]);
			hexY = parseInt(dvals[1]);
			var hex = new Hex(hexX, hexY);
			gstr = dvals[2];
			wstr = dvals[3];
			fstr = (dvals.length < 5 ? '' : dvals[4]);
			grns = [];
			wals = [];
			ftrs = [];
			if(gstr.length > 2) {
				gstr = gstr.substr(1, gstr.length - 2);
				grns = gstr.split(SEP_LIST);
				for(var g = 0; g < grns.length; g++) {
					comp = parseInt(grns[g].substr(0, grns[g].indexOf(SEP_PAIR)));
					matr = parseInt(grns[g].substr(grns[g].indexOf(SEP_PAIR) + 1));
					hex.addBase(comp, matr);
				}
			}
			if(wstr.length > 2) {
				wals = wstr.substr(1, wstr.length - 2).split(SEP_LIST);
				for(var w = 0; w < wals.length; w++) {
					comp = parseInt(wals[w].substr(0, wals[w].indexOf(SEP_PAIR)));
					matr = parseInt(wals[w].substr(wals[w].indexOf(SEP_PAIR) + 1));
					hex.addOverlay(comp, matr);
				}
			}
			if(fstr.length > 2) {
				ftrs = fstr.substr(1, fstr.length - 2).split(SEP_LIST);
				for(var f = 0; f < ftrs.length; f++) {
					feat = parseInt(ftrs[f].substr(0, ftrs[f].indexOf(SEP_PAIR)));
					matr = parseInt(ftrs[f].substr(ftrs[f].indexOf(SEP_PAIR) + 1));
					if(feat >= ANNOTATION_BASE) {
						hex.addAnnotation(feat, matr);
					}
					else {
						hex.addFeature(feat, matr);
					}
				}
			}
			hexarray[hexX][hexY] = hex;
		}
	}
	hideLoadOverlay();
	toggleMegahexes(megahexesOn);
	if(backdropImage != null) {
		backdropImage.addEventListener('load', function() {
			backdropShowing = true;
			document.getElementById("backdroptoggle").className = "palettetool" + (backdropShowing ? " toolactive" : "");
			drawMap();
		}, false);
	}
	drawMap();
	document.getElementById('mapname').value = loadedMapName;
	document.getElementById('mapauthor').value = loadedAuthorName;
}

function displayLoadSelector() {
	document.getElementById("overlay").className = "fileloaddiv";
	document.getElementById("shader").style.display = "block";
	document.getElementById("overlay").style.display = "block";
	document.getElementById("loadMapSelector").style.display = "block";
}

function displayBackdropSelector() {
	if(backdropLocked) {
		warnBackdropLocked();
	}
	else {
		document.getElementById("overlay").className = "fileloaddiv";
		document.getElementById("shader").style.display = "block";
		document.getElementById("overlay").style.display = "block";
		document.getElementById("loadBackdropSelector").style.display = "block";
	}
}

function hideLoadOverlay() {
	document.getElementById("loadmsg").innerHTML = "&nbsp;";
	document.getElementById("loadimgmsg").innerHTML = "&nbsp;";
	document.getElementById("loadMapSelector").style.display = "none";
	document.getElementById("loadBackdropSelector").style.display = "none";
	document.getElementById("overlay").style.display = "none";
	document.getElementById("shader").style.display = "none";
}

function saveImage() {
	var currMapSize = drawsize;
	if(drawsize != parseInt(document.getElementById("imagesavesize").value)) {
		drawsize = parseInt(document.getElementById("imagesavesize").value);
		recalculate();
		resizeCanvas();
		drawMap();
	}
	var mapname = getMapName();
	var png = canvas.toDataURL("image/png").replace("image/png", "image/octet-stream");
	var pseudoanchor = document.createElement("a");
	pseudoanchor.href = png;
	pseudoanchor.download = mapname + ".png";
	pseudoanchor.style.display = "none";
	document.body.appendChild(pseudoanchor);
	pseudoanchor.click();
	document.body.removeChild(pseudoanchor);
	if(drawsize != currMapSize) {
		drawsize = currMapSize;
		recalculate();
		resizeCanvas();
		drawMap();
	}
	hideSaveOverlay();
}

function imageSaveSizeSet() {
	var newsavesize = document.getElementById("imagesavesize").value;
	var savesizex = Math.ceil(((newsavesize * 1.5) * gridx) + (leftmargin * 2) + (newsavesize / 2));
	var savesizey = Math.ceil(((newsavesize * CALCSQRT) * gridy) + (topmargin * 2) + (newsavesize / 2) + (newsavesize / 4));
	document.getElementById("imgsavemsg").innerHTML = "<span class='sizemsg'>" + savesizex + " x " + savesizey + " pixels</span>";
}

function displaySaveSelector() {
	document.getElementById("imagesavesize").value = drawsize;
	imageSaveSizeSet();
	document.getElementById("overlay").className = "imagesavediv";
	document.getElementById("shader").style.display = "block";
	document.getElementById("overlay").style.display = "block";
	document.getElementById("imageSaveGadget").style.display = "block";
}

function hideSaveOverlay() {
	document.getElementById("imgsavemsg").innerHTML = "&nbsp;";
	document.getElementById("imageSaveGadget").style.display = "none";
	document.getElementById("overlay").style.display = "none";
	document.getElementById("shader").style.display = "none";
}

function updateBackdropWidgets() {
	if(backdropImage != null) {
		document.getElementById('showBackdropWidth').value = backdropImage.width;
		document.getElementById('showBackdropHeight').value = backdropImage.height;
		document.getElementById('showBackdropOffsetX').value = backdropOffset.x;
		document.getElementById('showBackdropOffsetY').value = backdropOffset.y;
	}
	else {
		document.getElementById('showBackdropWidth').value = 0;
		document.getElementById('showBackdropHeight').value = 0;
		document.getElementById('showBackdropOffsetX').value = 0;
		document.getElementById('showBackdropOffsetY').value = 0;
	}
}

function loadBackdropData(cmp) {
	var loadFile = cmp.files[0];
	var lastPeriod = loadFile.name.lastIndexOf(".");
	var extn = (lastPeriod < 0 ? "null" : loadFile.name.substr(lastPeriod).toLowerCase());
	if(".png" != extn && ".jpg" != extn && ".jpeg" != extn) {
		document.getElementById("loadimgmsg").innerHTML = "<span class='warning'><b>" + loadFile.name + "</b> does not have supported extension.<br/>Image must be a PNG or JPEG.</span>";
	}
	else {
		backdropImage = new Image();
		document.getElementById("loadimgmsg").innerHTML = "<span class='success'>Loading image <b>" + loadFile.name + "</b> ...</span>";
		var filereader = new FileReader();
		filereader.onload = function() {
			backdropImage.src = filereader.result;
		}
		filereader.readAsDataURL(loadFile);
		backdropImage.addEventListener('load', function() {
			hideOverlay();
			backdropShowing = true;
			document.getElementById("backdroptoggle").className = "palettetool" + (backdropShowing ? " toolactive" : "");
			setTranslucency(true);
			backdropToCorner();
			updateBackdropWidgets();
			drawMap();
		}, false);
		hideLoadOverlay();
	}
}

function toggleBackdrop() {
	backdropShowing = !backdropShowing;
	document.getElementById("backdroptoggle").className = "palettetool" + (backdropShowing ? " toolactive" : "");
	if(backdropShowing != translucencyEnabled) {
		toggleTranslucency();
	}
	drawMap();
}

function resetBackdrop() {
	backdropImage = null;
	changeBackdropLockState(false);
	updateBackdropWidgets();
	setTranslucency(false);
}

function removeBackdrop() {
	if(backdropLocked) {
		warnBackdropLocked();
	}
	else {
		resetBackdrop();
		drawMap();
	}
}

function backdropFitWidth() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropImage.width = canvas.width;
			backdropImage.height = backdropImage.naturalHeight * (canvas.width / backdropImage.naturalWidth);
			backdropOffset.x = 0;
			backdropOffset.y = 0;
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function backdropFitHeight() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropImage.width = backdropImage.naturalWidth * (canvas.height / backdropImage.naturalHeight);
			backdropImage.height = canvas.height;
			backdropOffset.x = 0;
			backdropOffset.y = 0;
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function backdropFitBoth() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropImage.width = canvas.width;
			backdropImage.height = canvas.height;
			backdropOffset.x = 0;
			backdropOffset.y = 0;
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function backdropFitActual() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropImage.width = backdropImage.naturalWidth;
			backdropImage.height = backdropImage.naturalHeight;
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function resizeBackdropWidth(amt) {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			var newWidth = backdropImage.width + amt;
			if(newWidth > 0 && newWidth < 10000) {
				backdropImage.width = newWidth;
				updateBackdropWidgets();
				drawMap();
			}
		}
	}
}

function resizeBackdropHeight(amt) {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			var newHeight = backdropImage.height + amt;
			if(newHeight > 0 && newHeight < 10000) {
				backdropImage.height = newHeight;
				updateBackdropWidgets();
				drawMap();
			}
		}
	}
}

function shiftBackdropHorizontal(amt) {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			var newOffsetX = backdropOffset.x + amt;
			if(newOffsetX > -(backdropImage.width + 1) && newOffsetX < canvas.width) {
				backdropOffset.x = newOffsetX;
				updateBackdropWidgets();
				drawMap();
			}
		}
	}
}

function shiftBackdropVertical(amt) {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			var newOffsetY = backdropOffset.y + amt;
			if(newOffsetY > -(backdropImage.height + 1) && newOffsetY < canvas.height) {
				backdropOffset.y = newOffsetY;
				updateBackdropWidgets();
				drawMap();
			}
		}
	}
}

function backdropToCorner() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropOffset.x = 0;
			backdropOffset.y = 0;
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function backdropToCenter() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropOffset.x = Math.round((canvas.width - backdropImage.width) / 2);
			backdropOffset.y = Math.round((canvas.height - backdropImage.height) / 2);
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function setTranslucency(enableTrans) {
	translucencyEnabled = enableTrans;
	document.getElementById("translucenttoggle").className = "palettetool" + (translucencyEnabled ? " toolactive" : "");
}

function toggleTranslucency() {
	setTranslucency(!translucencyEnabled);
	drawMap();
}

function writeVersionNumber() {
	document.write('(v' + VERSION_NUMBER + ')');
}

var activeCustomColor = 3;
var activeCustomLayer  = LAYER_BASE;

function parseColor(clr) {
	var clrR = 128;
	var clrG = 128;
	var clrB = 128;
	if(clr.startsWith('#')) {
		clrR = parseInt(clr.substr(1,2), 16);
		clrG = parseInt(clr.substr(3,2), 16);
		clrB = parseInt(clr.substr(5,2), 16);
	}
	var colorParse = [];
	colorParse[CLR_R] = clrR;
	colorParse[CLR_G] = clrG;
	colorParse[CLR_B] = clrB;
	return colorParse;
}

function encodeColor(clrArray) {
	var rval = clrArray[CLR_R].toString(16);
	var gval = clrArray[CLR_G].toString(16);
	var bval = clrArray[CLR_B].toString(16);
	return ('#' + (rval.length < 2 ? '0' + rval : rval) + (gval.length < 2 ? '0' + gval : gval) + (bval.length < 2 ? '0' + bval : bval));
}

function setCustomMaterial(clrId) {
	document.getElementById('customMaterial' + activeCustomColor).className = 'palettetool';
	activeCustomColor = clrId;
	document.getElementById('customMaterial' + activeCustomColor).className = 'palettetool toolactive';
	setCustomColorForm();
	updateCustomColorPreview(LAYER_ALL);
}

function setAnnotation(annot) {
	document.getElementById('annotate_' + activeAnnotation).className = 'annotatetool';
	activeAnnotation = annot;
	document.getElementById('annotate_' + activeAnnotation).className = 'annotatetool toolactive';
}

function setCustomColorForm() {
	document.getElementById('materialname').value = materialNames[activeCustomColor];
	var pclr = parseColor((activeCustomLayer == LAYER_OVERLAY ? hexOverlay[activeCustomColor] : hexBase[activeCustomColor]));
	document.getElementById('customRed').value = pclr[CLR_R];
	document.getElementById('showRedLevel').innerHTML = pclr[CLR_R];
	document.getElementById('customGreen').value = pclr[CLR_G];
	document.getElementById('showGreenLevel').innerHTML = pclr[CLR_G];
	document.getElementById('customBlue').value = pclr[CLR_B];
	document.getElementById('showBlueLevel').innerHTML = pclr[CLR_B];
	document.getElementById('customprevb').className = (activeCustomLayer == LAYER_BASE    ? 'palettetool customactive' : 'palettetool custominactive');
	document.getElementById('customprevo').className = (activeCustomLayer == LAYER_OVERLAY ? 'palettetool customactive' : 'palettetool custominactive');
}

function updateCustomColorPreview(lyr) {
	if(lyr == LAYER_BASE || lyr == LAYER_ALL) {
		var customb = document.getElementById('custommaterialb');
		var ctxb = customb.getContext('2d');
		var hexraw = getHexRawCoords(PREVIEW_SCALE, PREVIEW_SIZEX, PREVIEW_SIZEY);
		ctxb.clearRect(0, 0, PREVIEW_SIZEX * 2, PREVIEW_SIZEY * 2);
		drawComponent(ctxb, 0, 0, hexraw[0], hexraw[1], HEX_WHOLE, hexBase[activeCustomColor]);
		drawHexOutline(ctxb, hexraw[0], hexraw[1], '#000000');
	}
	if(lyr == LAYER_OVERLAY || lyr == LAYER_ALL) {
		var customo = document.getElementById('custommaterialo');
		var ctxo = customo.getContext('2d');
		var hexraw = getHexRawCoords(PREVIEW_SCALE, PREVIEW_SIZEX, PREVIEW_SIZEY);
		ctxo.clearRect(0, 0, PREVIEW_SIZEX * 2, PREVIEW_SIZEY * 2);
		drawComponent(ctxo, 0, 0, hexraw[0], hexraw[1], HEX_WHOLE, hexOverlay[activeCustomColor]);
		drawHexOutline(ctxo, hexraw[0], hexraw[1], '#000000');
	}
}

function setCustomColorLayer(lyr) {
	activeCustomLayer = lyr;
	document.getElementById('customprevb').className = (activeCustomLayer == LAYER_BASE    ? 'palettetool customactive' : 'palettetool custominactive');
	document.getElementById('customprevo').className = (activeCustomLayer == LAYER_OVERLAY ? 'palettetool customactive' : 'palettetool custominactive');
	setCustomColorForm();
}

function alterCustomColor(colorpart, slider) {
	var pclr = parseColor((activeCustomLayer == LAYER_OVERLAY ? hexOverlay[activeCustomColor] : hexBase[activeCustomColor]));
	pclr[colorpart] = parseInt(slider.value);
	document.getElementById('showRedLevel').innerHTML = pclr[CLR_R];
	document.getElementById('showGreenLevel').innerHTML = pclr[CLR_G];
	document.getElementById('showBlueLevel').innerHTML = pclr[CLR_B];
	if(activeCustomLayer == LAYER_OVERLAY) { hexOverlay[activeCustomColor] = encodeColor(pclr); }
	else { hexBase[activeCustomColor] = encodeColor(pclr); }
	updateCustomColorPreview(activeCustomLayer);
}

function goToHelp(sect) {
	var helpdiv = document.getElementById('paletteHelp');
	var helpsect = document.getElementById(sect);
	helpdiv.scrollTop = helpsect.offsetTop - helpdiv.offsetTop;
	return false;
}

function displayPrefSetter() {
	prefRestoreCreator = defaultAuthorName;
	prefRestoreGridshade = defaultGridGrey;
	document.getElementById("overlay").className = "setprefdiv";
	document.getElementById("shader").style.display = "block";
	document.getElementById("overlay").style.display = "block";
	document.getElementById("prefSetter").style.display = "block";
	document.getElementById('prefInputCreator').value = getMapAuthor();
	document.getElementById('prefInputRows').value = gridy;
	document.getElementById('prefInputCols').value = gridx;
	document.getElementById('prefMapgrain').value = mapGrain;
	document.getElementById('prefLineweight').value = lineWeight;
	document.getElementById('prefHexshade').value = defaultGridGrey;
	document.getElementById('showShadeLevelPref').innerHTML = (100 - Math.round((defaultGridGrey / 255) * 100)) + '%';
}

function cancelPrefSetter() {
	defaultAuthorName = prefRestoreCreator;
	document.getElementById('prefInputCreator').value = defaultAuthorName;
	defaultGridGrey = prefRestoreGridshade;
	document.getElementById('hexshade').value = defaultGridGrey;
	document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((defaultGridGrey / 255) * 100)) + '%';
	document.getElementById('prefHexshade').value = defaultGridGrey;
	document.getElementById('showShadeLevelPref').innerHTML = (100 - Math.round((defaultGridGrey / 255) * 100)) + '%';
	defaultGridColor = "rgb(" + defaultGridGrey + "," + defaultGridGrey + "," + defaultGridGrey + ")";
	hexOutline = defaultGridColor;
	drawMap(true);
}

function checkVersion() {
	var req = new XMLHttpRequest();
//	req.open('GET', 'blob:file%3A///D%3A/files/USB/shadekeep/shadekeep/shamat/v.html', true);
	req.onload = function() {
		debugMsg('Fetching...');
		debugMsg(req);
		debugMsg(req.response);
	}
	req.open('GET', 'http://shadekeep.com/index.html', true);
	req.send();
	window.alert("Version: " + VERSION_NUMBER);
}

function getLineIntersection(line1Start, line1End, line2Start, line2End) {
	var line1YDiff = line1End.y - line1Start.y;
	var line1XDiff = line1Start.x - line1End.x;
	var line1Slope = (line1YDiff * line1Start.x) + (line1XDiff * line1Start.y);
	var line2YDiff = line2End.y - line2Start.y;
	var line2XDiff = line2Start.x - line2End.x;
	var line2Slope = (line2YDiff * line2Start.x) + (line2XDiff * line2Start.y);
	var determin   = (line1YDiff * line2XDiff) - (line1XDiff * line2YDiff);
	if(determin == 0) {
		// parallel lines, no intersection
		return line1Start;
	}
	else {
		var interX = ((line2XDiff * line1Slope) - (line1XDiff * line2Slope)) / determin;
		var interY = ((line1YDiff * line2Slope) - (line2YDiff * line1Slope)) / determin;
		return new Point(Math.round(interX), Math.round(interY));
	}
}

function clearMapGrid() {
	for(x = 0; x < hexarray.length; x++) {
		for(y = 0; y < hexarray[x].length; y++) {
			hexarray[x][y].clearAll();
		}
	}
}

function generateRandomMap() {
	hideRandomMapOverlay();
	var maptype = "color";
	var maptypes = document.getElementsByName('rndmaptype');
	for(m = 0; m < maptypes.length; m++) {
		if(maptypes[m].checked) {
			maptype = maptypes[m].value;
		}
	}
	var colorMap = (maptype == "color");
	clearMapGrid();
	randomMapSeed(colorMap);
	while(randomMapHasBlanks()) {
		randomMapIterate();
	}
	if(document.getElementById('rndriver').checked) {
		randomMapRivers(colorMap);
	}
	if(document.getElementById('rndpath').checked) {
		randomMapPaths(colorMap);
	}
	randomMapFeatures(colorMap);
	drawMap();
}

function getRandom(upperval) {
	return Math.floor(Math.random() * upperval);
}

function randomMapSeed(colorMap) {
	var seedCount = Math.ceil((hexarray.length * hexarray[0].length) / 50);
	var seedPlant = seedCount;
	var seedValue = TERRAIN_EMPTY;
	var seedMater = 1;
	var baseValue = HEX_EMPTY;
	var baseMater = MATERIAL_EMPTY;
	var seedPool = [];
	if(document.getElementById('rndblank').checked) {
		seedPool.push('EMPTY');
		seedPool.push('EMPTY');
		seedPool.push('EMPTY');
		seedPool.push('EMPTY');
		seedPool.push('EMPTY');
	}
	if(document.getElementById('rndwater').checked) {
		seedPool.push('WATER');
	}
	if(document.getElementById('rndmountains').checked) {
		seedPool.push('MOUNTAIN');
	}
	if(document.getElementById('rndforest').checked) {
		seedPool.push('FOREST');
		seedPool.push('FOREST');
		seedPool.push('FOREST');
	}
	if(document.getElementById('rndswamp').checked) {
		seedPool.push('SWAMP');
	}
	if(document.getElementById('rndplains').checked) {
		seedPool.push('PLAINS');
		seedPool.push('PLAINS');
	}
	if(document.getElementById('rnddesert').checked) {
		seedPool.push('DESERT');
	}
	if(seedPool.length < 1) {
		seedPool.push('EMPTY');
	}
	while(seedPlant > 0) {
		var findHex = true;
		switch(seedPool[getRandom(seedPool.length)]) {
			case 'EMPTY':
				baseValue = HEX_EMPTY;
				baseMater = MATERIAL_EMPTY;
				seedValue = TERRAIN_EMPTY;
				seedMater = 0;
				break;
			case 'PLAINS':
				baseValue = HEX_WHOLE;
				baseMater = (colorMap ? 5 : MATERIAL_EMPTY);
				seedValue = TERRAIN_PLAINS;
				seedMater = (colorMap ? 15 : 9);
				break;
			case 'FOREST':
				baseValue = HEX_EMPTY;
				baseMater = (colorMap ? MATERIAL_EMPTY : MATERIAL_EMPTY);
				seedValue = TERRAIN_TREES;
				seedMater = (colorMap ? 7 : 3);
				break;
			case 'SWAMP':
				baseValue = HEX_WHOLE;
				baseMater = (colorMap ? 14 : MATERIAL_EMPTY);
				seedValue = TERRAIN_SWAMP;
				seedMater = (colorMap ? 1 : 1);
				break;
			case 'DESERT':
				baseValue = HEX_WHOLE;
				baseMater = (colorMap ? 10 : MATERIAL_EMPTY);
				seedValue = TERRAIN_DESERT;
				seedMater = (colorMap ? 4 : 9);
				break;
			case 'MOUNTAIN':
				baseValue = HEX_WHOLE;
				baseMater = (colorMap ? 3 : MATERIAL_EMPTY);
				seedValue = TERRAIN_MOUNTFIL;
				seedMater = (colorMap ? 6 : 3);
				break;
			case 'WATER':
				baseValue = HEX_EMPTY;
				baseMater = (colorMap ? MATERIAL_EMPTY : MATERIAL_EMPTY);
				seedValue = TERRAIN_WATER;
				seedMater = (colorMap ? 13 : 1);
				break;
		}
		while(findHex) {
			var randX = getRandom(hexarray.length);
			var randY = getRandom(hexarray[0].length);
			if(hexarray[randX][randY].features.length == 0) {
				hexarray[randX][randY].addBase(baseValue, baseMater);
				hexarray[randX][randY].addFeature(seedValue, seedMater);
				findHex = false;
			}
		}
		seedPlant--;
	}
	drawMap();
}

function randomMapHasBlanks() {
	for(x = 0; x < hexarray.length; x++) {
		for(y = 0; y < hexarray[x].length; y++) {
			if(hexarray[x][y].features.length == 0) {
				return true;
			}
		}
	}
	return false;
}

function randomMapIterate() {
	var nextPt;
	for(x = 0; x < hexarray.length; x++) {
		for(y = 0; y < hexarray[x].length; y++) {
			nextPt = getHexAt(x, y, getRandom(6));
			if(nextPt.x != OFFGRID && nextPt.y != OFFGRID) {
				if(hexarray[x][y].features.length > 0) {
					if(hexarray[nextPt.x][nextPt.y].features.length == 0) {
						hexarray[nextPt.x][nextPt.y].addBase(hexarray[x][y].bases[0].ctype, hexarray[x][y].bases[0].material);
						hexarray[nextPt.x][nextPt.y].addFeature(hexarray[x][y].features[0].ctype, hexarray[x][y].features[0].material);
					}
				}
				else {
					if(hexarray[nextPt.x][nextPt.y].features.length > 0) {
						hexarray[x][y].addBase(hexarray[nextPt.x][nextPt.y].bases[0].ctype, hexarray[nextPt.x][nextPt.y].bases[0].material);
						hexarray[x][y].addFeature(hexarray[nextPt.x][nextPt.y].features[0].ctype, hexarray[nextPt.x][nextPt.y].features[0].material);
					}
				}
			}
		}
	}
}

function randomMapPaths(colorMap) {
	var pathCount = Math.max((Math.ceil((hexarray.length * hexarray[0].length) / 500) + (getRandom(3) - 2)), 1);
	var pathLoop = 0;
	var pathMater = (colorMap ? 15 : 9);
	var addPath = true;
	while(pathLoop < pathCount) {
		addPath = true;
		var randX = getRandom(hexarray.length);
		var randY = getRandom(hexarray[0].length);
		var rota = getRandom(6);
		if(hexarray[randX][randY].features.length > 0) {
			if(hexarray[randX][randY].features[0].ctype == TERRAIN_WATER || hexarray[randX][randY].features[0].ctype == TERRAIN_MOUNTFIL) {
				addPath = false;
			}
		}
		if(addPath) {
			hexarray[randX][randY].addBase(HEX_CHANL + rota, pathMater);
			randomMapPathPlot(randX, randY, rota, pathMater);
			rota = rota - 3;
			if(rota < 0) { rota = rota + 6; }
			randomMapPathPlot(randX, randY, rota, pathMater);
			pathLoop++;
		}
	}
}

function randomMapPathPlot(x, y, direction, pathMater) {
	var hexPt = getHexAt(x, y, direction);
	if(hexPt.x == OFFGRID || hexPt.y == OFFGRID) {
		return;
	}
	if(hexarray[hexPt.x][hexPt.y].features.length > 0) {
		if(hexarray[hexPt.x][hexPt.y].features[0].ctype == TERRAIN_WATER || hexarray[hexPt.x][hexPt.y].features[0].ctype == TERRAIN_MOUNTFIL) {
			return;
		}
	}
	var pathAct = getRandom(8);
	switch(pathAct) {
		case 0:
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
			// Continue path
			hexarray[hexPt.x][hexPt.y].addBase(HEX_CHANL + direction, pathMater);
			break;
		case 6:
			// Turn counterclockwise
			direction = direction - 3;
			if(direction < 0) { direction = direction + 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JUNCT + direction, pathMater);
			direction = direction - 2;
			if(direction < 0) { direction += 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JUNCT + direction, pathMater);
			break;
		case 7:
			// Turn clockwise
			direction = direction - 3;
			if(direction < 0) { direction = direction + 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JUNCT + direction, pathMater);
			direction = direction + 2;
			if(direction > 5) { direction -= 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JUNCT + direction, pathMater);
			break;
	}
	randomMapPathPlot(hexPt.x, hexPt.y, direction, pathMater);
}

function randomMapRivers(colorMap) {
	var pathCount = Math.max((Math.ceil((hexarray.length * hexarray[0].length) / 750) + (getRandom(3) - 2)), 1);
	var pathLoop = 0;
	var pathMater = (colorMap ? 13 : 3);
	var addPath = true;
	while(pathLoop < pathCount) {
		addPath = true;
		var randX = getRandom(hexarray.length);
		var randY = getRandom(hexarray[0].length);
		var rota = getRandom(6);
		if(hexarray[randX][randY].features.length > 0) {
			if(hexarray[randX][randY].features[0].ctype == TERRAIN_WATER || hexarray[randX][randY].features[0].ctype == TERRAIN_MOUNTFIL) {
				addPath = false;
			}
		}
		if(addPath) {
			hexarray[randX][randY].addBase(HEX_WIDE + rota, pathMater);
			randomMapRiverPlot(randX, randY, rota, pathMater);
			rota = rota - 3;
			if(rota < 0) { rota = rota + 6; }
			randomMapRiverPlot(randX, randY, rota, pathMater);
			pathLoop++;
		}
	}
}

function randomMapRiverPlot(x, y, direction, pathMater) {
	var hexPt = getHexAt(x, y, direction);
	if(hexPt.x == OFFGRID || hexPt.y == OFFGRID) {
		return;
	}
	if(hexarray[hexPt.x][hexPt.y].features.length > 0) {
		if(hexarray[hexPt.x][hexPt.y].features[0].ctype == TERRAIN_WATER || hexarray[hexPt.x][hexPt.y].features[0].ctype == TERRAIN_MOUNTFIL) {
			return;
		}
	}
	var pathAct = getRandom(8);
	switch(pathAct) {
		case 0:
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
			// Continue path
			hexarray[hexPt.x][hexPt.y].addBase(HEX_WIDE + direction, pathMater);
			break;
		case 6:
			// Turn counterclockwise
			direction = direction - 3;
			if(direction < 0) { direction = direction + 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JOIN + direction, pathMater);
			direction = direction - 2;
			if(direction < 0) { direction += 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JOIN + direction, pathMater);
			break;
		case 7:
			// Turn clockwise
			direction = direction - 3;
			if(direction < 0) { direction = direction + 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JOIN + direction, pathMater);
			direction = direction + 2;
			if(direction > 5) { direction -= 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JOIN + direction, pathMater);
			break;
	}
	randomMapRiverPlot(hexPt.x, hexPt.y, direction, pathMater);
}

function randomMapFeatures(colorMap) {
	var randX = 0;
	var randY = 0;
	var isOkay = false;

	if(document.getElementById('rndcapitol').checked) {
		var searching = true;
		var attempts = 0;
		while(searching && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].features.length; i++) {
				if(
					hexarray[randX][randY].features[i].ctype == TERRAIN_WATER ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_MOUNTFIL ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addFeature(TERRAIN_CAPITOL, (colorMap ? 1 : 1)); // was 16
				searching = false;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndcity').checked) {
		var cityCount = Math.ceil((hexarray.length * hexarray[0].length) / 500);
		var ccounter = 0;
		var attempts = 0;
		while(ccounter < cityCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].features.length; i++) {
				if(
					hexarray[randX][randY].features[i].ctype == TERRAIN_WATER ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_MOUNTFIL ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addFeature(TERRAIN_CITY, (colorMap ? 1 : 1)); // was 13
				ccounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndvillage').checked) {
		var townCount = Math.ceil((hexarray.length * hexarray[0].length) / 250);
		var tcounter = 0;
		var attempts = 0;
		while(tcounter < townCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].features.length; i++) {
				if(
					hexarray[randX][randY].features[i].ctype == TERRAIN_WATER ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addFeature(TERRAIN_VILLAGE, (colorMap ? 1 : 1)); // was 11
				tcounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndfort').checked) {
		var townCount = Math.ceil((hexarray.length * hexarray[0].length) / 500);
		var tcounter = 0;
		var attempts = 0;
		while(tcounter < townCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].features.length; i++) {
				if(
					hexarray[randX][randY].features[i].ctype == TERRAIN_WATER ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addFeature(TERRAIN_FORT, (colorMap ? 1 : 1));
				tcounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndmine').checked) {
		var townCount = Math.ceil((hexarray.length * hexarray[0].length) / 750);
		var tcounter = 0;
		var attempts = 0;
		while(tcounter < townCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].features.length; i++) {
				if(
					hexarray[randX][randY].features[i].ctype == TERRAIN_WATER ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_SWAMP ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_TREES ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_DESERT ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addFeature(TERRAIN_MINE, (colorMap ? 1 : 1));
				tcounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndruin').checked) {
		var townCount = Math.ceil((hexarray.length * hexarray[0].length) / 500);
		var tcounter = 0;
		var attempts = 0;
		while(tcounter < townCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].features.length; i++) {
				if(
					hexarray[randX][randY].features[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addFeature(TERRAIN_RUIN, (colorMap ? 1 : 1));
				tcounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndinterest').checked) {
		var townCount = Math.ceil((hexarray.length * hexarray[0].length) / 600);
		var tcounter = 0;
		var attempts = 0;
		while(tcounter < townCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].features.length; i++) {
				if(
					hexarray[randX][randY].features[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].features[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addFeature(TERRAIN_INTEREST, (colorMap ? 1 : 1));
				tcounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}
}

function displayRandomMapSelector() {
	document.getElementById("overlay").className = "rndmaptoolsdiv";
	document.getElementById("shader").style.display = "block";
	document.getElementById("overlay").style.display = "block";
	document.getElementById("randomMapTools").style.display = "block";
}

function hideRandomMapOverlay() {
	document.getElementById("randomMapTools").style.display = "none";
	document.getElementById("overlay").style.display = "none";
	document.getElementById("shader").style.display = "none";
}

function generateRandomRoomDungeon() {
	hideRandomRoomsSettings();
	var minRooms = 5;
	var maxRooms = 12;
	var minRoomWidth = 4;
	var minRoomHeight = 4;
	var maxRoomWidth = 12;
	var maxRoomHeight = 12;
	var roomSpacing = 1;
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomCountMin.value)))  { minRooms = parseInt(document.randomroomelems.inpRoomCountMin.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomCountMax.value)))  { maxRooms = parseInt(document.randomroomelems.inpRoomCountMax.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomWidthMin.value)))  { minRoomWidth = parseInt(document.randomroomelems.inpRoomWidthMin.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomWidthMax.value)))  { maxRoomWidth = parseInt(document.randomroomelems.inpRoomWidthMax.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomHeightMin.value))) { minRoomHeight = parseInt(document.randomroomelems.inpRoomHeightMin.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomHeightMax.value))) { maxRoomHeight = parseInt(document.randomroomelems.inpRoomHeightMax.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomSpacing.value)))   { roomSpacing = parseInt(document.randomroomelems.inpRoomSpacing.value); }
	if(maxRooms < minRooms) { var tmp = maxRooms; maxRooms = minRooms; minRooms = tmp; }
	if(maxRoomWidth < minRoomWidth) { var tmp = maxRoomWidth; maxRoomWidth = minRoomWidth; minRoomWidth = tmp; }
	if(maxRoomHeight < minRoomHeight) { var tmp = maxRoomHeight; maxRoomHeight = minRoomHeight; minRoomHeight = tmp; }
	// generate random room set
	var rooms = [];
	rooms[0] = randomRoom(minRoomWidth, minRoomHeight, maxRoomWidth, maxRoomHeight);
	var numRooms = minRooms + getRandom(maxRooms - minRooms);
	var r = 1;
	var breakout = 0;
	while(r < numRooms && breakout < 500) {
		var newRoom = randomRoom(minRoomWidth, minRoomHeight, maxRoomWidth, maxRoomHeight);
		var isDistinct = true;
		for(var tr = 0; tr < rooms.length; tr++) {
			if(newRoom.overlaps(rooms[tr], roomSpacing)) {
				isDistinct = false;
			}
		}
		if(isDistinct) {
			rooms[r] = newRoom;
			r++;
			breakout = 0;
		}
		else {
			breakout++;
		}
	}
	// fill rooms on map
	clearMapGrid();
	var roomFill = activeMaterial;
	for(var rd = 0; rd < rooms.length; rd++) {
		for(var x = rooms[rd].startX; x <= rooms[rd].endX; x++) {
			for(var y = rooms[rd].startY; y <= rooms[rd].endY; y++) {
				hexarray[x][y].addBase(HEX_WHOLE, roomFill);
			}
		}
	}
	joinRooms(rooms, roomFill);
	drawMap();
}

function randomRoom(minW, minH, maxW, maxH) {
	var rWidth  = minW + getRandom(maxW - minW);
	var rHeight = minH + getRandom(maxH - minH);
	var rX = getRandom(hexarray.length - rWidth);
	var rY = getRandom(hexarray[0].length - rHeight);
	return new RoomDef(rX, rY, rWidth, rHeight);
}

function joinRooms(roomArray, mater) {
	var connections = [];
	var currentRoom = null;
	var compareRoom = null;
	var closestRoom = -1;
	var dist = Math.max(hexarray.length, hexarray[0].length);
	var thisdist = Math.max(hexarray.length, hexarray[0].length);
	var canAdd = true;
	for(var r = 0; r < roomArray.length; r++) {
		currentRoom = roomArray[r];
		closestRoom = -1;
		dist = Math.max(hexarray.length, hexarray[0].length);
		for(var r2 = 0; r2 < roomArray.length; r2++) {
			compareRoom = roomArray[r2];
			thisdist = Math.max(hexarray.length, hexarray[0].length);
			if(r == r2) {
				// same room, skip
			}
			else if((compareRoom.startX > currentRoom.endX || compareRoom.endX < currentRoom.startX) && (compareRoom.startY > currentRoom.endY || compareRoom.endY < currentRoom.startY)) {
				// not aligned
			}
			else {
				if(compareRoom.startX > currentRoom.endX) {
					thisdist = compareRoom.startX - currentRoom.endX;
				}
				else if(compareRoom.endX < currentRoom.startX) {
					thisdist = currentRoom.startX - compareRoom.endX;
				}
				else if(compareRoom.startY > currentRoom.endY) {
					thisdist = compareRoom.startY - currentRoom.endY;
				}
				else if(compareRoom.endY < currentRoom.startY) {
					thisdist = currentRoom.startY - compareRoom.endY;
				}
				if(thisdist < dist) {
					canAdd = true;
					if(typeof connections[r2] === 'undefined') {
						// not set yet, is okay
					}
					else {
						if(connections[r2] == r) {
							// already connected
							canAdd = false;
						}
					}
					if(canAdd) {
						dist = thisdist;
						closestRoom = r2;
					}
				}
			}
		}
		if(closestRoom > -1) {
		}
		connections[r] = closestRoom;
	}
	connectRooms(roomArray, connections, mater);
}

function connectRooms(roomArray, connections, mater) {
	var currRoom = null;
	var joinRoom = null;
	for(var r = 0; r < roomArray.length; r++) {
		if(connections[r] > -1) {
			currRoom = roomArray[r];
			joinRoom = roomArray[connections[r]];
			if(joinRoom.startX > currRoom.endX || joinRoom.endX < currRoom.startX) {
				// horizontal join
				var placeFrom = Math.max(currRoom.startY, joinRoom.startY);
				var placeTo = Math.min(currRoom.endY, joinRoom.endY);
				var place = getRandom(placeTo - placeFrom) + placeFrom;
				if(joinRoom.startX > currRoom.endX) {
					for(var x = currRoom.endX + 1; x <= joinRoom.startX - 1; x++) {
						hexarray[x][place].addBase(HEX_WHOLE, mater);
					}
				}
				else {
					for(var x = joinRoom.endX + 1; x <= currRoom.startX - 1; x++) {
						hexarray[x][place].addBase(HEX_WHOLE, mater);
					}
				}
			}
			else {
				// vertical join
				var placeFrom = Math.max(currRoom.startX, joinRoom.startX);
				var placeTo = Math.min(currRoom.endX, joinRoom.endX);
				var place = getRandom(placeTo - placeFrom) + placeFrom;
				if(joinRoom.startY > currRoom.endY) {
					for(var y = currRoom.endY + 1; y <= joinRoom.startY - 1; y++) {
						hexarray[place][y].addBase(HEX_WHOLE, mater);
					}
				}
				else {
					for(var y = joinRoom.endY + 1; y <= currRoom.startY - 1; y++) {
						hexarray[place][y].addBase(HEX_WHOLE, mater);
					}
				}
			}
		}
	}
}

function displayRandomRoomsSettings() {
	document.getElementById("overlay").className = "rndmaptoolsdiv";
	document.getElementById("shader").style.display = "block";
	document.getElementById("overlay").style.display = "block";
	document.getElementById("randomRoomsSettings").style.display = "block";
}

function hideRandomRoomsSettings() {
	document.getElementById("randomRoomsSettings").style.display = "none";
	document.getElementById("overlay").style.display = "none";
	document.getElementById("shader").style.display = "none";
}

function generateRandomTunnelDungeon() {
	hideRandomTunnelsSettings();
	var saturationPercent = 0.25;
	var densitySelector = document.randomtunnelelems.tunneldensity;
	for(var d = 0; d < densitySelector.length; d++) {
		if(densitySelector[d].checked) {
			saturationPercent = densitySelector[d].value;
		}
	}
	var totalHexes = hexarray.length * hexarray[0].length;
	var toFillHexes = Math.round(totalHexes * saturationPercent);
	clearMapGrid();
	var pathMater = activeMaterial;
	while(countMaterialFills(pathMater) < toFillHexes) {
		randomTunnel(pathMater);
	}
	var fillMater = (pathMater == 16 ? 12 : 16);
	joinTunnels(pathMater, fillMater);
	var fillSeed = fillFinder(fillMater);
	floodFill(hexarray, fillSeed.x, fillSeed.y, fillMater, pathMater);
	drawMap();
}

function randomTunnel(pathMater) {
	var addPath = false;
	while(!addPath) {
		addPath = true;
		var randX = getRandom(hexarray.length);
		var randY = getRandom(hexarray[0].length);
		var rota = getRandom(6);
		if(hexarray[randX][randY].bases[0].material != MATERIAL_EMPTY) {
			addPath = false;
		}
		if(addPath) {
			randomTunnelPlot(randX, randY, rota, pathMater);
		}
	}
}

function randomTunnelPlot(x, y, direction, pathMater) {
	var hexPt = getHexAt(x, y, direction);
	if(hexPt.x == OFFGRID || hexPt.y == OFFGRID) {
		return;
	}
	if(hexarray[hexPt.x][hexPt.y].bases[0].material != MATERIAL_EMPTY) {
		return;
	}
	var pathAct = getRandom(8);
	switch(pathAct) {
		case 0:
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
			// Continue path
			hexarray[hexPt.x][hexPt.y].addBase(HEX_WHOLE, pathMater);
			break;
		case 6:
			// Turn counterclockwise
			direction = direction - 5;
			if(direction < 0) { direction += 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_WHOLE, pathMater);
			break;
		case 7:
			// Turn clockwise
			direction = direction - 1;
			if(direction < 0) { direction = direction + 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_WHOLE, pathMater);
			break;
	}
	randomTunnelPlot(hexPt.x, hexPt.y, direction, pathMater);
}

function joinTunnels(pathMater, fillMater) {
	// fill from initial seed point
	var seeking = true;
	var seekCol = 0;
	var seekRow = 0;
	while(seeking) {
		if(hexarray[seekCol][seekRow].bases[0].material == pathMater) {
			floodFill(hexarray, seekCol, seekRow, pathMater, fillMater);
			seeking = false;
		}
		if(seeking) {
			seekRow++;
			if(seekRow >= hexarray[seekCol].length) {
				seekRow = 0;
				seekCol++;
				if(seekCol >= hexarray.length) {
					seeking = false;
				}
			}
		}
	}
	var tunnelCount = countMaterialFills(pathMater);
	while(tunnelCount > 0) {
		var closestFillHex = new Point(0, 0);
		var closestTunnelHex = new Point(0, 0);
		var shortestDistance = hexarray.length * hexarray[0].length;
		var testDistance = shortestDistance;
		for(var x = 0; x < hexarray.length; x++) {
			for(var y = 0; y < hexarray[x].length; y++) {
				if(hexarray[x][y].bases[0].material == fillMater) {
					for(var x2 = 0; x2 < hexarray.length; x2++) {
						for(var y2 = 0; y2 < hexarray[x2].length; y2++) {
							if(hexarray[x2][y2].bases[0].material == pathMater) {
								testDistance = Math.abs(x - x2) + Math.abs(y - y2);
								if(testDistance < shortestDistance) {
									shortestDistance = testDistance;
									closestFillHex.x = x;
									closestFillHex.y = y;
									closestTunnelHex.x = x2;
									closestTunnelHex.y = y2;
								}
							}
						}
					}
				}
			}
		}
		connectTunnels(hexarray, closestFillHex.x, closestFillHex.y, closestTunnelHex.x, closestTunnelHex.y, pathMater, fillMater);
		tunnelCount = countMaterialFills(pathMater);
	}
}

function connectTunnels(dataarray, startX, startY, endX, endY, pathMater, fillMater) {
	var posX = startX;
	var posY = startY;
	var nextX = 0;
	var nextY = 0;
	var fillX = -1;
	var fillY = -1;
	while(!(posX == endX && posY == endY)) {
		if(startX == endX) {
			nextX = 0;
		}
		else {
			if(posX < endX) {
				nextX = 1;
			}
			else if(posX > endX) {
				nextX = -1;
			}
			else {
				nextX = 0;
			}
		}
		if(startY == endY) {
			nextY = 0;
		}
		else {
			if(posY < endY) {
				nextY = 1;
			}
			else if(posY > endY) {
				nextY = -1;
			}
			else {
				nextY = 0;
			}
		}
		if(nextX == 0 && nextY == 1) {
			posY++;
		}
		else if(nextX == 0 && nextY == -1) {
			posY--;
		}
		else {
			// different horizontal rules for odd and even hexes
			if(posX % 2 == 1) {
				if(nextX == -1 && nextY == -1) {
					posX--;
				}
				else if(nextX == -1 && nextY == 0) {
					posX--;
				}
				else if(nextX == -1 && nextY == 1) {
					posX--;
					posY++;
				}
				else if(nextX == 1 && nextY == -1) {
					posX++;
				}
				else if(nextX == 1 && nextY == 0) {
					posX++;
				}
				else if(nextX == 1 && nextY == 1) {
					posX++;
					posY++;
				}
			}
			else {
				if(nextX == -1 && nextY == -1) {
					posX--;
					posY--;
				}
				else if(nextX == -1 && nextY == 0) {
					posX--;
				}
				else if(nextX == -1 && nextY == 1) {
					posX--;
				}
				else if(nextX == 1 && nextY == -1) {
					posX++;
					posY--;
				}
				else if(nextX == 1 && nextY == 0) {
					posX++;
				}
				else if(nextX == 1 && nextY == 1) {
					posX++;
				}
			}
		}
		dataarray[posX][posY].addBase(HEX_WHOLE, pathMater);
		if(fillX == -1) { fillX = posX; }
		if(fillY == -1) { fillY = posY; }
	}
	floodFill(dataarray, fillX, fillY, pathMater, fillMater);
}

function displayRandomTunnelsSettings() {
	document.getElementById("overlay").className = "rndmaptoolsdiv";
	document.getElementById("shader").style.display = "block";
	document.getElementById("overlay").style.display = "block";
	document.getElementById("randomTunnelsSettings").style.display = "block";
}

function hideRandomTunnelsSettings() {
	document.getElementById("randomTunnelsSettings").style.display = "none";
	document.getElementById("overlay").style.display = "none";
	document.getElementById("shader").style.display = "none";
}

function floodFill(dataarray, posX, posY, matchMater, fillMater) {
	if(dataarray[posX][posY].bases[0].material == matchMater) {
		dataarray[posX][posY].addBase(HEX_WHOLE, fillMater);
	}
	else {
		return;
	}
	if(posX >= 1) {
		floodFill(dataarray, posX - 1, posY, matchMater, fillMater);
	}
	if(posY >= 1) {
		floodFill(dataarray, posX, posY - 1, matchMater, fillMater);
	}
	if(posX < dataarray.length - 1) {
		floodFill(dataarray, posX + 1, posY, matchMater, fillMater);
	}
	if(posY < dataarray[posX].length - 1) {
		floodFill(dataarray, posX, posY + 1, matchMater, fillMater);
	}
	if(posX % 2 == 0) {
		if(posX >= 1 && posY >= 1) {
			floodFill(dataarray, posX - 1, posY - 1, matchMater, fillMater);
		}
		if(posX < dataarray.length - 1 && posY >= 1) {
			floodFill(dataarray, posX + 1, posY - 1, matchMater, fillMater);
		}
	}
	else {
		if(posX >= 1 && posY < dataarray[posX].length - 1) {
			floodFill(dataarray, posX - 1, posY + 1, matchMater, fillMater);
		}
		if(posX < dataarray.length - 1 && posY < dataarray[posX].length - 1) {
			floodFill(dataarray, posX + 1, posY + 1, matchMater, fillMater);
		}
	}
}

function countMaterialFills(mater) {
	var mcount = 0;
	for(var x = 0; x < hexarray.length; x++) {
		for(var y = 0; y < hexarray[x].length; y++) {
			if(hexarray[x][y].bases[0].material == mater) {
				mcount++;
			}
		}
	}
	return mcount;
}

function fillFinder(mater) {
	for(var x = 0; x < hexarray.length; x++) {
		for(var y = 0; y < hexarray[x].length; y++) {
			if(hexarray[x][y].bases[0].material == mater) {
				return new Point(x, y);
			}
		}
	}
	return new Point(-1, -1);
}